/*
 *  linux/arch/nds32/include/asm/assembler.h
 *  Copyright (C) 2008 Andes Technology Corporation
 */

#ifndef __NDS32_ASSEMBLER_H__
#define __NDS32_ASSEMBLER_H__

/*
 * Stack format
 */

#define S_PAD           0xac /* Padding for stack pointer 8-byte alignment. */
#define S_FUCOP_CTL     0xa8
#define S_LP		0xa4
#define S_GP		0xa0
#define S_FP		0x9c
#define S_R25		0x98
#define S_R24		0x94
#define S_R23		0x90
#define S_R22		0x8c
#define S_R21		0x88
#define S_R20		0x84
#define S_R19		0x80
#define S_R18		0x7c
#define S_R17		0x78
#define S_R16		0x74
#define S_R15		0x70
#define S_R14		0x6c
#define S_R13		0x68
#define S_R12		0x64
#define S_R11		0x60
#define S_R10		0x5c
#define S_R9		0x58
#define S_R8		0x54
#define S_R7		0x50
#define S_R6		0x4c
#define S_R5		0x48
#define S_R4		0x44
#define S_R3		0x40
#define S_R2		0x3c
#define S_R1		0x38
#define S_R0		0x34
#define S_D1LO		0x30
#define S_D1HI		0x2c
#define S_D0LO		0x28
#define S_D0HI		0x24
#define S_PP1		0x20
#define S_PP0		0x1c
#define S_PIPC		0x18
#define S_PIPSW		0x14
#define S_ORIG_R0	0x10
#define S_SP		0xc
#define S_IPC		0x8
#define S_IPSW		0x4
#define S_IR0		0x0

#ifdef __ASSEMBLY__
.macro	get_thread_info, rd
	srli	\rd, $sp, #13
	slli	\rd, \rd, #13
.endm

.macro gie_disable
	setgie.d
	dsb
.endm

.macro gie_enable
	setgie.e
	dsb
.endm

.macro gie_save oldpsw
	mfsr \oldpsw, $ir0
	setgie.d
        dsb
.endm

.macro gie_restore oldpsw
	andi \oldpsw, \oldpsw, #0x1
	beqz \oldpsw, 7001f
	setgie.e
	dsb
7001:
.endm


#define USER(insn,  reg, addr, opr)	\
9999:	insn  reg, addr, opr;		\
	.section __ex_table,"a";	\
	.align 3;			\
	.long	9999b, 9001f;		\
	.previous

#else /* __ASSEMBLY__ */

__asm__ (".macro gie_disable\n\t"
         "setgie.d\n\t"
         "dsb\n\t"
         ".endm\n\t"
         );

__asm__ (".macro gie_enable\n\t"
         "setgie.e\n\t"
         "dsb\n\t"
         ".endm\n\t"
         );

#endif /* !__ASSEMBLY__ */
#endif /* __NDS32_ASSEMBLER_H__ */
