/*
 *            Copyright Andes Technology Corporation 2007-2008
 *                         All Rights Reserved.
 *
 *  Revision History:
 *
 *    Aug.21.2007     Created.
 *
 *  DESCRIPTION
 *
 *   DMA controller driver internal supplement library.
 *
 */

#ifndef __NDS_DMAD_ATF_INC__
#define __NDS_DMAD_ATF_INC__

/*****************************************************************************
 * Configuration section
*****************************************************************************/

/* Debug trace enable switch */
#define DMAD_ERROR_TRACE            1       /* message for fatal errors */
#define DMAD_DEBUG_TRACE            0       /* message for debug trace */

typedef u32 addr_t;

/*****************************************************************************
 * DMAC - AG101 AHB
*****************************************************************************/
/* Device base address */

#define DMAC_BASE		(ahb_base)
#define PMU_BASE		(pmu_base)

/* DMA controller registers (8-bit width) */
#define DMAC_INT                    (DMAC_BASE + 0x00)
#define DMAC_INT_TC                 (DMAC_BASE + 0x04)
#define DMAC_INT_TC_CLR             (DMAC_BASE + 0x08)
#define DMAC_INT_ERRABT             (DMAC_BASE + 0x0c)
#define DMAC_INT_ERRABT_CLR         (DMAC_BASE + 0x10)
#define DMAC_TC                     (DMAC_BASE + 0x14)
#define DMAC_ERRABT                 (DMAC_BASE + 0x18)
#define DMAC_CH_EN                  (DMAC_BASE + 0x1c)
#define DMAC_CH_BUSY                (DMAC_BASE + 0x20)
#define DMAC_CSR                    (DMAC_BASE + 0x24)
#define DMAC_SYNC                   (DMAC_BASE + 0x28)

/* DMA channel registers base address */
#define DMAC_C0_OFFSET              0x100
#define DMAC_C1_OFFSET              0x120
#define DMAC_C2_OFFSET              0x140
#define DMAC_C3_OFFSET              0x160
#define DMAC_C4_OFFSET              0x180
#define DMAC_C5_OFFSET              0x1a0
#define DMAC_C6_OFFSET              0x1c0
#define DMAC_C7_OFFSET              0x1e0

#define DMAC_C0_BASE                (DMAC_BASE + DMAC_C0_OFFSET)
#define DMAC_C1_BASE                (DMAC_BASE + DMAC_C1_OFFSET)
#define DMAC_C2_BASE                (DMAC_BASE + DMAC_C2_OFFSET)
#define DMAC_C3_BASE                (DMAC_BASE + DMAC_C3_OFFSET)
#define DMAC_C4_BASE                (DMAC_BASE + DMAC_C4_OFFSET)
#define DMAC_C5_BASE                (DMAC_BASE + DMAC_C5_OFFSET)
#define DMAC_C6_BASE                (DMAC_BASE + DMAC_C6_OFFSET)
#define DMAC_C7_BASE                (DMAC_BASE + DMAC_C7_OFFSET)

#define DMAC_MAX_CHANNELS           8
#define DMAC_BASE_CH(n)             (DMAC_C0_BASE + \
                                     (DMAC_C1_OFFSET - DMAC_C0_OFFSET) * \
                                     (addr_t)(n))  /* n = 0 ~ 3 */

#define DMAC_CSR_OFFSET             0x00
#define DMAC_CFG_OFFSET             0x04
#define DMAC_SRC_ADDR_OFFSET        0x08
#define DMAC_DST_ADDR_OFFSET        0x0c
#define DMAC_LLP_OFFSET             0x10
#define DMAC_SIZE_OFFSET            0x14

/* DMA channel 0 registers (32-bit width) */
#define DMAC_C0_CSR                 (DMAC_C0_BASE + DMAC_CSR_OFFSET)
#define DMAC_C0_CFG                 (DMAC_C0_BASE + DMAC_CFG_OFFSET)
#define DMAC_C0_SRC_ADDR            (DMAC_C0_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C0_DST_ADDR            (DMAC_C0_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C0_LLP                 (DMAC_C0_BASE + DMAC_LLP_OFFSET)
#define DMAC_C0_SIZE                (DMAC_C0_BASE + DMAC_SIZE_OFFSET)

/* DMA channel 1 registers (32-bit width) */
#define DMAC_C1_CSR                 (DMAC_C1_BASE + DMAC_CSR_OFFSET)
#define DMAC_C1_CFG                 (DMAC_C1_BASE + DMAC_CFG_OFFSET)
#define DMAC_C1_SRC_ADDR            (DMAC_C1_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C1_DST_ADDR            (DMAC_C1_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C1_LLP                 (DMAC_C1_BASE + DMAC_LLP_OFFSET)
#define DMAC_C1_SIZE                (DMAC_C1_BASE + DMAC_SIZE_OFFSET)

/* DMA channel 2 registers (32-bit width) */
#define DMAC_C2_CSR                 (DMAC_C2_BASE + DMAC_CSR_OFFSET)
#define DMAC_C2_CFG                 (DMAC_C2_BASE + DMAC_CFG_OFFSET)
#define DMAC_C2_SRC_ADDR            (DMAC_C2_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C2_DST_ADDR            (DMAC_C2_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C2_LLP                 (DMAC_C2_BASE + DMAC_LLP_OFFSET)
#define DMAC_C2_SIZE                (DMAC_C2_BASE + DMAC_SIZE_OFFSET)

/* DMA channel 3 registers (32-bit width) */
#define DMAC_C3_CSR                 (DMAC_C3_BASE + DMAC_CSR_OFFSET)
#define DMAC_C3_CFG                 (DMAC_C3_BASE + DMAC_CFG_OFFSET)
#define DMAC_C3_SRC_ADDR            (DMAC_C3_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C3_DST_ADDR            (DMAC_C3_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C3_LLP                 (DMAC_C3_BASE + DMAC_LLP_OFFSET)
#define DMAC_C3_SIZE                (DMAC_C3_BASE + DMAC_SIZE_OFFSET)

/* DMA channel 4 registers (32-bit width) */
#define DMAC_C4_CSR                 (DMAC_C4_BASE + DMAC_CSR_OFFSET)
#define DMAC_C4_CFG                 (DMAC_C4_BASE + DMAC_CFG_OFFSET)
#define DMAC_C4_SRC_ADDR            (DMAC_C4_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C4_DST_ADDR            (DMAC_C4_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C4_LLP                 (DMAC_C4_BASE + DMAC_LLP_OFFSET)
#define DMAC_C4_SIZE                (DMAC_C4_BASE + DMAC_SIZE_OFFSET)

/* DMA channel 5 registers (32-bit width) */
#define DMAC_C5_CSR                 (DMAC_C5_BASE + DMAC_CSR_OFFSET)
#define DMAC_C5_CFG                 (DMAC_C5_BASE + DMAC_CFG_OFFSET)
#define DMAC_C5_SRC_ADDR            (DMAC_C5_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C5_DST_ADDR            (DMAC_C5_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C5_LLP                 (DMAC_C5_BASE + DMAC_LLP_OFFSET)
#define DMAC_C5_SIZE                (DMAC_C5_BASE + DMAC_SIZE_OFFSET)

/* DMA channel 6 registers (32-bit width) */
#define DMAC_C6_CSR                 (DMAC_C6_BASE + DMAC_CSR_OFFSET)
#define DMAC_C6_CFG                 (DMAC_C6_BASE + DMAC_CFG_OFFSET)
#define DMAC_C6_SRC_ADDR            (DMAC_C6_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C6_DST_ADDR            (DMAC_C6_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C6_LLP                 (DMAC_C6_BASE + DMAC_LLP_OFFSET)
#define DMAC_C6_SIZE                (DMAC_C6_BASE + DMAC_SIZE_OFFSET)

/* DMA channel 7 registers (32-bit width) */
#define DMAC_C7_CSR                 (DMAC_C7_BASE + DMAC_CSR_OFFSET)
#define DMAC_C7_CFG                 (DMAC_C7_BASE + DMAC_CFG_OFFSET)
#define DMAC_C7_SRC_ADDR            (DMAC_C7_BASE + DMAC_SRC_ADDR_OFFSET)
#define DMAC_C7_DST_ADDR            (DMAC_C7_BASE + DMAC_DST_ADDR_OFFSET)
#define DMAC_C7_LLP                 (DMAC_C7_BASE + DMAC_LLP_OFFSET)
#define DMAC_C7_SIZE                (DMAC_C7_BASE + DMAC_SIZE_OFFSET)

/*****************************************************************************
 * DMAC defs - AG101 AHB
*****************************************************************************/

/* Interrupt status register (+00)  */
#define DMAC_INT0_MASK              0x01
#define DMAC_INT0_BIT               0
#define DMAC_INT1_MASK              0x02
#define DMAC_INT1_BIT               1
#define DMAC_INT2_MASK              0x04
#define DMAC_INT2_BIT               2
#define DMAC_INT3_MASK              0x08
#define DMAC_INT3_BIT               3

/* Interrupt for terminal count status register (+0x04) */
#define DMAC_INT_TC0_MASK           0x01
#define DMAC_INT_TC0_BIT            0
#define DMAC_INT_TC1_MASK           0x02
#define DMAC_INT_TC1_BIT            1
#define DMAC_INT_TC2_MASK           0x04
#define DMAC_INT_TC2_BIT            2
#define DMAC_INT_TC3_MASK           0x08
#define DMAC_INT_TC3_BIT            3
#define DMAC_INT_TC4_MASK           0x01
#define DMAC_INT_TC4_BIT            0
#define DMAC_INT_TC5_MASK           0x02
#define DMAC_INT_TC5_BIT            1
#define DMAC_INT_TC6_MASK           0x04
#define DMAC_INT_TC6_BIT            2
#define DMAC_INT_TC7_MASK           0x08
#define DMAC_INT_TC7_BIT            3

#define DMAC_INT_TC_MASK            0xff
#define DMAC_INT_TC_SHIFT           0

/* Interrupt for terminal count clear register (+0x08) */
#define DMAC_INT_TC0_CLR_MASK       0x01
#define DMAC_INT_TC0_CLR_BIT        0
#define DMAC_INT_TC1_CLR_MASK       0x02
#define DMAC_INT_TC1_CLR_BIT        1
#define DMAC_INT_TC2_CLR_MASK       0x04
#define DMAC_INT_TC2_CLR_BIT        2
#define DMAC_INT_TC3_CLR_MASK       0x08
#define DMAC_INT_TC3_CLR_BIT        3

#define DMAC_INT_TC_CLR_MASK        0x0f
#define DMAC_INT_TC_CLR_SHIFT       0

/* Interrupt for error/abort status register (+0x0c, 32-bits width) */
#define DMAC_INT_ERR0_MASK          0x00000001
#define DMAC_INT_ERR0_BIT           0
#define DMAC_INT_ERR1_MASK          0x00000002
#define DMAC_INT_ERR1_BIT           1
#define DMAC_INT_ERR2_MASK          0x00000004
#define DMAC_INT_ERR2_BIT           2
#define DMAC_INT_ERR3_MASK          0x00000008
#define DMAC_INT_ERR3_BIT           3

#define DMAC_INT_ERR_MASK           0x0000000f
#define DMAC_INT_ERR_SHIFT          0

#define DMAC_INT_ABT0_MASK          0x00010000
#define DMAC_INT_ABT0_BIT           16
#define DMAC_INT_ABT1_MASK          0x00020000
#define DMAC_INT_ABT1_BIT           17
#define DMAC_INT_ABT2_MASK          0x00040000
#define DMAC_INT_ABT2_BIT           18
#define DMAC_INT_ABT3_MASK          0x00080000
#define DMAC_INT_ABT3_BIT           19

#define DMAC_INT_ABT_MASK           0x000f0000
#define DMAC_INT_ABT_SHIFT          16

/* Interrupt for error/abort status clear register (+0x10, 32-bits width) */
#define DMAC_INT_ERR0_CLR_MASK      0x00000001
#define DMAC_INT_ERR0_CLR_BIT       0
#define DMAC_INT_ERR1_CLR_MASK      0x00000002
#define DMAC_INT_ERR1_CLR_BIT       1
#define DMAC_INT_ERR2_CLR_MASK      0x00000004
#define DMAC_INT_ERR2_CLR_BIT       2
#define DMAC_INT_ERR3_CLR_MASK      0x00000008
#define DMAC_INT_ERR3_CLR_BIT       3

#define DMAC_INT_ERR_CLR_MASK       0x0000000f
#define DMAC_INT_ERR_CLR_SHIFT      0

#define DMAC_INT_ABT0_CLR_MASK      0x00010000
#define DMAC_INT_ABT0_CLR_BIT       16
#define DMAC_INT_ABT1_CLR_MASK      0x00020000
#define DMAC_INT_ABT1_CLR_BIT       17
#define DMAC_INT_ABT2_CLR_MASK      0x00040000
#define DMAC_INT_ABT2_CLR_BIT       18
#define DMAC_INT_ABT3_CLR_MASK      0x00080000
#define DMAC_INT_ABT3_CLR_BIT       19

#define DMAC_INT_ABT_CLR_MASK       0x000f0000
#define DMAC_INT_ABT_CLR_SHIFT      16

/* Terminal count status register (+0x14) */
#define DMAC_TC0_MASK               0x01
#define DMAC_TC0_BIT                0
#define DMAC_TC1_MASK               0x02
#define DMAC_TC1_BIT                1
#define DMAC_TC2_MASK               0x04
#define DMAC_TC2_BIT                2
#define DMAC_TC3_MASK               0x08
#define DMAC_TC3_BIT                3

/* Error/abort status register (+0x18, 32-bits width) */
#define DMAC_ERR0_MASK              0x00000001
#define DMAC_ERR0_BIT               0
#define DMAC_ERR1_MASK              0x00000002
#define DMAC_ERR1_BIT               1
#define DMAC_ERR2_MASK              0x00000004
#define DMAC_ERR2_BIT               2
#define DMAC_ERR3_MASK              0x00000008
#define DMAC_ERR3_BIT               3

#define DMAC_ABT0_MASK              0x00010000
#define DMAC_ABT0_BIT               16
#define DMAC_ABT1_MASK              0x00020000
#define DMAC_ABT1_BIT               17
#define DMAC_ABT2_MASK              0x00040000
#define DMAC_ABT2_BIT               18
#define DMAC_ABT3_MASK              0x00080000
#define DMAC_ABT3_BIT               19

/* Channel enable status register (+0x1c) */
#define DMAC_CH0_EN_MASK            0x01
#define DMAC_CH0_EN_BIT             0
#define DMAC_CH1_EN_MASK            0x02
#define DMAC_CH1_EN_BIT             1
#define DMAC_CH2_EN_MASK            0x04
#define DMAC_CH2_EN_BIT             2
#define DMAC_CH3_EN_MASK            0x08
#define DMAC_CH3_EN_BIT             3

/* Channel busy status register (+0x20) */
#define DMAC_CH0_BUSY_MASK          0x01
#define DMAC_CH0_BUSY_BIT           0
#define DMAC_CH1_BUSY_MASK          0x02
#define DMAC_CH1_BUSY_BIT           1
#define DMAC_CH2_BUSY_MASK          0x04
#define DMAC_CH2_BUSY_BIT           2
#define DMAC_CH3_BUSY_MASK          0x08
#define DMAC_CH3_BUSY_BIT           3

/* Main configuration status register (+0x24) */
#define DMAC_DMACEN_MASK            0x01
#define DMAC_DMACEN_BIT             0
#define DMAC_M0ENDIAN_MASK          0x02
#define DMAC_M0ENDIAN_BIT           1
#define DMAC_M1ENDIAN_MASK          0x04
#define DMAC_M1ENDIAN_BIT           2

    #define DMAC_ENDIAN_LITTLE         0
    #define DMAC_ENDIAN_BIG            1

/* Sync register (+0x28) */
#define DMAC_SYNC0_MASK             0x01
#define DMAC_SYNC0_BIT              0
#define DMAC_SYNC1_MASK             0x02
#define DMAC_SYNC1_BIT              1
#define DMAC_SYNC2_MASK             0x04
#define DMAC_SYNC2_BIT              2
#define DMAC_SYNC3_MASK             0x08
#define DMAC_SYNC3_BIT              3

/* DMA channel 0~n Control Registers (CH[n]_BASE + 0x00) */
#define DMAC_CSR_CH_EN_MASK         0x00000001
#define DMAC_CSR_CH_EN_BIT          0

#define DMAC_CSR_DST_SEL_MASK       0x00000002
#define DMAC_CSR_DST_SEL_BIT        1
#define DMAC_CSR_SRC_SEL_MASK       0x00000004
#define DMAC_CSR_SRC_SEL_BIT        2
    #define DMAC_CSR_SEL_MASTER0          0x00
    #define DMAC_CSR_SEL_MASTER1          0x01

#define DMAC_CSR_DSTAD_CTL_MASK     0x00000018
#define DMAC_CSR_DSTAD_CTL_SHIFT    3
#define DMAC_CSR_SRCAD_CTL_MASK     0x00000060
#define DMAC_CSR_SRCAD_CTL_SHIFT    5
    #define DMAC_CSR_AD_INC               0x00
    #define DMAC_CSR_AD_DEC               0x01
    #define DMAC_CSR_AD_FIX               0x02

#define DMAC_CSR_MODE_MASK          0x00000080
#define DMAC_CSR_MODE_BIT           7
    #define DMAC_CSR_MODE_NORMAL          0x00
    #define DMAC_CSR_MODE_HSHK            0x01

#define DMAC_CSR_DST_WIDTH_MASK     0x00000700
#define DMAC_CSR_DST_WIDTH_SHIFT    8
#define DMAC_CSR_SRC_WIDTH_MASK     0x00003800
#define DMAC_CSR_SRC_WIDTH_SHIFT    11
    #define DMAC_CSR_WIDTH_8              0x00
    #define DMAC_CSR_WIDTH_16             0x01
    #define DMAC_CSR_WIDTH_32             0x02

#ifdef CONFIG_PLATFORM_AHBDMA
#define DMAC_CYCLE_TO_BYTES(cycle, width) ((cycle) << (width))
#define DMAC_BYTES_TO_CYCLE(bytes, width) ((bytes) >> (width))
#else
#define DMAC_CYCLE_TO_BYTES(cycle, width) 0
#define DMAC_BYTES_TO_CYCLE(bytes, width) 0
#endif  /* CONFIG_PLATFORM_AHBDMA */

#define DMAC_CSR_ABT                0x00008000
#define DMAC_CSR_ABT_BIT            15

#define DMAC_CSR_SRC_SIZE_MASK      0x00070000
#define DMAC_CSR_SRC_SIZE_SHIFT     16
    #define DMAC_CSR_SIZE_1              0x00
    #define DMAC_CSR_SIZE_4              0x01
    #define DMAC_CSR_SIZE_8              0x02
    #define DMAC_CSR_SIZE_16             0x03
    #define DMAC_CSR_SIZE_32             0x04
    #define DMAC_CSR_SIZE_64             0x05
    #define DMAC_CSR_SIZE_128            0x06
    #define DMAC_CSR_SIZE_256            0x07

#define DMAC_CSR_PROT1              0x00080000
#define DMAC_CSR_PROT1_BIT          19
#define DMAC_CSR_PROT2              0x00100000
#define DMAC_CSR_PROT2_BIT          20
#define DMAC_CSR_PROT3              0x00200000
#define DMAC_CSR_PROT3_BIT          21

#define DMAC_CSR_CHPRI_MASK         0x00c00000
#define DMAC_CSR_CHPRI_SHIFT        22
    #define DMAC_CSR_CHPRI_0              0x00
    #define DMAC_CSR_CHPRI_1              0x01
    #define DMAC_CSR_CHPRI_2              0x02
    #define DMAC_CSR_CHPRI_3              0x03

#define DMAC_CSR_FF_TH_MASK         0x07000000
#define DMAC_CSR_FF_TH_SHIFT        24
    #define DMAC_CSR_FF_TH_1              0x00
    #define DMAC_CSR_FF_TH_2              0x01
    #define DMAC_CSR_FF_TH_4              0x02
    #define DMAC_CSR_FF_TH_8              0x03
    #define DMAC_CSR_FF_TH_16             0x04

#define DMAC_CSR_TC_MSK_MSK         0x80000000
#define DMAC_CSR_TC_MSK_BIT         31

/* DMA channel 0~n Configuration Registers (CH[n]_BASE + 0x04) */
#define DMAC_CFG_INT_TC_MSK         0x00000001
#define DMAC_CFG_INT_TC_MSK_BIT     0
#define DMAC_CFG_INT_ERR_MSK        0x00000002
#define DMAC_CFG_INT_ERR_MSK_BIT    1
#define DMAC_CFG_INT_ABT_MSK        0x00000004
#define DMAC_CFG_INT_ABT_MSK_BIT    2

#define DMAC_CFG_INT_SRC_RS_MASK    0x00000078
#define DMAC_CFG_INT_SRC_RS_SHIFT   3
#define DMAC_CFG_INT_SRC_HE_MASK    0x00000080
#define DMAC_CFG_INT_SRC_HE_BIT     7

#define DMAC_CFG_BUSY_MASK          0x00000100
#define DMAC_CFG_BUSY_BIT           8

#define DMAC_CFG_INT_DST_RS_MASK    0x00001e00
#define DMAC_CFG_INT_DST_RS_SHIFT   9
#define DMAC_CFG_INT_DST_HE_MASK    0x00002000
#define DMAC_CFG_INT_DST_HE_BIT     13

#define DMAC_REQN_NONE          PMU_REQN_NONE
#define DMAC_REQN_CFC           PMU_REQN_CFC
#define DMAC_REQN_SSP           PMU_REQN_SSP
#define DMAC_REQN_UART1TX       PMU_REQN_UART1TX
#define DMAC_REQN_UART1RX       PMU_REQN_UART1RX
#define DMAC_REQN_UART2TX       PMU_REQN_UART2TX
#define DMAC_REQN_UART2RX       PMU_REQN_UART2RX
#define DMAC_REQN_SDC           PMU_REQN_SDC
#define DMAC_REQN_I2SAC97TX     PMU_REQN_I2SAC97TX
#define DMAC_REQN_I2SAC97RX     PMU_REQN_I2SAC97RX
#define DMAC_REQN_I2SAC97TX_AMERALD			PMU_REQN_I2SAC97TX_AMERALD
#define DMAC_REQN_I2SAC97RX_AMERALD			PMU_REQN_I2SAC97RX_AMERALD
#define DMAC_REQN_USB           PMU_REQN_USB
#define DMAC_REQN_EXT0          PMU_REQN_EXT0
#define DMAC_REQN_EXT1          PMU_REQN_EXT1
#define DMAC_REQN_MAX           PMU_REQN_MAX
#define DMAC_CFG_INT_LLPCNT_MASK    0x000f0000
#define DMAC_CFG_INT_LLPCNT_SHIFT   16

/* DMA channel 0~n Linked List Descriptor Registers (CH[n]_BASE + 0x10) */
#define DMAC_LLP_ADDR_MASK          0xfffffffc
#define DMAC_LLP_ADDR_SHIFT         2
#define DMAC_LLP_MASTER_MASK        0x00000001
#define DMAC_LLP_MASTER_BIT         0
#define DMAC_LLP_MASTER_0                0
#define DMAC_LLP_MASTER_1                1

/* DMA channel 0~3 Transfer Size Registers (CH[n]_BASE + 0x14) */
#define DMAC_TOT_SIZE_MASK          0x003fffff
#define DMAC_TOT_SIZE_SHIFT         0

/*****************************************************************************
 * APBBR - AG101 AHB to APB Bridge
*****************************************************************************/
/* Device base address */
#ifdef CONFIG_PLAT_AG102
#define APBBR_BASE                  APBBR_VA_BASE
#else
#define APBBR_BASE                  APBBRG_FTAPBBRG020S_0_VA_BASE
#endif

/* DMA channel A registers (32-bit width) */
#define APBBR_DMAA_BASE             (APBBR_BASE + 0x80)
#define APBBR_DMAB_BASE             (APBBR_BASE + 0x90)
#define APBBR_DMAC_BASE             (APBBR_BASE + 0xa0)
#define APBBR_DMAD_BASE             (APBBR_BASE + 0xb0)

#define APBBR_DMA_MAX_CHANNELS      APBBRG_FTAPBBRG020S_IRQ_COUNT
/* n = 0 ~ APBBRG_FTAPBBRG020S_IRQ_COUNT */
#define APBBR_DMA_BASE_CH(n)        (APBBR_DMAA_BASE + \
                                     (APBBR_DMAB_BASE - APBBR_DMAA_BASE) * \
                                     (addr_t)(n))

#define APBBR_DMA_SAD_OFFSET        0x00
#define APBBR_DMA_DAD_OFFSET        0x04
#define APBBR_DMA_CYC_OFFSET        0x08
#define APBBR_DMA_CMD_OFFSET        0x0c

/*****************************************************************************
 * APBBR defs - AG101 AHB to APB Bridge
*****************************************************************************/

/* APBBR slave#n (n = 1~6, 8, 11, 16~23) base/size register */
#define APBBR_SLAVE_SIZE_MASK       0x000f0000      /* Size of address space */
#define APBBR_SLAVE_SIZE_SHIFT      16
    #define APBBR_SIZE_1M                    0
    #define APBBR_SIZE_2M                    1
    #define APBBR_SIZE_4M                    2
    #define APBBR_SIZE_8M                    3
    #define APBBR_SIZE_16M                   4
    #define APBBR_SIZE_32M                   5
    #define APBBR_SIZE_64M                   6
    #define APBBR_SIZE_128M                  7
    #define APBBR_SIZE_256M                  8

#define APBBR_SLAVE_BASE_MASK       0x3ff00000
#define APBBR_SLAVE_BASE_SHIFT      20

/* APBBR DMA channel transfer cycles register
 *   DMA cycles (data size), 1 or 4 bus data transfer cycles per DMA cycle
 *   => transfer size = cycles * data_width * burst(1 or 4)
 *      so, max = 16M*4*4 = 256M
 */
#define APBBR_DMA_CYC_MASK          0x00ffffff
#define APBBR_DMA_CYC_SHIFT         0

/* APBBR DMA channel command register */
#define APBBR_DMA_CHEN_MASK         0x00000001
#define APBBR_DMA_CHEN_BIT          0

#define APBBR_DMA_FINTST_MASK       0x00000002
#define APBBR_DMA_FINTST_BIT        1
#define APBBR_DMA_FINTEN_MASK       0x00000004
#define APBBR_DMA_FINTEN_BIT        2

#define APBBR_DMA_BURST_MASK        0x00000008
#define APBBR_DMA_BURST_BIT         3

#define APBBR_DMA_ERRINTST_MASK     0x00000010
#define APBBR_DMA_ERRINTST_BIT      4
#define APBBR_DMA_ERRINTEN_MASK     0x00000020
#define APBBR_DMA_ERRINTEN_BIT      5

#define APBBR_DMA_SRCADDRSEL_MASK   0x00000040
#define APBBR_DMA_SRCADDRSEL_BIT    6
#define APBBR_DMA_DSTADDRSEL_MASK   0x00000080
#define APBBR_DMA_DSTADDRSEL_BIT    7
#define APBBR_ADDRSEL_APB                0
#define APBBR_ADDRSEL_AHB                1

#define APBBR_DMA_SRCADDRINC_MASK   0x00000700
#define APBBR_DMA_SRCADDRINC_SHIFT  8
#define APBBR_DMA_DSTADDRINC_MASK   0x00007000
#define APBBR_DMA_DSTADDRINC_SHIFT  12
#define APBBR_ADDRINC_FIXED              0      /* no increment */
#define APBBR_ADDRINC_I1X                1      /* +1,  +4 (burst) */
#define APBBR_ADDRINC_I2X                2      /* +2,  +8 (burst) */
#define APBBR_ADDRINC_I4X                3      /* +4, +16 (burst) */
#define APBBR_ADDRINC_D1                 5      /* -1 */
#define APBBR_ADDRINC_D2                 6      /* -2 */
#define APBBR_ADDRINC_D4                 7      /* -4 */

#define APBBR_DMA_DREQSEL_MASK      0x000f0000
#define APBBR_DMA_DREQSEL_SHIFT     16
#define APBBR_DMA_SREQSEL_MASK      0x0f000000
#define APBBR_DMA_SREQSEL_SHIFT     24

#define APBBR_REQN_NONE                  0
#define APBBR_REQN_CFC                   1
#define APBBR_REQN_SSP                   2
#define APBBR_REQN_SDC                   5
#define APBBR_REQN_SDC_AMERALD	     	 7
#define APBBR_REQN_I2SAC97TX_AMERALD	 8
#define APBBR_REQN_I2SAC97RX_AMERALD     9

#define APBBR_REQN_I2SAC97TX             6
#define APBBR_REQN_SSP2                  8
#define APBBR_REQN_STUART                9      /* UART1 ? */
#define APBBR_REQN_BTUART               10      /* UART2 ? */
#define APBBR_REQN_IRDA                 11
#define APBBR_REQN_SMMC                 12
#define APBBR_REQN_I2SAC97RX            13
#define APBBR_REQN_FUSB220              14
#define APBBR_REQN_MMSC                 15
#define APBBR_REQN_MAX                  15

#define APBBR_DMA_DATAWIDTH_MASK    0x00300000      /* Data width of transfer */
#define APBBR_DMA_DATAWIDTH_SHIFT   20
#define APBBR_DATAWIDTH_4                0      /* word */
#define APBBR_DATAWIDTH_2                1      /* half-word */
#define APBBR_DATAWIDTH_1                2      /* byte */

#ifdef CONFIG_PLATFORM_APBDMA
#define APBBR_DMA_CYCLE_TO_BYTES(cycle, width) ((cycle) << (2-(width)))
#define APBBR_DMA_BYTES_TO_CYCLE(bytes, width) ((bytes) >> (2-(width)))
#else
#define APBBR_DMA_CYCLE_TO_BYTES(cycle, width) 0
#define APBBR_DMA_BYTES_TO_CYCLE(bytes, width) 0
#endif  /* CONFIG_PLATFORM_APBDMA */



/*****************************************************************************
 * PMU - AG101 Core APB
*****************************************************************************/
/* Device base address */

/* PMU registers (32-bit width) */
#define PMU_AHBDMA_REQACK           (PMU_BASE + 0x90)

#define PMU_CFC_REQACK_CFG_OFF		(0xa0)
#define PMU_SSP1_REQACK_CFG_OFF		(0xa4)
#define PMU_UART1TX_REQACK_CFG_OFF	(0xa8)
#define PMU_UART1RX_REQACK_CFG_OFF	(0xac)
#define PMU_UART2TX_REQACK_CFG_OFF	(0xb0)
#define PMU_UART2RX_REQACK_CFG_OFF	(0xb4)
#define PMU_SDC_REQACK_CFG_OFF		(0xb8)
#define PMU_I2SAC97TX_REQACK_CFG_OFF	(0xbc)
#define PMU_I2SAC97RX_REQACK_CFG_OFF	(0xc4)
#define PMU_UART3TX_REQACK_CFG_OFF	(0xc0)
#define PMU_UART3RX_REQACK_CFG_OFF	(0xcc)
#define PMU_USB_REQACK_CFG_OFF		(0xc8)
#define PMU_IRDA_REQACK_CFG_OFF		(0xd0)
#define PMU_EXT0_REQACK_CFG_OFF		(0xd4)
#define PMU_EXT1_REQACK_CFG_OFF		(0xd8)
#define PMU_NONE			(0x00)

/*****************************************************************************
 * PMU - AG101 Core APB
*****************************************************************************/

/* Driving capability and slew rate control register 2 (+0x48) */
#define PMU_STUART_DCSR_MASK        0x0000000f
#define PMU_STUART_DCSR_SHIFT       0
#define PMU_BTUART_DCSR_MASK        0x00000f00
#define PMU_BTUART_DCSR_SHIFT       8
/*#define PMU_FFUART_DCSR_MASK        0x0000f000*/
/*#define PMU_FFUART_DCSR_SHIFT       12        */
#define PMU_PMU_DCSR_MASK           0x000f0000
#define PMU_PMU_DCSR_SHIFT          16
#define PMU_I2SAC97_DCSR_MASK       0x00f00000
#define PMU_I2SAC97_DCSR_SHIFT      20
#define PMU_SSP_DCSR_MASK           0x0f000000
#define PMU_SSP_DCSR_SHIFT          24
#define PMU_SD_DCSR_MASK            0xf0000000
#define PMU_SD_DCSR_SHIFT           28

/* AHB DMA REQ/ACK connection configuration status register (+0x90) */
#define PMU_CH0_REQACK_MASK         0x0000000f
#define PMU_CH0_REQACK_SHIFT        0
#define PMU_CH1_REQACK_MASK         0x000000f0
#define PMU_CH1_REQACK_SHIFT        4
#define PMU_CH2_REQACK_MASK         0x00000f00
#define PMU_CH2_REQACK_SHIFT        8
#define PMU_CH3_REQACK_MASK         0x0000f000
#define PMU_CH3_REQACK_SHIFT        12
#define PMU_CH4_REQACK_MASK         0x000f0000
#define PMU_CH4_REQACK_SHIFT        16
#define PMU_CH5_REQACK_MASK         0x00f00000
#define PMU_CH5_REQACK_SHIFT        20
#define PMU_CH6_REQACK_MASK         0x0f000000
#define PMU_CH6_REQACK_SHIFT        24
#define PMU_CH7_REQACK_MASK         0xf0000000
#define PMU_CH7_REQACK_SHIFT        28

#define PMU_REQN_NONE                    0
#define PMU_REQN_CFC                     1
#define PMU_REQN_SSP                     2
#define PMU_REQN_UART1TX                 3
#define PMU_REQN_UART1RX                 4
#define PMU_REQN_UART2TX                 5
#define PMU_REQN_UART2RX                 6
#define PMU_REQN_SDC                     7
#define PMU_REQN_I2SAC97TX               8
#define PMU_REQN_I2SAC97RX              10
#define PMU_REQN_I2SAC97TX_AMERALD		 8
#define PMU_REQN_I2SAC97RX_AMERALD		 9
#define PMU_REQN_USB                    11
#define PMU_REQN_EXT0                   14
#define PMU_REQN_EXT1                   15
#define PMU_REQN_MAX                    15

/* CFC ..., etc, REQ/ACK connection configuration registers (0xa0 ~ 0xd8) */
#define PMU_CHANNEL_MASK            0x00000007
#define PMU_CHANNEL_SHIFT           0
#define PMU_DMACUSED_MASK           0x00000008
#define PMU_DMACUSED_BIT            3
/*****************************************************************************
 * DMAD globals section
 */

enum DMAD_DMAC_CORE {
	DMAD_DMAC_AHB_CORE,
	DMAD_DMAC_APB_CORE
};

enum DMAD_CHREG_FLAGS {
	DMAD_FLAGS_NON_BLOCK    = 0x00000000,
	DMAD_FLAGS_SLEEP_BLOCK  = 0x00000001,
	DMAD_FLAGS_SPIN_BLOCK   = 0x00000002,
	DMAD_FLAGS_RING_MODE    = 0x00000008,   /* ring submission mode */
	DMAD_FLAGS_BIDIRECTION  = 0x00000010,   /* indicates both tx and rx */
};

enum DMAD_CHDIR
{
	DMAD_DIR_A0_TO_A1       = 0,
	DMAD_DIR_A1_TO_A0       = 1,
};

/* AHB Channel Request
 *
 * Notes for developers:
 *   These should be channel-only properties. Controller-specific properties
 *   should be separated as other driver structure or driver buildin-hardcode.
 *   If controller properties are embeded in this union, request for a channel
 *   may unexpectedly override the controller setting of the request of other
 *   channels.
 */
typedef struct dmad_ahb_chreq
{
	/* channel property */
	u32 sync;                       /* (in)  different clock domain */
	u32 priority;                   /* (in)  DMAC_CSR_CHPRI_xxx */
	u32 hw_handshake;               /* (in)  hardware handshaking on/off */
	u32 burst_size;                 /* (in)  DMAC_CSR_SIZE_xxx */

	/* source property */
	union {
		u32 src_width;          /* (in)  DMAC_CSR_WIDTH_xxx */
		u32 addr0_width;        /* (in)  bi-direction mode alias */
		u32 ring_width;         /* (in)  ring-mode alias */
	};
	union {
		u32 src_ctrl;           /* (in)  DMAC_CSR_AD_xxx */
		u32 addr0_ctrl;         /* (in)  bi-direction mode alias */
		u32 ring_ctrl;          /* (in)  ring-mode alias */
	};
	union {
		u32 src_reqn;           /* (in)  DMAC_REQN_xxx */
		u32 addr0_reqn;         /* (in)  bi-direction mode alias */
		u32 ring_reqn;          /* (in)  ring-mode alias */
	};

	/* destination property */
	union {
		u32 dst_width;          /* (in)  DMAC_CSR_WIDTH_xxx */
		u32 addr1_width;        /* (in)  bi-direction mode alias */
		u32 dev_width;          /* (in)  ring-mode alias */
	};
	union {
		u32 dst_ctrl;           /* (in)  DMAC_CSR_AD_xxx */
		u32 addr1_ctrl;         /* (in)  bi-direction mode alias */
		u32 dev_ctrl;           /* (in)  ring-mode alias */
	};
	union {
		u32 dst_reqn;           /* (in)  DMAC_REQN_xxx */
		u32 addr1_reqn;         /* (in)  bi-direction mode alias */
		u32 dev_reqn;           /* (in)  ring-mode alias */
	};

	/* (in)  transfer direction, valid only if following flags were set ...
	 *         DMAD_FLAGS_BIDIRECTION or
	 *         DMAD_FLAGS_RING_MODE
	 *       value:
	 *         0 (addr0 -> addr1, or ring-buff to device)
	 *         1 (addr0 <- addr1, or device to ring-buff)
	 */
	u32 tx_dir;
} dmad_ahb_chreq;

/* APB Channel Request
 *
 * Notes for developers:
 *   These should be channel-only properties. Controller-specific properties
 *   should be separated as other driver structure or driver buildin-hardcode.
 *   If controller properties are embeded in this union, request for a channel
 *   may unexpectedly override the controller setting of the request of other
 *   channels.
 */
typedef struct dmad_apb_chreq
{
	/* controller property (removed! should not exist in this struct) */

	/* channel property */
	u32 burst_mode;                 /* (in)  Burst mode (0/1) */
	u32 data_width;                 /* (in)  APBBR_DATAWIDTH_xxx */

	/* source property */
	union {
		u32 src_ctrl;           /* (in)  APBBR_ADDRINC_xxx */
		u32 addr0_ctrl;         /* (in)  bi-direction mode alias */
		u32 ring_ctrl;          /* (in)  ring-mode alias */
	};
	union {
		u32 src_reqn;           /* (in)  APBBR_REQN_xxx */
		u32 addr0_reqn;         /* (in)  bi-direction mode alias */
		u32 ring_reqn;          /* (in)  ring-mode alias */
	};

	/* destination property */
	union {
		u32 dst_ctrl;           /* (in)  APBBR_ADDRINC_xxx */
		u32 addr1_ctrl;         /* (in)  bi-direction mode alias */
		u32 dev_ctrl;           /* (in)  ring-mode alias */
	};
	union {
		u32 dst_reqn;           /* (in)  APBBR_REQN_xxx */
		u32 addr1_reqn;         /* (in)  bi-direction mode alias */
		u32 dev_reqn;           /* (in)  ring-mode alias */
	};

	/* (in)  transfer direction, valid only if following flags were set ...
	 *         DMAD_FLAGS_BIDIRECTION or
	 *         DMAD_FLAGS_RING_MODE
	 *       value:
	 *         0 (addr0 -> addr1, or ring-buff to device)
	 *         1 (addr0 <- addr1, or device to ring-buff)
	 */
	u32 tx_dir;

} dmad_apb_chreq;

/* Channel Request Descriptor */
typedef struct dmad_chreq
{
	/* common fields */
	u32     controller;                 /* (in)  enum DMAD_DMAC_CORE */
	u32	flags;                      /* (in)  enum DMAD_CHREQ_FLAGS */

	/**********************************************************************
	 * ring mode specific fields (valid only for DMAD_FLAGS_RING_MODE)
	 * note:
	 *  - size fields are in unit of data width
	 *    * for AHB, ring size is limited to 4K * data_width of data if
	 *      hw-LLP is not used
	 *    * for AHB, ring size is limited to 4K * data_width * LLP-count
	 *      hw-if LLP is used
	 *    * for APB, ring size is limited to 16M * data_width of data
	 *  - currently sw ring mode dma supports only fixed or incremental
	 *    src/dst addressing
	 *  - ring_size shoule >= periods * period_size
	 */
	dma_addr_t ring_base;               /* (in)  ring buffer base (pa) */
	dma_addr_t ring_size;               /* (in)  unit of data width */
	addr_t     dev_addr;                /* (in)  device data port address */
	dma_addr_t periods;                 /* (in)  number of ints per ring */
	dma_addr_t period_size;             /* (in)  size per int, data-width */


	/* channel-wise completion callback - called when hw-ptr catches sw-ptr
	 * (i.e., channel stops)
	 *
	 * completion_cb:   (in) client supplied callback function, executed in
	 *                       interrupt context.
	 * completion_data: (in) client private data to be passed to data
	 *                       argument of completion_cb().
	 */
	void (*completion_cb)(int channel, u16 status, void *data);
	void *completion_data;
	/*********************************************************************/

	/* channel allocation output */
	u32     channel;                    /* (out) allocated channel */
	void    *drq;                       /* (out) internal use (DMAD_DRQ *)*/

	/* channel-alloc parameters (channel-wise properties) */
	union {
#ifdef CONFIG_PLATFORM_AHBDMA
		dmad_ahb_chreq ahb_req;     /* (in)  for AHB DMA parameters */
#endif
#ifdef CONFIG_PLATFORM_APBDMA
		dmad_apb_chreq apb_req;     /* (in)  APB Bridge DMA params */
#endif
	};

} dmad_chreq;

/* drb states are mutual exclusive */
enum DMAD_DRB_STATE
{
	DMAD_DRB_STATE_FREE             = 0,
	DMAD_DRB_STATE_READY            = 0x00000001,
	DMAD_DRB_STATE_SUBMITTED        = 0x00000002,
	DMAD_DRB_STATE_EXECUTED         = 0x00000004,
	DMAD_DRB_STATE_COMPLETED        = 0x00000008,
	//DMAD_DRB_STATE_ERROR          = 0x00000010,
	DMAD_DRB_STATE_ABORT            = 0x00000020,
};

/* DMA request block
 * todo: replaced link with kernel struct list_head ??
 */
typedef struct dmad_drb
{
	u32  prev;                       /* (internal) previous node */
	u32  next;                       /* (internal) next node */
	u32  node;                       /* (internal) this node */

	u32  state;                      /* (out) DRB's current state */

	union {
		dma_addr_t src_addr;     /* (in)  source pa */
		dma_addr_t addr0;        /* (in)  bi-direction mode alias */
	};

	union {
		dma_addr_t dst_addr;     /* (in)  destination pa */
		dma_addr_t addr1;        /* (in)  bi-direction mode alias */
	};

	/* (in) AHB DMA (22 bits): 0 ~ 4M-1, unit is "data width"
	 *      APB DMA (24 bits): 0 ~ 16M-1, unit is "data width * burst size"
	 *      => for safe without mistakes, use dmad_make_req_cycles() to
	 *         compose this value if the addressing mode is incremental
	 *         mode (not working yet for decremental mode).
	 */
	dma_addr_t req_cycle;

	/* (in)  if non-null, this sync object will be signaled upon dma
	 * completion (for blocked-waiting dma completion)
	 */
	struct completion *sync;

} dmad_drb;


/******************************************************************************
 * Debug Trace Mechanism
 */
#if (DMAD_ERROR_TRACE)
#define dmad_err(format, arg...)  printk(KERN_ERR format , ## arg)
#else
#define dmad_err(format, arg...)  (void)(0)
#endif

#if (DMAD_DEBUG_TRACE)
#define dmad_dbg(format, arg...)  printk(KERN_INFO format , ## arg)
#else
#define dmad_dbg(format, arg...)  (void)(0)
#endif

#if (defined(CONFIG_PLATFORM_AHBDMA) || defined(CONFIG_PLATFORM_APBDMA))

/******************************************************************************
 * DMAD Driver Interface
******************************************************************************/

extern int dmad_channel_alloc(dmad_chreq *ch_req);
extern int dmad_channel_free(dmad_chreq *ch_req);
extern int dmad_channel_enable(const dmad_chreq *ch_req, u8 enable);
extern u32 dmad_max_size_per_drb(dmad_chreq *ch_req);
extern u32 dmad_bytes_to_cycles(dmad_chreq *ch_req, u32 byte_size);

extern int dmad_kickoff_requests(dmad_chreq *ch_req);
extern int dmad_drain_requests(dmad_chreq *ch_req, u8 shutdown);

/* for performance reason, these two functions are platform-specific */
#ifdef CONFIG_PLATFORM_AHBDMA
extern int dmad_probe_irq_source_ahb(void);
#endif
#ifdef CONFIG_PLATFORM_APBDMA
extern int dmad_probe_irq_source_apb(void);
#endif

/* note: hw_ptr here is phyical address of dma source or destination */
extern dma_addr_t dmad_probe_hw_ptr_src(dmad_chreq *ch_req);
extern dma_addr_t dmad_probe_hw_ptr_dst(dmad_chreq *ch_req);

/*****************************************************************************
 * routines only valid in discrete (non-ring) mode
 */
extern int dmad_config_channel_dir(dmad_chreq *ch_req, u8 dir);
extern int dmad_alloc_drb(dmad_chreq *ch_req, dmad_drb **drb);
extern int dmad_free_drb(dmad_chreq *ch_req, dmad_drb *drb);
extern int dmad_submit_request(dmad_chreq *ch_req,
			       dmad_drb *drb, u8 keep_fired);
extern int dmad_withdraw_request(dmad_chreq *ch_req, dmad_drb *drb);
/****************************************************************************/

/*****************************************************************************
 * routines only valid in ring mode
 * note: sw_ptr and hw_ptr are values offset from the ring buffer base
 *       unit of sw_ptr is data-width
 *       unit of hw_ptr returned is byte
 */
extern int dmad_update_ring(dmad_chreq *ch_req);
extern int dmad_update_ring_sw_ptr(dmad_chreq *ch_req,
				   dma_addr_t sw_ptr, u8 keep_fired);
extern dma_addr_t dmad_probe_ring_hw_ptr(dmad_chreq *ch_req);
/****************************************************************************/

#else  /* CONFIG_PLATFORM_AHBDMA || CONFIG_PLATFORM_APBDMA */

static inline int dmad_channel_alloc(dmad_chreq *ch_req) { return -EFAULT; }
static inline int dmad_channel_free(dmad_chreq *ch_req) { return -EFAULT; }
static inline int dmad_channel_enable(const dmad_chreq *ch_req, u8 enable)
	{ return -EFAULT; }
static inline u32 dmad_max_size_per_drb(dmad_chreq *ch_req) { return 0; }
static inline u32 dmad_bytes_to_cycles(dmad_chreq *ch_req, u32 byte_size)
	{ return 0; }
static inline int dmad_kickoff_requests(dmad_chreq *ch_req) { return -EFAULT; }
static inline int dmad_drain_requests(dmad_chreq *ch_req, u8 shutdown)
	{ return -EFAULT; }
static inline int dmad_probe_irq_source_ahb(void) { return -EFAULT; }
static inline int dmad_probe_irq_source_apb(void) { return -EFAULT; }
static inline dma_addr_t dmad_probe_hw_ptr_src(dmad_chreq *ch_req)
	{ return (dma_addr_t)NULL; }
static inline dma_addr_t dmad_probe_hw_ptr_dst(dmad_chreq *ch_req)
	{ return (dma_addr_t)NULL; }
static inline int dmad_config_channel_dir(dmad_chreq *ch_req, u8 dir)
	{ return -EFAULT; }
static inline int dmad_alloc_drb(dmad_chreq *ch_req, dmad_drb **drb)
	{ return -EFAULT; }
static inline int dmad_free_drb(dmad_chreq *ch_req, dmad_drb *drb)
	{ return -EFAULT; }
static inline int dmad_submit_request(dmad_chreq *ch_req,
       dmad_drb *drb, u8 keep_fired) { return -EFAULT; }
static inline int dmad_withdraw_request(dmad_chreq *ch_req, dmad_drb *drb)
	{ return -EFAULT; }
static inline int dmad_update_ring(dmad_chreq *ch_req)
	{ return -EFAULT; }
static inline int dmad_update_ring_sw_ptr(dmad_chreq *ch_req,
	dma_addr_t sw_ptr, u8 keep_fired) { return -EFAULT; }
static inline dma_addr_t dmad_probe_ring_hw_ptr(dmad_chreq *ch_req)
	{ return (dma_addr_t)NULL; }

#endif  /* CONFIG_PLATFORM_AHBDMA || CONFIG_PLATFORM_APBDMA */

#endif  /* __NDS_DMAD_ATF_INC__ */
