/*
 *  linux/arch/nds32/include/asm/atomic.h
 *  Copyright (C) 2008 Andes Technology Corporation
 */

#ifndef __NDS32_ATOMIC_H__
#define __NDS32_ATOMIC_H__

#include <linux/types.h>
#include <asm/barrier.h>
#include <asm/cmpxchg.h>

#if defined(CONFIG_SMP) || !defined(CONFIG_CPU_DCACHE_WRITETHROUGH)

#define ATOMIC_INIT(i)	{ (i) }
#define atomic_set(v,i)	(((v)->counter) = (i))
#define atomic_read(v) (*(volatile int *)&(v)->counter)

static inline void atomic_add(int i, atomic_t *v)
{
	int temp;
	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%0, [%1+$r15]\n"
		"\tadd\t%0, %0, %2\n"
		"\tscw\t%0, [%1+$r15]\n"
		"\tbeqz\t%0, 1b\n"
		: "=&r" (temp) : "r" (&v->counter), "r" (i));
}

static inline void atomic_sub(int i, atomic_t *v)
{
	int temp;
	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%0, [%1+$r15]\n"
		"\tsub\t%0, %0, %2\n"
		"\tscw\t%0, [%1+$r15]\n"
		"\tbeqz\t%0, 1b\n"
		: "=&r" (temp) : "r" (&v->counter), "r" (i));
}

#define atomic_inc(v)       atomic_add(1, v)
#define atomic_dec(v)       atomic_sub(1, v)

static inline int atomic_inc_return(atomic_t *v)
{
	int temp1, temp2;
	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%1, [%2+$r15]\n"
		"\taddi\t%1, %1, #1\n"
		"\tori\t%0, %1, #0\n"
		"\tscw\t%1, [%2+$r15]\n"
		"\tbeqz\t%1, 1b\n"
		: "=&r" (temp2), "=&r" (temp1)
		: "r" (&v->counter));
	return temp2;
}

static inline int atomic_dec_return(atomic_t *v)
{
	int temp1, temp2;
	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%1, [%2+$r15]\n"
		"\taddi\t%1, %1, #-1\n"
		"\tori\t%0, %1, #0\n"
		"\tscw\t%1, [%2+$r15]\n"
		"\tbeqz\t%1, 1b\n"
		: "=&r" (temp2), "=&r" (temp1)
		: "r" (&v->counter));
	return temp2;
}

static inline int atomic_add_return(int i, atomic_t *v)
{
	int temp1, temp2;

	smp_mb();

	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%1, [%2+$r15]\n"
		"\tadd\t%1, %1, %3\n"
		"\tori\t%0, %1, #0\n"
		"\tscw\t%1, [%2+$r15]\n"
		"\tbeqz\t%1, 1b\n"
		: "=&r" (temp2), "=&r" (temp1)
		: "r" (&v->counter), "r" (i));

	smp_mb();

	return temp2;
}

static inline int atomic_sub_return(int i, atomic_t *v)
{
	int temp1, temp2;

	smp_mb();

	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%1, [%2+$r15]\n"
		"\tsub\t%1, %1, %3\n"
		"\tori\t%0, %1, #0\n"
		"\tscw\t%1, [%2+$r15]\n"
		"\tbeqz\t%1, 1b\n"
		: "=&r" (temp2), "=&r" (temp1)
		: "r" (&v->counter), "r" (i));

	smp_mb();

	return temp2;
}


#define atomic_inc_and_test(v)  	(atomic_add_return(1, v) == 0)
#define atomic_dec_and_test(v)		(atomic_sub_return(1, v) == 0)
#define atomic_sub_and_test(i,v)	(atomic_sub_return((i),(v)) == 0)
#define atomic_add_negative(i,v)	(atomic_add_return((i),(v)) < 0)


static inline int atomic_cmpxchg(atomic_t *v, int old, int new)
{
	int temp1, temp2, temp3;

	smp_mb();

	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%0, [%3+$r15]\n"
		"\tsub\t%2, %0, %5\n"
		"\tcmovz\t%1, %4, %2\n"
		"\tcmovn\t%1, %0, %2\n"
		"\tscw\t%1, [%3+$r15]\n"
		"\tbeqz\t%1, 1b\n"
		: "=&r" (temp3), "=&r" (temp2), "=&r" (temp1)
		: "r" (&v->counter), "r" (new), "r" (old));

	smp_mb();

	return temp3;
}

static inline int atomic_xchg(atomic_t *v, int new)
{
	int temp1, temp2;
	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%0, [%2+$r15]\n"
		"\tori\t%1, %3, #0\n"
		"\tscw\t%1, [%2+$r15]\n"
		"\tbeqz\t%1, 1b\n"
		: "=&r" (temp2), "=&r" (temp1)
		: "r" (&v->counter), "r" (new));
	return temp2;
}

static inline int __atomic_add_unless(atomic_t *v, int a, int u)
{
    int c, old;

    c = atomic_read(v);
    while (c != u && (old = atomic_cmpxchg((v), c, c + a)) != c)
        c = old;
    return c;
}

static inline int atomic_inc_not_zero(atomic_t *v)
{
	int temp1, temp2;
	__asm__ __volatile__(
		"movi\t$r15, #0\n"
		"1:\n"
		"\tllw\t%1, [%2+$r15]\n"
		"\tslt\t%0, $r15, %1\n"
		"\tadd\t%1, %1, %0\n"
		"\tscw\t%1, [%2+$r15]\n"
		"\tbeqz\t%1, 1b\n"
		: "=&r" (temp2), "=&r" (temp1)
		: "r" (&v->counter));
	return temp2;
}

#define smp_mb__before_atomic_dec()	barrier()
#define smp_mb__after_atomic_dec()	barrier()
#define smp_mb__before_atomic_inc()	barrier()
#define smp_mb__after_atomic_inc()	barrier()

#else /* CONFIG_SMP*/



#define ATOMIC_INIT(i)	{ (i) }

#ifdef __KERNEL__

#include <asm-generic/atomic.h>

static inline int atomic_dec_if_positive(atomic_t * v)
{
	unsigned long flags;
	int result;

	local_irq_save(flags);
	result = v->counter;
	result -= 1;
	if (result >= 0)
		v->counter = result;
	local_irq_restore(flags);
	return result;
}

/* Atomic operations are already serializing on ARM */
#define smp_mb__before_atomic_dec()	barrier()
#define smp_mb__after_atomic_dec()	barrier()
#define smp_mb__before_atomic_inc()	barrier()
#define smp_mb__after_atomic_inc()	barrier()

#endif

#endif /* CONFIG_SMP */

#ifndef CONFIG_GENERIC_ATOMIC64
typedef struct {
	    u64 __aligned(8) counter;
} atomic64_t;
#define ATOMIC64_INIT(i) { (i) }
static inline u64 atomic64_add_return(u64 i, atomic64_t *v){ return 0; }
#endif /* CONFIG_GENERIC_ATOMIC64 */

#endif /* __NDS32_ATOMIC_H__ */
