/*
 *  linux/arch/nds32/include/asm/fpu.h
 *  Copyright (C) 2008 Andes Technology Corporation
 */

#ifndef __ASM_NDS32_FPU_H
#define __ASM_NDS32_FPU_H

#ifndef __ASSEMBLY__
#include <linux/preempt.h>
#include <asm/ptrace.h>

extern void save_fpu(struct task_struct *__tsk);
extern void fpload(const struct fpu_struct *fpregs);
extern void do_fpu_exception(unsigned long error_code, struct pt_regs *regs,
			     unsigned long eva);
extern int do_fpu_inst(unsigned short, struct pt_regs *);

#ifdef CONFIG_FPU

#define test_tsk_fpu(regs)	(regs->NDS32_FUCOP_CTL & FUCOP_CTL_mskCP0EN)

struct task_struct;

static inline void release_fpu(struct pt_regs *regs)
{
	regs->NDS32_FUCOP_CTL &= ~FUCOP_CTL_mskCP0EN;
}

static inline void grab_fpu(struct pt_regs *regs)
{
	regs->NDS32_FUCOP_CTL |= FUCOP_CTL_mskCP0EN;
}

static inline void enable_fpu(void)
{
        SET_FUCOP_CTL(GET_FUCOP_CTL() | FUCOP_CTL_mskCP0EN);
}

static inline void disable_fpu(void)
{
        SET_FUCOP_CTL(GET_FUCOP_CTL() & ~FUCOP_CTL_mskCP0EN);
}

static inline void lose_fpu(int save)
{
	preempt_disable();
#ifndef CONFIG_UNLAZY_FPU
	if (last_task_used_math == current) {
#else
	if (test_tsk_fpu(task_pt_regs(current))) {
#endif
		if (save)
		{
			save_fpu(current);
# ifndef CONFIG_UNLAZY_FPU
			last_task_used_math=NULL;
# endif
		}
	}
	release_fpu(task_pt_regs(current));
	preempt_enable();
}

static inline void own_fpu(int restore)
{
	preempt_disable();
#ifndef CONFIG_UNLAZY_FPU
	if (last_task_used_math != current) {
#else
	if (!test_tsk_fpu(task_pt_regs(current))) {
#endif
		if (restore)
		{
# ifdef CONFIG_UNLAZY_FPU
			fpload(&current->thread.fpu);
# else
			if(last_task_used_math!=NULL)
				save_fpu(last_task_used_math);
			fpload(&current->thread.fpu);
			last_task_used_math=current;
#endif
		}
	}
	grab_fpu(task_pt_regs(current));
	preempt_enable();
}
# ifdef CONFIG_UNLAZY_FPU
static inline void unlazy_fpu(struct task_struct *tsk)
{
	preempt_disable();
	if (test_tsk_fpu(task_pt_regs(tsk)))
		save_fpu(tsk);
	preempt_enable();
}
# endif /* CONFIG_UNLAZY_FPU */
static inline void clear_fpu(struct pt_regs *regs)
{
	preempt_disable();
	if (test_tsk_fpu(regs)) {
		release_fpu(regs);
	}
	preempt_enable();
}
#endif /* CONFIG_FPU */
#endif /* __ASSEMBLY__ */

#endif /* __ASM_NDS32_FPU_H */
