/* ============================================================================
 *
 *  linux/arch/nds32/include/asm/kgdb.h
 *
 *  Copyright (C) 2007 Andes Technology Corporation
 *  This file is part of Linux and should be licensed under the GPL.
 *  See the file COPYING for conditions for redistribution.
 *
 *  Abstract:
 *
 *    This program is for NDS32 KGDB support.
 *
 *  Author: Harry Pan
 *
 *  Revision History:
 *
 *    Jul.14.2007     Initial ported by Harry.
 *
 *  Note:
 *
 * ============================================================================
 */
#ifndef __ASM_NDS32_KGDB_H__
#define __ASM_NDS32_KGDB_H__

#include <asm/ptrace.h>

#define BREAK_INSTR_SIZE        2
#define CACHE_FLUSH_IS_SAFE     1

#ifndef __ASSEMBLY__

/*
 * Define numbers of registers we have in NDS32 arch
 */
#define NDS32_NUM_GR    32      // general registers.
#define NDS32_NUM_SPR   1       // special registers. (PC)
#define NDS32_NUM_CR    6       // ctrl registers.
#define NDS32_NUM_IR    16      // interruption registers.
#define NDS32_NUM_MR    11      // MMU registers.
#define NDS32_NUM_DR    48      // debug registers.
#define NDS32_NUM_PFR   4       // performance monitoring registers.
#define NDS32_NUM_DMAR  11      // local memory DMA registers
#define NDS32_NUM_RACR  1       // resource access control registers.
#define NDS32_NUM_IDR   2       // implementation dependent registers.
#define NDS32_NUM_SR    (NDS32_NUM_CR + NDS32_NUM_IR + NDS32_NUM_MR + \
                         NDS32_NUM_DR + NDS32_NUM_PFR + NDS32_NUM_DMAR + \
                         NDS32_NUM_RACR + NDS32_NUM_IDR)
#define NDS32_NUM_REGS  (NDS32_NUM_GR + NDS32_NUM_SPR)
#define DBG_MAX_REG_NUM  NDS32_NUM_REGS

#define KGDB_MAX_NO_CPUS        1
#define BUFMAX                  2048
#define NUMREGBYTES             (NDS32_NUM_REGS << 2)

/*
 * NDS32 virtual registers layout for GDB.
 */
enum nds32_regnum
{
    NDS32_R0_REGNUM = 0,    // first integer-like argument.
    NDS32_R5_REGNUM = 5,    // last integer-like argument.
    NDS32_FP_REGNUM = 28,   // frame register
    NDS32_LP_REGNUM = 30,   // link pointer
    NDS32_SP_REGNUM = 31,   // address of stack top.
    NDS32_PC_REGNUM = 32,
    NDS32_CR0_REGNUM = 33,
    NDS32_IR0_REGNUM = (NDS32_CR0_REGNUM + NDS32_NUM_CR),
    NDS32_MR0_REGNUM = (NDS32_IR0_REGNUM + NDS32_NUM_IR),
    NDS32_DR0_REGNUM = (NDS32_MR0_REGNUM + NDS32_NUM_MR),
    NDS32_PFR0_REGNUM = (NDS32_DR0_REGNUM + NDS32_NUM_DR),
    NDS32_DMAR0_REGNUM = (NDS32_PFR0_REGNUM + NDS32_NUM_PFR),
    NDS32_RACR0_REGNUM = (NDS32_DMAR0_REGNUM + NDS32_NUM_DMAR),
    NDS32_IDR0_REGNUM = (NDS32_RACR0_REGNUM + NDS32_NUM_RACR),
    /* nds32 calling convention. */
    NDS32_ARG0_REGNUM = NDS32_R0_REGNUM,
    NDS32_ARGN_REGNUM = NDS32_R5_REGNUM,
    NDS32_RET_REGNUM = NDS32_R0_REGNUM,
};

extern void compiled_breakinst(void);
extern void arch_kgdb_breakpoint(void);



#endif /* !__ASSEMBLY__ */
#endif /* __ASM_NDS32_KGDB_H__ */
