/*
 * linux/arch/nds32/include/asm/nds32.h -- Andes NDS32 processor register interface
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2006 Andes Technology Corporation
 *
 */
#ifndef _ASM_NDS32_NDS32_H_
#define _ASM_NDS32_NDS32_H_

#include <asm/bitfield.h>
#include <asm/reg_access.h>
#include <asm/cachectl.h>

#define MSYNC( subtype) __asm__ ("\n\tmsync "#subtype);
#define STANDBY( cond) __asm__ ("\n\tstandby "#cond);

#ifndef __ASSEMBLY__

static inline void ISB( void) { __asm__ ("\n\tisb"); }
static inline void DSB( void) { __asm__ ("\n\tdsb"); }

static inline void GIE_ENABLE( void)
{
	__asm__ ("gie_enable\n\t"
		 :
		 :
		 :"memory");
}

static inline void GIE_DISABLE( void)
{
        __asm__ ("gie_disable\n\t"
		 :
		 :
		 :"memory");
}

static inline unsigned long CACHE_SET(unsigned char cache){

	if( cache == ICACHE)
		return 64 << ( ( GET_ICM_CFG() & ICM_CFG_mskISET) >> ICM_CFG_offISET);
	else
		return 64 << ( ( GET_DCM_CFG() & DCM_CFG_mskDSET) >> DCM_CFG_offDSET);
}

static inline unsigned long CACHE_WAY(unsigned char cache){

	if( cache == ICACHE)
		return 1 + ( ( GET_ICM_CFG() & ICM_CFG_mskIWAY) >> ICM_CFG_offIWAY);
	else
		return 1 + ( ( GET_DCM_CFG() & DCM_CFG_mskDWAY) >> DCM_CFG_offDWAY);
}

static inline unsigned long CACHE_LINE_SIZE(unsigned char cache){

	if( cache == ICACHE)
		return 8 << ( ( ( GET_ICM_CFG() & ICM_CFG_mskISZ) >> ICM_CFG_offISZ) - 1);
	else
		return 8 << ( ( ( GET_DCM_CFG() & DCM_CFG_mskDSZ) >> DCM_CFG_offDSZ) - 1);
}

static inline void GIE_SAVE( unsigned long *var){

	*var = GET_PSW();
	GIE_DISABLE();
}

static inline void GIE_RESTORE( unsigned long var){

	if( var & PSW_mskGIE){
		GIE_ENABLE();
	}
}

#endif /* __ASSEMBLY__ */

#define IVB_BASE		PHYS_OFFSET	/* in user space for intr/exc/trap/break table base, 64KB aligned 
						 * We defined at the start of the physical memory */

/* The following dispatching entry */
#define ENTRY_TLB_MISC		(IVB_BASE + nrTLB_MISC*vENTRY_SZ)	/* TLB misc eh# */
/* dispatched sub-entry exception handler numbering */
#define RD_PROT			0  /* read protrection */
#define WRT_PROT		1  /* write protection */
#define NOEXEC			2  /* non executable */
#define PAGE_MODIFY		3  /* page modified */
#define ACC_BIT			4  /* access bit */
#define RESVED_PTE		5  /* reserved PTE attribute */
/* reserved 6 ~ 16 */

#endif /* _ASM_NDS32_NDS32_H_ */
