/*
 *  linux/arch/nds32/include/asm/page.h
 *
 *  Copyright (C) 1995-2003 Russell King
 *  Copyright (C) 2008 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#ifndef _ASMNDS32_PAGE_H
#define _ASMNDS32_PAGE_H


#ifdef CONFIG_ANDES_PAGE_SIZE_4KB
#define PAGE_SHIFT      12
#endif
#ifdef CONFIG_ANDES_PAGE_SIZE_8KB
#define PAGE_SHIFT      13
#endif
#include <linux/const.h>
#define PAGE_SIZE       (_AC(1,UL) << PAGE_SHIFT)
#define PAGE_MASK       (~(PAGE_SIZE-1))
#define PTE_MASK        PAGE_MASK

/* PAGE_SHIFT determines the page size */

#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_16KB
#define LARGE_PAGE_SHIFT        14
#define HPAGE_SHIFT     14
#endif

#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_64KB
#define LARGE_PAGE_SHIFT        16
#define HPAGE_SHIFT     16
#endif

#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_256KB
#define LARGE_PAGE_SHIFT        18
#define HPAGE_SHIFT     18
#endif
#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_1MB
#define LARGE_PAGE_SHIFT        20
#define HPAGE_SHIFT     20
#endif

#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_4MB
#define LARGE_PAGE_SHIFT        22
#define HPAGE_SHIFT     22
#endif

#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_16MB
#define LARGE_PAGE_SHIFT        24
#define HPAGE_SHIFT     24
#endif

#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_64MB
#define LARGE_PAGE_SHIFT        26
#define HPAGE_SHIFT     26
#endif

#ifdef CONFIG_ANDES_HUGETLB_PAGE_SIZE_256MB
#define LARGE_PAGE_SHIFT        28
#define HPAGE_SHIFT     28
#endif

#ifdef CONFIG_HUGETLB_PAGE
// taken for i386 style
#define LARGE_PAGE_SIZE (1UL << LARGE_PAGE_SHIFT)
#define LARGE_PAGE_MASK (~(LARGE_PAGE_SIZE-1))
// taken for SH style
#define HPAGE_SIZE              (1UL << HPAGE_SHIFT)
#define HPAGE_MASK              (~(HPAGE_SIZE-1))
#define HUGETLB_PAGE_ORDER      (HPAGE_SHIFT-PAGE_SHIFT)
#endif


#ifdef __KERNEL__

#ifndef __ASSEMBLY__

struct page;
struct vm_area_struct;
#ifndef CONFIG_CPU_CACHE_NONALIASING
extern void copy_user_highpage(struct page *to, struct page *from,
                               unsigned long vaddr, struct vm_area_struct *vma);
extern void clear_user_highpage(struct page *page, unsigned long vaddr);

#define __HAVE_ARCH_COPY_USER_HIGHPAGE
#define clear_user_highpage	clear_user_highpage
#else
#define clear_user_page(page, vaddr, pg)        clear_page(page)
#define copy_user_page(to, from, vaddr, pg)     copy_page(to, from)
#endif

void clear_page(void *page);
void copy_page(void *to, void *from);

#undef STRICT_MM_TYPECHECKS

#ifdef STRICT_MM_TYPECHECKS
/*
 * These are used to make use of C type-checking..
 */
typedef struct { unsigned long pte; } pte_t;
typedef struct { unsigned long pmd; } pmd_t;
typedef struct { unsigned long pgd; } pgd_t;
typedef struct { unsigned long pgprot; } pgprot_t;

#define pte_val(x)      ((x).pte)
#define pmd_val(x)      ((x).pmd)
#define pgd_val(x)	((x).pgd)
#define pgprot_val(x)   ((x).pgprot)

#define __pte(x)        ((pte_t) { (x) } )
#define __pmd(x)        ((pmd_t) { (x) } )
#define __pgd(x)	((pgd_t) { (x) } )
#define __pgprot(x)     ((pgprot_t) { (x) } )

#else
/*
 * .. while these make it easier on the compiler
 */
typedef unsigned long pte_t;
typedef unsigned long pmd_t;
typedef unsigned long pgd_t;
typedef unsigned long pgprot_t;

#define pte_val(x)      (x)
#define pmd_val(x)      (x)
#define pgd_val(x)	(x)
#define pgprot_val(x)   (x)

#define __pte(x)        (x)
#define __pmd(x)        (x)
#define __pgd(x)        (x)
#define __pgprot(x)     (x)

#endif /* STRICT_MM_TYPECHECKS */
typedef struct page *pgtable_t;

#include <asm/memory.h>
#include <asm-generic/getorder.h>

#endif /* !__ASSEMBLY__ */

#define VM_DATA_DEFAULT_FLAGS	(VM_READ | VM_WRITE | VM_EXEC | \
				 VM_MAYREAD | VM_MAYWRITE | VM_MAYEXEC)

#endif /* __KERNEL__ */

#endif
