/*
 * linux/arch/nds32/include/asm/processor.h
 */
/* Copyright (C) 1995-1999 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
/* ============================================================================
 *  Copyright (C) 2007 Andes Technology Corporation
 *  This file is part of Linux and should be licensed under the GPL.
 *  See the file COPYING for conditions for redistribution.
 *
 *  Abstract:
 *
 *    This program is ptrace relative code for Andes NDS32 architecture.
 *    Original referred from ARM, fit to NDS32.
 *
 *  Revision History:
 *
 *    Oct.03.2007     Original from Tom, Shawn and Steven, refined by Harry.
 *
 *  Note:
 *
 * ============================================================================
 */
#ifndef __ASM_NDS32_PROCESSOR_H
#define __ASM_NDS32_PROCESSOR_H

/*
 * Default implementation of macro that returns current
 * instruction pointer ("program counter").
 */
#define current_text_addr() ({ __label__ _l; _l: &&_l;})

#ifdef __KERNEL__

#include <asm/ptrace.h>
#include <asm/procinfo.h>
#include <asm/types.h>
#include <asm/cpuver.h>
#include <asm/sigcontext.h>

#define KERNEL_STACK_SIZE	PAGE_SIZE
#define STACK_TOP	TASK_SIZE
#define STACK_TOP_MAX   TASK_SIZE

struct debug_info {
	u32	address;
	u16	insn;
	u8	valid;
};

struct thread_struct {
							/* fault info	  */
	unsigned long		address;
	unsigned long		trap_no;
	unsigned long		error_code;

        struct fpu_struct fpu;				/* Saved fpu/fpu emulator stuff. */
	struct debug_info	debug;			/* debugging      */
};

#define INIT_THREAD  {	}

#ifdef __NDS32_EB__
#define PSW_DE	PSW_mskBE
#else
#define PSW_DE	0x0
#endif

#ifdef CONFIG_WBNA
#define PSW_valWBNA	PSW_mskWBNA
#else
#define PSW_valWBNA	0x0
#endif

#ifdef CONFIG_HWZOL
#define	PSW_valINIT (PSW_CPL_ANY | PSW_mskAEN | PSW_valWBNA | PSW_mskDT | PSW_mskIT | PSW_DE | PSW_mskGIE)
#else
#define	PSW_valINIT (PSW_CPL_ANY | PSW_valWBNA | PSW_mskDT | PSW_mskIT | PSW_DE | PSW_mskGIE)
#endif

#define start_thread(regs,pc,sp)					\
({									\
	unsigned long *stack = (unsigned long *)sp;			\
	set_fs(USER_DS);						\
	memzero(regs->uregs, sizeof(regs->uregs));			\
	regs->NDS32_ipsw = PSW_valINIT;					\
	regs->NDS32_ir0 = (PSW_CPL_ANY | PSW_valWBNA | PSW_mskDT | PSW_mskIT | PSW_DE | PSW_SYSTEM | PSW_INTL_1);			\
	regs->NDS32_ipc = pc;		/* pc */			\
	regs->NDS32_sp = sp;		/* $sp */			\
	regs->NDS32_r2 = stack[4];	/* $r2 (envp) */		\
	regs->NDS32_r1 = stack[1];	/* $r1 (argv) */		\
	regs->NDS32_r0 = stack[0];	/* $r0 (argc) */		\
})


/* Forward declaration, a strange C thing */
struct task_struct;

/* Free all resources held by a thread. */
extern void release_thread(struct task_struct *);
#ifdef CONFIG_FPU
#ifndef CONFIG_UNLAZU_FPU //lazy fpu
extern struct task_struct *last_task_used_math;
#endif
#endif

/* Prepare to copy thread state - unlazy all lazy status */
#define prepare_to_copy(tsk)	do { } while (0)

unsigned long get_wchan(struct task_struct *p);

#define cpu_relax()			barrier()

#define task_pt_regs(task) \
	((struct pt_regs *) (task_stack_page(task) + THREAD_SIZE \
		- 8) - 1)

/*
 * Create a new kernel thread
 */
extern int kernel_thread(int (*fn)(void *), void *arg, unsigned long flags);

#define KSTK_EIP(tsk)	(((unsigned long *)(4096+(unsigned long)task_thread_info(tsk)))[1019])
#define KSTK_ESP(tsk)	(((unsigned long *)(4096+(unsigned long)task_thread_info(tsk)))[1017])


#endif

#endif /* __ASM_NDS32_PROCESSOR_H */
