/*
 *  linux/arch/nds32/include/asm/tlbflush.h
 *  Copyright (C) 2008 Andes Technology Corporation
 */

#ifndef _ASMNDS32_TLBFLUSH_H
#define _ASMNDS32_TLBFLUSH_H

#include <linux/spinlock.h>
#include <linux/mm.h>
#include <nds32_intrinsic.h>

static inline void local_flush_tlb_all(void)
{
	asm("tlbop FLUA\n");
	__nds32__isb();
}
static inline void local_flush_tlb_mm(struct mm_struct *mm)
{
	asm("tlbop FLUA\n");
	__nds32__isb();
}
static inline void local_flush_tlb_kernel_range(unsigned long start,
						unsigned long end)
{
	while(start < end) {
		asm("tlbop %0, INV"::"r" (start));
		__nds32__isb();
		start += PAGE_SIZE;
	}
}

#ifndef CONFIG_CPU_NO_CONTEXT_ID
void local_flush_tlb_range(struct vm_area_struct *vma,
			   unsigned long start, unsigned long end);
void local_flush_tlb_page(struct vm_area_struct *vma, unsigned long addr);
#else
static inline void local_flush_tlb_range(struct vm_area_struct *vma,
					 unsigned long start, unsigned long end)
{
	if ((end - start) > 0x400000) {
		asm("tlbop FLUA");
		__nds32__isb();
		return ;
	}
	while(start < end) {
		asm("tlbop %0, INV"::"r" (start));
		__nds32__isb();
		start += PAGE_SIZE;
	}
}

static inline void local_flush_tlb_page(struct vm_area_struct *vma,
					unsigned long addr)
{
	asm("tlbop %0, INV"::"r" (addr));
	__nds32__isb();
}
#endif

#ifndef CONFIG_SMP
#define flush_tlb_all		local_flush_tlb_all
#define flush_tlb_mm		local_flush_tlb_mm
#define flush_tlb_range		local_flush_tlb_range
#define flush_tlb_page		local_flush_tlb_page
#define flush_tlb_kernel_range	local_flush_tlb_kernel_range
#else
void flush_tlb_all(void);
void flush_tlb_mm(struct mm_struct *mm);
void flush_tlb_range(struct vm_area_struct *vma,
		     unsigned long start, unsigned long end);
void flush_tlb_page(struct vm_area_struct *vma, unsigned long addr);
void flush_tlb_kernel_range(unsigned long start, unsigned long end);
#endif

void update_mmu_cache(struct vm_area_struct *vma,
		      unsigned long address, pte_t* pte);
void tlb_migrate_finish(struct mm_struct *mm);

#endif
