/*
 *  linux/arch/nds32/include/asm/traps.h
 *  Copyright (C) 2008 Andes Technology Corporation
 */

#ifndef _ASMNDS32_TRAP_H
#define _ASMNDS32_TRAP_H

#include <linux/list.h>
typedef void (* NMI_handler_t)(void);

struct undef_hook {
	struct list_head node;
	u32 instr_mask;
	u32 instr_val;
	u32 cpsr_mask;
	u32 cpsr_val;
	int (*fn)(struct pt_regs *regs, unsigned int instr);
};
void register_NMI_handler(NMI_handler_t handler);
void unregister_NMI_handler(void);

void register_undef_hook(struct undef_hook *hook);
void unregister_undef_hook(struct undef_hook *hook);

extern void __init early_trap_init(void);


/* Exception entry point number */
#define ENTRY_POINT_NUM 9

/* Exception enrty point*/
#define RESET_NMI_EXCEPTION 0
#define TLB_FILL_EXCEPTION 1
#define PTE_NOT_PRESETN_EXCEPTION 2
#define TLB_MISC_EXCEPTION 3
#define TLB_VLPT_MISS_EXCEPTION 4
#define MACHINE_ERROR_EXCEPTION 5
#define DEBUG_RELATED_EXCEPTION 6
#define GENERAL_EXCEPTION 7
#define SYSCALL_EXCEPTION 8


/* The event type for general exception */
#define ALIGNMENT_CHECK 0
#define RESERVED_INSTRUCTION 1
#define TRAP 2
#define ARITHMETIC 3
#define PRECISE_BUS_ERROR 4
#define IMPRECISE_BUS_ERROR 5
#define COPROCESSOR 6
#define PRIVILEGED_INSTRUCTION 7
#define RESERVED_VALUE 8
#define NONEXISTENT_MEMORY_ADDRESS 9
#define MPPZIU_CONTROL 10
#define NEXT_PRECISE_STACK_OVERFLOW 11
#define GENERAL_EXCEPTION_RESERVED_1 12
#define GENERAL_EXCEPTION_RESERVED_2 13
#define GENERAL_EXCEPTION_RESERVED_3 14
#define GENERAL_EXCEPTION_RESERVED_4 15

/* The SUB_TYPE filed in itype register for coprocess-issuing exception */
#define INT_DIVIDE_BY_ZERO 1
#define INT_OVERFLOW 2

/* The SUB_TYPE filed in itype register for coprocess-issuing exception */
#define COPROCESS_RESERVED_1 0
#define COPROCESS_DISABLE_EXCEPTION 1
#define COPROCESS_EXCEPTION 2
#define AUDIO_EXT_DISABLE 3
#define AUDIO_EXT_ZERO_OVERHEAD_LOOP_ERR 4
#define COPROCESS_IMPRECISE_EXCEPTION 5
#define COPROCESS_RESERVED_2 6
#define ACE_DISABLE_EXCEPTION 7
#define ACE_EXCEPTION_1 8
#define ACE_EXCEPTION_2 9
#define ACE_EXCEPTION_3 10
#define ACE_EXCEPTION_4 11
#define ACE_EXCEPTION_5 12
#define ACE_EXCEPTION_6 13
#define ACE_EXCEPTION_7 14
#define ACE_EXCEPTION_8 15

/* the maximum supported coprocessor number */
#define MAX_COPROCESSOR_NUMBER 4

struct coprocsr_trap_t {
	void (*disable_handler) (unsigned long itype, struct pt_regs *regs,
				 unsigned long eva);
	void (*excpt_handler) (unsigned long itype, struct pt_regs *regs,
			       unsigned long eva);
	void (*impre_excpt_handler) (unsigned long itype, struct pt_regs *regs,
				     unsigned long eva);
};

extern struct coprocsr_trap_t coprocsr_trap[MAX_COPROCESSOR_NUMBER];


#endif
