/*
 *  linux/arch/nds32/include/asm/elf.h
 *  Copyright (C) 2008 Andes Technology Corporation
 */

#ifndef __ASMNDS32_ELF_H
#define __ASMNDS32_ELF_H

/*
 * ELF register definitions..
 */

#include <asm/ptrace.h>
#include <asm/user.h>

typedef unsigned long elf_greg_t;
typedef unsigned long elf_freg_t[3];

extern unsigned int elf_hwcap;

#define EM_NDS32		167


#define R_NDS32_NONE			0
#define R_NDS32_16_RELA			19
#define R_NDS32_32_RELA			20
#define R_NDS32_9_PCREL_RELA		22
#define R_NDS32_15_PCREL_RELA		23
#define R_NDS32_17_PCREL_RELA		24
#define R_NDS32_25_PCREL_RELA		25
#define R_NDS32_HI20_RELA		26
#define R_NDS32_LO12S3_RELA		27
#define R_NDS32_LO12S2_RELA		28
#define R_NDS32_LO12S1_RELA		29
#define R_NDS32_LO12S0_RELA		30
#define R_NDS32_SDA15S3_RELA    31
#define R_NDS32_SDA15S2_RELA    32
#define R_NDS32_SDA15S1_RELA    33
#define R_NDS32_SDA15S0_RELA    34
#define R_NDS32_GOT20			37
#define R_NDS32_25_PLTREL		38
#define R_NDS32_COPY			39
#define R_NDS32_GLOB_DAT		40
#define R_NDS32_JMP_SLOT		41
#define R_NDS32_RELATIVE		42
#define R_NDS32_GOTOFF			43
#define R_NDS32_GOTPC20			44
#define R_NDS32_GOT_HI20		45
#define R_NDS32_GOT_LO12		46
#define R_NDS32_GOTPC_HI20		47
#define R_NDS32_GOTPC_LO12		48
#define R_NDS32_GOTOFF_HI20		49
#define R_NDS32_GOTOFF_LO12		50
#define R_NDS32_INSN16			51
#define R_NDS32_LABEL			52
#define R_NDS32_LONGCALL1		53
#define R_NDS32_LONGCALL2		54
#define R_NDS32_LONGCALL3		55
#define R_NDS32_LONGJUMP1		56
#define R_NDS32_LONGJUMP2		57
#define R_NDS32_LONGJUMP3		58
#define R_NDS32_LOADSTORE		59
#define R_NDS32_9_FIXED_RELA		60
#define R_NDS32_15_FIXED_RELA		61
#define R_NDS32_17_FIXED_RELA		62
#define R_NDS32_25_FIXED_RELA		63
#define R_NDS32_PLTREL_HI20		64
#define R_NDS32_PLTREL_LO12		65
#define R_NDS32_PLT_GOTREL_HI20		66
#define R_NDS32_PLT_GOTREL_LO12		67
#define R_NDS32_LO12S0_ORI_RELA		72
#define R_NDS32_DWARF2_OP1_RELA     77
#define R_NDS32_DWARF2_OP2_RELA     78
#define R_NDS32_DWARF2_LEB_RELA     79
#define R_NDS32_WORD_9_PCREL_RELA	94
#define R_NDS32_LONGCALL4 			107
#define R_NDS32_RELA_NOP_MIX		192
#define R_NDS32_RELA_NOP_MAX		255

#define ELF_NGREG (sizeof (struct pt_regs) / sizeof(elf_greg_t))
typedef elf_greg_t elf_gregset_t[ELF_NGREG];

typedef struct user_fp elf_fpregset_t;

struct elf32_hdr;
extern int elf_check_arch(const struct elf32_hdr *hdr);

/*
 * These are used to set parameters in the core dumps.
 */
#define ELF_CLASS	ELFCLASS32
#ifdef __NDS32_EB__
#define ELF_DATA	ELFDATA2MSB;
#else
#define ELF_DATA	ELFDATA2LSB;
#endif
#define ELF_ARCH	EM_NDS32
#define USE_ELF_CORE_DUMP
#define ELF_EXEC_PAGESIZE	PAGE_SIZE

/* This is the location that an ET_DYN program is loaded if exec'ed.  Typical
   use of this is to invoke "./ld.so someprog" to test out a new version of
   the loader.  We need to make sure that it is out of the way of the program
   that it will "exec", and that there is sufficient room for the brk.  */

#define ELF_ET_DYN_BASE	(2 * TASK_SIZE / 3)

/* When the program starts, a1 contains a pointer to a function to be
   registered with atexit, as per the SVR4 ABI.  A value of 0 means we
   have no such handler.  */
#define ELF_PLAT_INIT(_r, load_addr)	(_r)->NDS32_r0 = 0

/* This yields a mask that user programs can use to figure out what
   instruction set this cpu supports. */

#define ELF_HWCAP	(elf_hwcap)

/* This yields a string that ld.so will use to load implementation
   specific libraries for optimization.  This is more specific in
   intent than poking at uname or /proc/cpuinfo. */

/* For now we just provide a fairly general string that describes the
   processor family.  This could be made more specific later if someone
   implemented optimisations that require it.  26-bit CPUs give you
   "v1l" for ARM2 (no SWP) and "v2l" for anything else (ARM1 isn't
   supported).  32-bit CPUs give you "v3[lb]" for anything based on an
   ARM6 or ARM7 core and "armv4[lb]" for anything based on a StrongARM-1
   core.  */

#define ELF_PLATFORM_SIZE 16
extern char elf_platform[];
#define ELF_PLATFORM	(elf_platform)

#ifdef __KERNEL__

/* Old NetWinder binaries were compiled in such a way that the iBCS
   heuristic always trips on them.  Until these binaries become uncommon
   enough not to care, don't trust the `ibcs' flag here.  In any case
   there is no other ELF system currently supported by iBCS.
   @@ Could print a warning message to encourage users to upgrade.  */
#define SET_PERSONALITY(ex)	set_personality(PER_LINUX)

#endif

#endif
