/*
 *  linux/arch/nds32/include/asm/ptrace.h
 *
 *  Copyright (C) 1996-2003 Russell King
 *  Copyright (C) 2008 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#ifndef __ASM_NDS32_PTRACE_H
#define __ASM_NDS32_PTRACE_H

#define PTRACE_GETREGS		12
#define PTRACE_SETREGS		13
#define PTRACE_GETFPREGS	14
#define PTRACE_SETFPREGS	15
#define PTRACE_GETAUREGS	18
#define PTRACE_SETAUREGS	19

#define PTRACE_OLDSETOPTIONS	21

#define PTRACE_GET_THREAD_AREA	22

#ifndef __ASSEMBLY__
/* this struct defines the way the registers are stored on the
   stack during a system call. */

struct pt_regs {
	long uregs[44];
};
#define NDS32_osp       uregs[43]
#define NDS32_FUCOP_CTL uregs[42]
#define NDS32_lp        uregs[41]
#define NDS32_gp        uregs[40]
#define NDS32_fp        uregs[39]
#define NDS32_r25       uregs[38]
#define NDS32_r24       uregs[37]
#define NDS32_r23       uregs[36]
#define NDS32_r22       uregs[35]
#define NDS32_r21       uregs[34]
#define NDS32_r20       uregs[33]
#define NDS32_r19       uregs[32]
#define NDS32_r18       uregs[31]
#define NDS32_r17       uregs[30]
#define NDS32_r16       uregs[29]
#define NDS32_r15       uregs[28]
#define NDS32_r14       uregs[27]
#define NDS32_r13       uregs[26]
#define NDS32_r12       uregs[25]
#define NDS32_r11       uregs[24]
#define NDS32_r10       uregs[23]
#define NDS32_r9        uregs[22]
#define NDS32_r8        uregs[21]
#define NDS32_r7        uregs[20]
#define NDS32_r6        uregs[19]
#define NDS32_r5        uregs[18]
#define NDS32_r4        uregs[17]
#define NDS32_r3        uregs[16]
#define NDS32_r2        uregs[15]
#define NDS32_r1        uregs[14]
#define NDS32_r0        uregs[13]
#if defined(CONFIG_HWZOL)
#define NDS32_lc        uregs[11]
#define NDS32_le        uregs[10]
#define NDS32_lb        uregs[9]
#endif
#define NDS32_pp1       uregs[8]
#define NDS32_pp0       uregs[7]
#define NDS32_pipc      uregs[6]
#define NDS32_pipsw     uregs[5]
#define NDS32_ORIG_r0   uregs[4]
#define NDS32_sp        uregs[3]
#define NDS32_ipc       uregs[2]
#define NDS32_ipsw      uregs[1]
#define NDS32_ir0       uregs[0]

#ifdef __KERNEL__
#include <asm/bitfield.h>

#define arch_has_single_step()	(1)
struct task_struct;
extern void user_enable_single_step(struct task_struct *);
extern void user_disable_single_step(struct task_struct *);
extern void user_direct_debug_exception_to_IVB(struct task_struct *);
extern void user_direct_debug_exception_to_DIMU(struct task_struct *);
extern void arch_ptrace_attach(struct task_struct *);
#define arch_ptrace_attach(child) arch_ptrace_attach(child)

#define user_mode(regs)	(((regs)->NDS32_ipsw & PSW_mskPOM) == 0)

#define interrupts_enabled(regs)         (!((regs)->NDS32_ipsw & ~PSW_mskGIE))

extern void show_regs(struct pt_regs *);

/* Are the current registers suitable for user mode?
 * (used to maintain security in signal handlers)
 */
static inline int valid_user_regs(struct pt_regs *regs)
{
	return user_mode(regs) && ((regs->NDS32_ipsw & PSW_mskGIE) == 1);
}

static inline unsigned long regs_return_value(struct pt_regs *regs)
{
        return regs->NDS32_r0;
}


#define instruction_pointer(regs)	((regs)->NDS32_ipc)
#define user_stack_pointer(regs)        ((regs)->NDS32_sp)

#ifdef CONFIG_SMP
extern unsigned long profile_pc(struct pt_regs *regs);
#else
#define profile_pc(regs) instruction_pointer(regs)
#endif

/*
 * kprobe-based event tracer support
 */
#include <linux/stddef.h>
#include <linux/types.h>
#define MAX_REG_OFFSET (offsetof(struct pt_regs, NDS32_osp))

extern int regs_query_register_offset(const char *name);
extern const char *regs_query_register_name(unsigned int offset);
extern bool regs_within_kernel_stack(struct pt_regs *regs, unsigned long addr);
extern unsigned long regs_get_kernel_stack_nth(struct pt_regs *regs,
                                               unsigned int n);

/**
 * regs_get_register() - get register value from its offset
 * @regs:      pt_regs from which register value is gotten
 * @offset:    offset number of the register.
 *
 * regs_get_register returns the value of a register whose offset from @regs.
 * The @offset is the offset of the register in struct pt_regs.
 * If @offset is bigger than MAX_REG_OFFSET, this returns 0.
 */
static inline unsigned long regs_get_register(struct pt_regs *regs,
                                              unsigned int offset)
{
        if (unlikely(offset > MAX_REG_OFFSET))
                return 0;
        return *(unsigned long *)((unsigned long)regs + offset);
}

/* Valid only for Kernel mode traps. */
static inline unsigned long kernel_stack_pointer(struct pt_regs *regs)
{
        return regs->NDS32_sp;
}

#endif /* __KERNEL__ */

#endif /* !__ASSEMBLY__ */

#endif /* __ASM_NDS32_PTRACE_H */
