/*
 * Copyright (C) 1995-2003 Russell King
 *               2001-2002 Keith Owens
 * Copyright (C) 2009 Andes Technology Corporation
 *
 * Generate definitions needed by assembly language modules.
 * This code generates raw asm output which is post-processed to extract
 * and format the required data.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/sched.h>
#include <linux/mm.h>
#include <asm/thread_info.h>
#include <asm/memory.h>
#include <asm/procinfo.h>

/*
 * Make sure that the compiler and target are compatible.
 */

#if __GNUC__ < 3 || \
   (__GNUC__ == 3 && __GNUC_MINOR__ < 4) || \
   (__GNUC__ == 3 && __GNUC_MINOR__ == 4 && __GNUC_PATCHLEVEL__ != 0 && \
					     __GNUC_PATCHLEVEL__ < 4) || \
   (__GNUC__ == 4 && __GNUC_MINOR__ < 2)
#error Your compiler is too buggy; it is known to miscompile kernels.
#error    Known good compilers: 3.4.4, 4.2
#endif

/* Use marker if you need to separate the values later */

#define DEFINE(sym, val) \
        asm volatile("\n->" #sym " %0 " #val : : "i" (val))

#define BLANK() asm volatile("\n->" : : )

int main(void)
{
	DEFINE(TSK_ACTIVE_MM, offsetof(struct task_struct, active_mm));
	BLANK();
	DEFINE(TI_FLAGS, offsetof(struct thread_info, flags));
	DEFINE(TI_PREEMPT, offsetof(struct thread_info, preempt_count));
	DEFINE(TI_ADDR_LIMIT, offsetof(struct thread_info, addr_limit));
	DEFINE(TI_TASK, offsetof(struct thread_info, task));
	DEFINE(TI_EXEC_DOMAIN, offsetof(struct thread_info, exec_domain));
	DEFINE(TI_CPU, offsetof(struct thread_info, cpu));
	DEFINE(TI_SP_SAVE, offsetof(struct thread_info, sp_save));
	BLANK();
	DEFINE(S_FRAME_SIZE, sizeof(struct pt_regs));
	BLANK();
	DEFINE(MM_CONTEXT_ID, offsetof(struct mm_struct, context.id));
	BLANK();
	DEFINE(VMA_VM_MM, offsetof(struct vm_area_struct, vm_mm));
	DEFINE(VMA_VM_FLAGS, offsetof(struct vm_area_struct, vm_flags));
	BLANK();
	DEFINE(VM_EXEC, VM_EXEC);
	BLANK();
	DEFINE(VIRT_OFFSET, PAGE_OFFSET);
	BLANK();
	DEFINE(PROC_INFO_SZ, sizeof(struct proc_info_list));

	return 0;
}
