#include <linux/linkage.h>
#include <asm/memory.h>
#include <asm/nds32.h>
#include <asm/assembler.h>
#include <asm/errno.h>
#include <asm/asm-offsets.h>
#include <asm/page.h>

	.macro push_zol
#ifdef CONFIG_HWZOL
	mfusr	$r10, $LB
	mfusr	$r11, $LE
	mfusr	$r12, $LC
#endif
	.endm
	.macro	save_user_regs
	/* $p1 is VA of percpu data */
	sethi	$p0, hi20(PAGE_OFFSET - PHYS_OFFSET)
	add	$p1, $p1, $p0

	/* change to kernel stack */
	mfsr	$p0, $IPSW
	andi	$p0, $p0, #PSW_mskPOM
	cmovz	$sp, $r25, $p0

	/* 8 byte aligned */
	movi	$r25, #~7
	and	$r25, $sp, $r25
	smw.adm	$sp, [$r25], $sp, #0x1
	move	$sp, $r25

#if defined(CONFIG_FPU)
        mfsr    $r25, $FUCOP_CTL
        push    $r25
        bclr    $r25, $r25, #FUCOP_CTL_offCP0EN
	mtsr    $r25, $FUCOP_CTL
#else
        addi    $sp, $sp, -4
#endif
	pushm	$r28, $r30

	/* zero fp if user mode*/
	movi	$r25, #0x0
	cmovz	$fp, $r25, $p0

	lwi	$r25, [$p1+#0x0]
	pushm	$r0, $r25
#ifdef CONFIG_HWZOL
	push_zol
#endif
	mfsr	$r9, $P_P1
	mfsr	$r8, $P_P0
	mfsr	$r7, $P_IPC
	mfsr	$r6, $P_IPSW
	mfsr	$r18, $IPC
	mfsr	$r17, $IPSW
	mfsr	$r16, $PSW
	pushm	$r6, $r13
	push	$r0
	lwi	$p0, [$p1+#0x8]
	push	$p0
	pushm	$r16, $r18

	andi	$r19, $r16, #PSW_mskINTL
	slti	$r20, $r19, #4
	bnez	$r20, 1f
	addi	$r21, $r16, #-2
	mtsr	$r21, $PSW
	isb
1:
	.endm

	.text

/*
 * Exception Vector
 */
exception_handlers:
	.long	do_NMI			!Reset/NMI
	.long	unhandled_exceptions	!TLB fill
	.long	do_page_fault		!PTE not present
	.long	do_dispatch_tlb_misc	!TLB misc
	.long	unhandled_exceptions	!TLB VLPT
	.long	unhandled_exceptions	!Machine Error
	.long	do_debug_trap		!Debug related
	.long	do_dispatch_general	!General exception
	.long	eh_syscall		!Syscall
	.long	asm_do_IRQ		!IRQ

common_exception_handler:
	save_user_regs
	lwi	$p0, [$p1+#0x4]
	andi	$p1, $p0, #0x78
	bnez	$p1, 1f
	sethi	$lp, hi20(ret_from_exception)
	ori	$lp, $lp, lo12(ret_from_exception)
	sethi	$p1, hi20(exception_handlers)
	ori	$p1, $p1, lo12(exception_handlers)
	lw	$p1, [$p1+$p0<<2]
	move	$r0, $p0
	mfsr	$r1, $EVA
	mfsr	$r2, $ITYPE
	move	$r3, $sp
	mfsr    $r4, $OIPC
	/* enable gie if it is enabled in IPSW. */
	mfsr	$r21, $PSW
	andi	$r17, $r17, #0x1
	or	$r21, $r21, $r17
	mtsr	$r21, $PSW
	dsb
	jr	$p1

	/* syscall */
1:
	addi	$p1, $p0, #-8
	bnez	$p1, 2f
	sethi	$lp, hi20(ret_from_exception)
	ori	$lp, $lp, lo12(ret_from_exception)
	sethi	$p1, hi20(exception_handlers)
	ori	$p1, $p1, lo12(exception_handlers)
	lwi	$p1, [$p1+#0x8<<2]
	jr	$p1

	/* interrupt */
2:
#ifdef CONFIG_TRACE_IRQFLAGS
	jal     arch_trace_hardirqs_off
#endif
#if defined(CONFIG_EVIC) || defined(CONFIG_IVIC)
	addi	$r0, $p0, #-9
#else
# ifdef CONFIG_IVIC_INTC
	jal	get_IntSrc
# else
# error "Not configure Vector Interrupt Controller  mode"
# endif
#endif
	move	$r1, $sp
	sethi	$lp, hi20(ret_from_intr)
	ori	$lp, $lp, lo12(ret_from_intr)
	sethi	$p0, hi20(exception_handlers)
	ori	$p0, $p0, lo12(exception_handlers)
	lwi	$p0, [$p0+#0x9<<2]
	jr	$p0

	.macro	EXCEPTION_VECTOR, num
	.align 6
	.ifc \num, 6
	mfsr    $p0, $EDM_CTL
	andi    $p0, $p0, EDM_CTL_mskV3_EDM_MODE
	tnez    $p0, 0x1a
	.endif
	/* $p1 is PA of percpu data */
	mfsr	$p1, $CORE_ID
	slli	$p1, $p1, #0x9
	li	$p0, #(PHYS_OFFSET+0x6000)
	or	$p1, $p1, $p0

	/*
	 * 0x0 saved $r25
	 * 0x4 vector number
	 * 0x8 user stack pointer
	 * 0xc kernel stack pointer
	 */
	swi	$r25, [$p1+0x0]
	movi	$p0, \num
	swi	$p0, [$p1+#0x4]
	swi	$sp, [$p1+#0x8]
	lwi	$r25, [$p1+#0xc]
	sethi	$p0, hi20(common_exception_handler)
	ori	$p0, $p0, lo12(common_exception_handler)
	jral.ton	$p0, $p0
	.endm

	.macro	IPI_VECTOR, num
	.align 6
	mfsr	$p1, $CORE_ID
	slli	$p1, $p1, #0x9
	li	$p0, #(PHYS_OFFSET+0x6000)
	or	$p1, $p1, $p0

	/*
	 * 0x10 indicator of CPU is initialized
	 * 0x14 CPU init function
	 */
	swi	$r25, [$p1+0x0]
	movi	$p0, \num
	swi	$p0, [$p1+#0x4]
	swi	$sp, [$p1+#0x8]
	lwi	$r25, [$p1+#0xc]
	lwi	$p0, [$p1+#0x10]
	beqz	$p0, 1f
	sethi	$p0, hi20(common_exception_handler)
	ori	$p0, $p0, lo12(common_exception_handler)
	jral.ton	$p0, $p0
1:
	lwi	$p0, [$p1+#0x14]
	jr	$p0

	.endm

	.section	".text.init", #alloc, #execinstr
	.global	exception_vector
exception_vector:
	EXCEPTION_VECTOR 0
	EXCEPTION_VECTOR 1
	EXCEPTION_VECTOR 2
	EXCEPTION_VECTOR 3
	EXCEPTION_VECTOR 4
	EXCEPTION_VECTOR 5
	EXCEPTION_VECTOR 6
	EXCEPTION_VECTOR 7
	EXCEPTION_VECTOR 8
#ifdef CONFIG_EVIC
	EXCEPTION_VECTOR 9
#else
	IPI_VECTOR 9
#endif
	EXCEPTION_VECTOR 10
	EXCEPTION_VECTOR 11
	EXCEPTION_VECTOR 12
	EXCEPTION_VECTOR 13
	EXCEPTION_VECTOR 14
	EXCEPTION_VECTOR 15
	EXCEPTION_VECTOR 16
	EXCEPTION_VECTOR 17
	EXCEPTION_VECTOR 18
	EXCEPTION_VECTOR 19
	EXCEPTION_VECTOR 20
	EXCEPTION_VECTOR 21
	EXCEPTION_VECTOR 22
	EXCEPTION_VECTOR 23
	EXCEPTION_VECTOR 24
	EXCEPTION_VECTOR 25
	EXCEPTION_VECTOR 26
	EXCEPTION_VECTOR 27
	EXCEPTION_VECTOR 28
	EXCEPTION_VECTOR 29
	EXCEPTION_VECTOR 30
	EXCEPTION_VECTOR 31
	EXCEPTION_VECTOR 32
	EXCEPTION_VECTOR 33
	EXCEPTION_VECTOR 34
	EXCEPTION_VECTOR 35
	EXCEPTION_VECTOR 36
	EXCEPTION_VECTOR 37
	EXCEPTION_VECTOR 38
	EXCEPTION_VECTOR 39
	EXCEPTION_VECTOR 40
#ifdef CONFIG_EVIC
	IPI_VECTOR 41
#else
	EXCEPTION_VECTOR 41
#endif
	EXCEPTION_VECTOR 42
	EXCEPTION_VECTOR 43
	EXCEPTION_VECTOR 44
	EXCEPTION_VECTOR 45
	EXCEPTION_VECTOR 46
	EXCEPTION_VECTOR 47
	EXCEPTION_VECTOR 48
	EXCEPTION_VECTOR 49
	EXCEPTION_VECTOR 50
	EXCEPTION_VECTOR 51
	EXCEPTION_VECTOR 52
	EXCEPTION_VECTOR 53
	EXCEPTION_VECTOR 54
	EXCEPTION_VECTOR 55
	EXCEPTION_VECTOR 56
	EXCEPTION_VECTOR 57
	EXCEPTION_VECTOR 58
	EXCEPTION_VECTOR 59
	EXCEPTION_VECTOR 60
	EXCEPTION_VECTOR 61
	EXCEPTION_VECTOR 62
	EXCEPTION_VECTOR 63
	EXCEPTION_VECTOR 64
	EXCEPTION_VECTOR 65
	EXCEPTION_VECTOR 66
	EXCEPTION_VECTOR 67
	EXCEPTION_VECTOR 68
	EXCEPTION_VECTOR 69
	EXCEPTION_VECTOR 70
	EXCEPTION_VECTOR 71
	EXCEPTION_VECTOR 72
