#include <linux/linkage.h>
#include <asm/unistd.h>
#include <asm/assembler.h>
#include <asm/nds32.h>
#include <asm/asm-offsets.h>
#include <asm/thread_info.h>

#define why	$r8            //  Linux syscall (!= 0)
#define tsk	$r9            //  current thread_info


	.macro pop_zol
#ifdef CONFIG_HWZOL
	mtusr	$r10, $LB
	mtusr	$r11, $LE
	mtusr	$r12, $LC
#endif
	.endm

	.macro	restore_user_regs_first
	setgie.d
	isb
	/* $p1 is VA of percpu data */
	mfsr	$p1, $CORE_ID
	andi	$p1, $p1, #0x3
	slli	$p1, $p1, #0x9
	sethi	$p0, #0xc0006
	or	$p1, $p1, $p0

	popm	$r16, $r18
	mtsr	$r16, $PSW
	pop	$r16
	swi	$r16, [$p1+#0x8]
	addi	$sp, $sp, #4	!pop $r0
	popm	$r6, $r13
	mtsr	$r17, $IPSW
	mtsr	$r18, $IPC
	mtsr	$r6, $P_IPSW
	mtsr	$r7, $P_IPC
	mtsr	$r8, $P_P0
	mtsr	$r9, $P_P1
#ifdef CONFIG_HWZOL
	pop_zol
#endif
	.endm

	.macro	restore_user_regs_last
	popm	$r28, $r30
#if defined(CONFIG_FPU)
        pop     $p0
        mtsr    $p0, $FUCOP_CTL
#else
	addi    $sp, $sp, 4
#endif

	pop	$p0
	cmovn	$sp, $p0, $p0

	mfsr	$p0, $IPSW
	andi	$p0, $p0, #PSW_mskPOM
	bnez	$p0, 1f
	swi	$sp, [$p1+#0xc]
	lwi	$sp, [$p1+#0x8]
1:
	// This is SW workaround for Bug #6294
	li	$p0, 0xc0000000
	cctl	$p0, L1D_VA_INVAL
	lwi	$p1, [$p0]

	iret
	nop

	.endm

	.macro	restore_user_regs
	restore_user_regs_first
	popm	$r0, $r25
	restore_user_regs_last
	.endm

	.macro	fast_restore_user_regs
	restore_user_regs_first
	addi	$sp, $sp, #4
	popm	$r1, $r25
	restore_user_regs_last
	.endm

#ifdef CONFIG_PREEMPT
	.macro	preempt_stop
	.endm
#else
	.macro	preempt_stop
	setgie.d
	isb
	.endm
#define	resume_kernel	no_work_pending
#endif

ENTRY(ret_from_exception)
	preempt_stop
ENTRY(ret_from_intr)
	get_thread_info tsk
	move	why, #0				! not system call

/*
 * judge Kernel or user mode
 *
 */
	lwi	$p0, [$sp+(#S_IPSW)]		! Check if in nested interrupt
	andi	$p0, $p0, #PSW_mskINTL
	bnez	$p0, resume_kernel		! done with iret
	j	resume_userspace


/*
 * This is the fast syscall return path.  We do as little as
 * possible here, and this includes saving $r0 back into the SVC
 * stack.
 * fixed: tsk - $r9, why - $r8, $r7 - syscall #, $r8 - syscall table pointer
 */
ENTRY(ret_fast_syscall)
	gie_disable
	lwi	$r1, [tsk+#TI_FLAGS]
	andi	$p1, $r1, #_TIF_WORK_MASK
	bnez	$p1, fast_work_pending
	fast_restore_user_regs			! iret

/*
 * Ok, we need to do extra processing,
 * enter the slow path returning from syscall, while pending work.
 */
fast_work_pending:
	swi	$r0, [$sp+(#S_R0)]		! what is different from ret_from_exception
	move	why, #1				! come from a syscall
work_pending:
	andi	$p1, $r1, #_TIF_NEED_RESCHED
	bnez	$p1, work_resched

	andi	$p1, $r1, #_TIF_SIGPENDING|#_TIF_NOTIFY_RESUME
	beqz	$p1, no_work_pending

	move	$r0, $sp			! 'regs'
	move	$r2, why
	gie_enable
	bal	do_notify_resume

	b	ret_slow_syscall		! return from slow_restore_user_regs

work_resched:
	bal	schedule			! path, return to user mode

/*
 * "slow" syscall return path.
 * "why" tells us if this was a real syscall.
 */
ENTRY(resume_userspace)
ENTRY(ret_slow_syscall)
	gie_disable
	lwi	$p0, [$sp+(#S_IPSW)]		! Check if in nested interrupt
	andi	$p0, $p0, #PSW_mskINTL
	bnez	$p0, no_work_pending		! done with iret
	lwi	$r1, [tsk+#TI_FLAGS]
	andi	$p1, $r1, #_TIF_WORK_MASK
	bnez	$p1, work_pending		! handle work_resched, sig_pend

no_work_pending:
#ifdef CONFIG_TRACE_IRQFLAGS
	lwi	$p0, [$sp+(#S_IPSW)]
	andi	$p0, $p0, #0x1
	la	$r10, arch_trace_hardirqs_off
	la	$r9, arch_trace_hardirqs_on
	cmovz	$r9, $p0, $r10
	jral	$r9
#endif
	restore_user_regs			! return from iret


/*
 * preemptive kernel
 */
#ifdef CONFIG_PREEMPT
resume_kernel:
	gie_disable
	lwi	$t0, [tsk+#TI_PREEMPT]
	bnez	$t0, no_work_pending
need_resched:
	lwi	$t0, [tsk+#TI_FLAGS]
	andi	$p1, $t0, #_TIF_NEED_RESCHED
	beqz	$p1, no_work_pending

	lwi	$t0, [$sp+(#S_IPSW)]		! Interrupts off?
	andi	$t0, $t0, #1
	beqz	$t0, no_work_pending

	jal	preempt_schedule_irq
	b	need_resched
#endif

/*
 * This is how we return from a fork.
 */
ENTRY(ret_from_fork)
	bal	schedule_tail
	beqz	$r6, 1f				! r6 stores fn for kernel thread
	move	$r0, $r7			! prepare kernel thread arg
	jral	$r6
1:
	get_thread_info tsk
	lwi	$r1, [tsk+#TI_FLAGS]		! check for syscall tracing
	move	why, #1
	andi	$p1, $r1, #_TIF_WORK_SYSCALL_LEAVE	! are we tracing syscalls?
	beqz	$p1, ret_slow_syscall
	move    $r0, $sp
	bal	syscall_trace_leave
	b	ret_slow_syscall

