#include <linux/linkage.h>
#include <asm/unistd.h>
#include <asm/assembler.h>
#include <asm/nds32.h>
#include <asm/asm-offsets.h>
#include <asm/thread_info.h>

/*
 * $r0 = previous task_struct,
 * $r1 = previous thread_info,
 * $r2 = next thread_info
 * previous and next are guaranteed not to be the same.
 */

ENTRY(_switch)
	smw.adm $r6, [$sp], $r14, #0xa
	swi	$sp, [$r1+ TI_SP_SAVE]	/* Save $sp to task structure */
	lwi	$sp, [$r2+ TI_SP_SAVE]	/* Get $sp from task structure */
	lmw.bim $r6, [$sp], $r14, #0xa
	ret

/*
 * These are the registers used in the syscall handler, and allow us
 * to have in theory up to 6 arguments to a function - $r0 to $r5.
 * $r7 is reserved for the system call number for Andes architecture.
 *
 * Note that tbl == why is intentional.
 * We must set at least "tsk" and "why" when calling ret_with_reschedule.
 */
#define tbl	$r8            //  syscall table pointer
#define why	$r8            //  Linux syscall (!= 0)
#define tsk	$r9            //  current thread_info

/*
 * Get the system call number. let $r7 take a system call nr.
 *
 * if it is called from library function, $r7 is filled by library call
 * when user program make a system call. and swid bitfield of $ir6 will
 * always be encoded as 0x7fff.
 *
 * if it is called from kernel code, $r7 will be writen as syscall nr
 * by retrieving from $ir6 'swid' bitfiled
 */
	.macro	get_scno
	mfsr	$r7, $ITYPE
	srli	$r7, $r7, #ITYPE_offSWID
	.endm

	.macro	updateipc
	addi	$r17, $r18, #4
	swi	$r17, [$sp + S_IPC]
	.endm

	.equ NR_syscalls,0
#define CALL(x) .equ NR_syscalls,NR_syscalls+1
#include "calls.S"
#undef CALL
#define CALL(x) .long x

ENTRY(eh_syscall)
	updateipc

	get_scno
	gie_enable
	get_thread_info tsk
	lwi	$p0, [tsk+#TI_FLAGS]		! check for syscall tracing

	andi	$p1, $p0, #_TIF_WORK_SYSCALL_ENTRY	! are we tracing syscalls?
	bnez	$p1, __sys_trace

	la	$lp, ret_fast_syscall		! return address

	li	$p0, __NR_NDS32_BASE
	addi	$p0, $p0, -1
	slt	$p0, $p0, $r7
	beqz	$p0, 1f
	addi    $r1, $SP, #0
	move	$r0, $r7
	b	nds32_syscall
1:
						! check upper syscall limit,
	andi	$r7, $r7, #0xfff
	addi	$p1, $r7, #-NR_syscalls		! syscall number of syscall instruction is guarded by addembler
	bgez	$p1, _SCNO_EXCEED		! call sys_* routine
	la	tbl, sys_call_table		! load syscall table pointer
	slli	$p1, $r7, #2
	add	$p1, tbl, $p1
	lwi	$p1, [$p1]
	jr	$p1				! no return

_SCNO_EXCEED:
	ori	$r0, $r7, #0
        ori	$r1, $sp, #0
	b	bad_syscall

/*
 * This is the really slow path.  We're going to be doing
 * context switches, and waiting for our parent to respond.
 */
__sys_trace:
	move	$r1, $sp
	move	$r0, $r7			! trace entry [IP = 0]
	bal	syscall_trace_enter
	move	$r7, $r0
	la	$lp, __sys_trace_return		! return address

	li	$p0, __NR_NDS32_BASE
	addi	$p0, $p0, -1
	slt	$p0, $p0, $r7
	beqz	$p0, 1f
	addi    $r1, $SP, #0
	b	nds32_syscall
1:
	andi	$r7, $r7, #0xfff
	addi	$p1, $sp, #S_R0			! pointer to regs
	addi	$p0, $r7, -#NR_syscalls		! check upper syscall limit
	bgez	$p0, _SCNO_EXCEED
	lmw.bi	$r0, [$p1], $r5			! have to reload $r0 - $r5

	slli	$p0, $r7, #2			! call sys_* routine
	la	tbl, sys_call_table		! load syscall table pointer
	add	$p0, tbl, $p0
	lwi	$p0, [$p0]
	jr	$p0

__sys_trace_return:
	swi	$r0, [$sp+#S_R0]		! T: save returned $r0
	move	$r0, $sp			! set pt_regs for syscall_trace_leave
	bal	syscall_trace_leave
	b	ret_slow_syscall

    .type   sys_call_table, #object
ENTRY(sys_call_table)
#include "calls.S"

/*
 * Special system call wrappers
 *
 * $r0 = syscall number
 * $r8 = syscall table
 */
       .type   sys_syscall, #function
sys_syscall:

	li	$p0, __NR_NDS32_BASE
	addi	$p0, $p0, -1
	slt	$p0, $p0, $r0
	beqz	$p0, 1f
	addi    $r1, $SP, #0
	b	nds32_syscall
1:
	bltz	$r0, 3f			! Guard whether syscall number is between 0~0x7fff.
	andi	$r0, $r0, #0xfff
	addi	$p1, $r0, #-NR_syscalls
	bgtz	$p1, 3f
	move	$p1, $r0
	move	$r0, $r1
	move	$r1, $r2
	move	$r2, $r3
	move	$r3, $r4
	move	$r4, $r5
! add for syscall 6 args
	lwi	$r5, [$sp + #S_SP ]
	lwi	$r5, [$r5]
! ~add for syscall 6 args

	lw	$p1, [tbl+$p1<<2]
	jr	$p1
3:	b	sys_ni_syscall

sys_sigreturn_wrapper:
	addi	$r0, $sp, #0
	b	sys_sigreturn

sys_rt_sigreturn_wrapper:
	addi	$r0, $sp, #0
	b	sys_rt_sigreturn

