/*
 * arch/nds32/kernel/fpu.c
 *
 * Copyright (C) 2001  Manuela Cirronis, Paolo Alberelli
 * Copyright (C) 2002  STMicroelectronics Limited
 *   Author : Stuart Menefy
 * Copyright (C) 2009 Andes Technology Corporation
 *
 * Started from SH4 version:
 *   Copyright (C) 1999, 2000  Kaz Kojima & Niibe Yutaka
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 */
#include <linux/sched.h>
#include <linux/signal.h>
#include <asm/processor.h>
#include <asm/user.h>
#include <asm/io.h>
#include <asm/bitfield.h>
#include <asm/fpu.h>
#include <asm/traps.h>
#include "fpu.h"

extern void do_revinsn(struct pt_regs *regs);
extern int do_fpu_denorm(struct pt_regs *regs, struct fpu_struct *fpu);

const struct fpu_struct init_fpuregs = {
	.fd_regs = {[0 ... 31] = sNAN64},
	.fpcsr = FPCSR_INIT
};

void save_fpu(struct task_struct *tsk)
{
	unsigned int fpcfg, fpcsr;
	enable_fpu();
	asm volatile ("fmfcfg	%0\n\t":"=&r" (fpcfg));
	fpcfg = ((fpcfg & FPCFG_mskFREG) >> FPCFG_offFREG);

	switch (fpcfg) {
	case 3:
		asm volatile ("fsdi $fd31, [%0+0xf8]\n\t"
			      "fsdi $fd30, [%0+0xf0]\n\t"
			      "fsdi $fd29, [%0+0xe8]\n\t"
			      "fsdi $fd28, [%0+0xe0]\n\t"
			      "fsdi $fd27, [%0+0xd8]\n\t"
			      "fsdi $fd26, [%0+0xd0]\n\t"
			      "fsdi $fd25, [%0+0xc8]\n\t"
			      "fsdi $fd24, [%0+0xc0]\n\t"
			      "fsdi $fd23, [%0+0xb8]\n\t"
			      "fsdi $fd22, [%0+0xb0]\n\t"
			      "fsdi $fd21, [%0+0xa8]\n\t"
			      "fsdi $fd20, [%0+0xa0]\n\t"
			      "fsdi $fd19, [%0+0x98]\n\t"
			      "fsdi $fd18, [%0+0x90]\n\t"
			      "fsdi $fd17, [%0+0x88]\n\t"
			      "fsdi $fd16, [%0+0x80]\n\t"
			      :	/* no output */
			      :"r" (&tsk->thread.fpu)
			      :"memory");
		/* fall through */
	case 2:
		asm volatile ("fsdi $fd15, [%0+0x78]\n\t"
			      "fsdi $fd14, [%0+0x70]\n\t"
			      "fsdi $fd13, [%0+0x68]\n\t"
			      "fsdi $fd12, [%0+0x60]\n\t"
			      "fsdi $fd11, [%0+0x58]\n\t"
			      "fsdi $fd10, [%0+0x50]\n\t"
			      "fsdi $fd9,  [%0+0x48]\n\t"
			      "fsdi $fd8,  [%0+0x40]\n\t"
			      :	/* no output */
			      :"r" (&tsk->thread.fpu)
			      :"memory");
		/* fall through */
	case 1:
		asm volatile ("fsdi $fd7,  [%0+0x38]\n\t"
			      "fsdi $fd6,  [%0+0x30]\n\t"
			      "fsdi $fd5,  [%0+0x28]\n\t"
			      "fsdi $fd4,  [%0+0x20]\n\t"
			      :	/* no output */
			      :"r" (&tsk->thread.fpu)
			      :"memory");
		/* fall through */
	case 0:
		asm volatile ("fsdi $fd3,  [%1+0x18]\n\t"
			      "fsdi $fd2,  [%1+0x10]\n\t"
			      "fsdi $fd1,  [%1+0x8]\n\t"
			      "fsdi $fd0,  [%1+0x0]\n\t"
			      "fmfcsr	%0\n\t"
			      "swi  %0, [%1+0x100]\n\t"
                              :"=&r" (fpcsr)
			      :"r"(&tsk->thread.fpu)
			      :"memory");
	}
	disable_fpu();
}

void fpload(const struct fpu_struct *fpregs)
{
	unsigned int fpcfg, fpcsr;
	enable_fpu();

	asm volatile ("fmfcfg    %0\n\t":"=&r" (fpcfg));
	fpcfg = ((fpcfg & FPCFG_mskFREG) >> FPCFG_offFREG);

	switch (fpcfg) {
	case 3:
		asm volatile ("fldi $fd31, [%0+0xf8]\n\t"
			      "fldi $fd30, [%0+0xf0]\n\t"
			      "fldi $fd29, [%0+0xe8]\n\t"
			      "fldi $fd28, [%0+0xe0]\n\t"
			      "fldi $fd27, [%0+0xd8]\n\t"
			      "fldi $fd26, [%0+0xd0]\n\t"
			      "fldi $fd25, [%0+0xc8]\n\t"
			      "fldi $fd24, [%0+0xc0]\n\t"
			      "fldi $fd23, [%0+0xb8]\n\t"
			      "fldi $fd22, [%0+0xb0]\n\t"
			      "fldi $fd21, [%0+0xa8]\n\t"
			      "fldi $fd20, [%0+0xa0]\n\t"
			      "fldi $fd19, [%0+0x98]\n\t"
			      "fldi $fd18, [%0+0x90]\n\t"
			      "fldi $fd17, [%0+0x88]\n\t"
			      "fldi $fd16, [%0+0x80]\n\t"
			      :	/* no output */
			      :"r" (fpregs));
		/* fall through */
	case 2:
		asm volatile ("fldi $fd15, [%0+0x78]\n\t"
			      "fldi $fd14, [%0+0x70]\n\t"
			      "fldi $fd13, [%0+0x68]\n\t"
			      "fldi $fd12, [%0+0x60]\n\t"
			      "fldi $fd11, [%0+0x58]\n\t"
			      "fldi $fd10, [%0+0x50]\n\t"
			      "fldi $fd9,  [%0+0x48]\n\t"
			      "fldi $fd8,  [%0+0x40]\n\t"
			      :	/* no output */
			      :"r" (fpregs));
		/* fall through */
	case 1:
		asm volatile ("fldi $fd7,  [%0+0x38]\n\t"
			      "fldi $fd6,  [%0+0x30]\n\t"
			      "fldi $fd5,  [%0+0x28]\n\t"
			      "fldi $fd4,  [%0+0x20]\n\t"
			      :	/* no output */
			      :"r" (fpregs));
		/* fall through */
	case 0:
		asm volatile ("fldi $fd3,  [%1+0x18]\n\t"
			      "fldi $fd2,  [%1+0x10]\n\t"
			      "fldi $fd1,  [%1+0x8]\n\t"
			      "fldi $fd0,  [%1+0x0]\n\t"
			      "lwi  %0, [%1+0x100]\n\t"
			      "fmtcsr	%0\n\t":"=&r" (fpcsr)
			      :"r"(fpregs));
	}
	disable_fpu();
}

void do_fpu_context_switch(unsigned long itype, struct pt_regs *regs,
			   unsigned long eva)
{
	preempt_disable();
	/* Enable to use FPU. */
	
	if (!user_mode(regs)) {
		printk(KERN_ERR "BUG: FPU is used in kernel mode.\n");
		BUG();
		goto end;
	}

	grab_fpu(regs);
#ifndef CONFIG_UNLAZY_FPU	//Lazy FPU is used
	if (last_task_used_math == current)
		goto end;
	if (last_task_used_math != NULL)
		/* Other processes fpu state, save away */
		save_fpu(last_task_used_math);
	last_task_used_math = current;
#endif
	if (used_math()) {
		fpload(&current->thread.fpu);
	} else {
		/* First time FPU user.  */
		fpload(&init_fpuregs);
		set_used_math();
	}
 end:
	preempt_enable();

}

void do_fpu_exception(unsigned long itype, struct pt_regs *regs,
		      unsigned long eva)
{

	siginfo_t si = { 0 };
	unsigned int fpcsr;
	enable_fpu();
	asm volatile ("fmfcsr    %0\n\t":"=&r" (fpcsr));
	disable_fpu();
	
	if (fpcsr & FPCSR_mskALLT) {
		si.si_signo = SIGFPE;
		/* Exception handling, denorm input, UDF and OVF */
		if (fpcsr & FPCSR_mskDNIT) {
			unsigned int rfpcsr;
			lose_fpu(1);
			si.si_signo = do_fpu_denorm(regs, &current->thread.fpu);
			own_fpu(1);

			if (si.si_signo == SIGFPE) {
				rfpcsr = current->thread.fpu.fpcsr;

				if (rfpcsr & FPCSR_mskIVO)
					si.si_code = FPE_FLTINV;
				if (rfpcsr & FPCSR_mskDBZ)
					si.si_code = FPE_FLTDIV;
				if (rfpcsr & FPCSR_mskOVF)
					si.si_code = FPE_FLTOVF;
				if (rfpcsr & FPCSR_mskUDF)
					si.si_code = FPE_FLTUND;
				if (rfpcsr & FPCSR_mskIEX)
					si.si_code = FPE_FLTRES;
			} else if (si.si_code == SIGILL)
				show_regs(regs);
			else if (si.si_code == SIGBUS)
				si.si_code = BUS_ADRERR;
		} else if (fpcsr & FPCSR_mskRIT) {
			printk("Reserved Instruction\n");
			show_regs(regs);
			if (!user_mode(regs))
				do_exit(SIGILL);
			si.si_signo = SIGILL;
		} else if (fpcsr & FPCSR_mskUDFT)
			si.si_code = FPE_FLTUND;
		else if (fpcsr & FPCSR_mskOVFT)
			si.si_code = FPE_FLTOVF;
		else if (fpcsr & FPCSR_mskIVOT)
			si.si_code = FPE_FLTINV;
		else if (fpcsr & FPCSR_mskDBZT)
			si.si_code = FPE_FLTDIV;
		else if (fpcsr & FPCSR_mskIEXT)
			si.si_code = FPE_FLTRES;
		/* If something went wrong, signal */
		if (si.si_signo) {
			if (si.si_code == SIGILL) {
				force_sig(si.si_signo, current);
			} else {
				si.si_addr = (void __user *)
				    instruction_pointer(regs);
				force_sig_info(si.si_signo, &si, current);
			}
		}
	} else {
		printk("Bad FPU exception\n");
		BUG();
	}
}

#define FPU_COPROCESSOR_ID 0
static int __init register_FPU_excpt_handler(void)
{
	if (GET_FUCOP_EXIST() &
	    (FUCOP_EXIST_mskCP0ISFPU | FUCOP_EXIST_mskCP0EX)) {
		coprocsr_trap[FPU_COPROCESSOR_ID].disable_handler
		    = do_fpu_context_switch;
		coprocsr_trap[FPU_COPROCESSOR_ID].excpt_handler
		    = do_fpu_exception;
	}
	return 0;
}

arch_initcall(register_FPU_excpt_handler);
