#include <linux/module.h>
#include <linux/ftrace.h>
#include <linux/uaccess.h>
#include <linux/stop_machine.h>
#include <linux/byteorder/generic.h>
#include <asm/ftrace.h>
#include <asm/cacheflush.h>

#ifdef __NDS32_EL__
#define INSN_NOP 0x09000040
#define INSN_SIZE(insn) (((insn & 0x00000080) == 0) ? 4 : 2)
#define IS_SETHI(insn) ((insn & 0x000000fe) == 0x00000046)
#define ENDIAN_CONVERT(insn) be32_to_cpu(insn)
#else /* __NDS32_EB__ */
#define INSN_NOP 0x40000009
#define INSN_SIZE(insn) (((insn & 0x80000000) == 0) ? 4 : 2)
#define IS_SETHI(insn) ((insn & 0xfe000000) == 0x46000000)
#define ENDIAN_CONVERT(insn) (insn)
#endif

#ifndef CONFIG_DYNAMIC_FTRACE

void __naked noinline ftrace_stub(unsigned long ip, unsigned long parent_ip,
                                  struct ftrace_ops *op, struct pt_regs *regs)
{
        __asm__ ("");  // avoid to optimize as pure function
        return;
}

extern void (*ftrace_trace_function) (unsigned long, unsigned long,
                                      struct ftrace_ops*, struct pt_regs*);
extern int ftrace_graph_entry_stub(struct ftrace_graph_ent *trace);
extern void ftrace_graph_caller(void);

void noinline _mcount(unsigned long parent_ip)
{
        /* save all state needed by the compiler prologue */

        unsigned long ip = (unsigned long)__builtin_return_address(0);

        if (function_trace_stop)
                return;

        if (ftrace_trace_function != ftrace_stub)
                ftrace_trace_function(ip - MCOUNT_INSN_SIZE, parent_ip,
                                      NULL, NULL);

#ifdef CONFIG_FUNCTION_GRAPH_TRACER
        if (ftrace_graph_return != (trace_func_graph_ret_t)ftrace_stub ||
            ftrace_graph_entry != ftrace_graph_entry_stub)
                ftrace_graph_caller();
#endif
        return;

        /* restore all state needed by the compiler epilogue */
}

#else /* CONFIG_DYNAMIC_FTRACE */

void __naked noinline ftrace_stub(unsigned long ip, unsigned long parent_ip,
                                  struct ftrace_ops *op, struct pt_regs *regs)
{
        __asm__ ("");  // avoid to optimize as pure function
        return;
}

void __naked noinline _mcount(unsigned long parent_ip)
{
        __asm__ ("");  // avoid to optimize as pure function
        return;
}

void _ftrace_caller(unsigned long parent_ip)
{
        /* save all state needed by the compiler prologue */

        if (function_trace_stop)
                return;

        /*
         * push arguments for real tracing function
         * first  arg : __builtin_return_address(0) - MCOUNT_INSN_SIZE
         * second arg : parent_ip
         */
        __asm__ __volatile__ ("move $r1, %0        \n\t"
                              "move $r0, %1        \n\t"
                              "addi $r0, $r0, #-12 \n\t"
                              :
                              :"r" (parent_ip),
                               "r" (__builtin_return_address(0)));

        /* a placeholder for the call to a real tracing function */
        __asm__ __volatile__ ("ftrace_call:        \n\t"
                              "nop                 \n\t"
                              "nop                 \n\t"
                              "nop                 \n\t");

#ifdef CONFIG_FUNCTION_GRAPH_TRACER
        /* a placeholder for the call to ftrace_graph_caller */
        __asm__ __volatile__ ("ftrace_graph_call:  \n\t"
                              "nop                 \n\t"
                              "nop                 \n\t"
                              "nop                 \n\t");
#endif
        /* restore all state needed by the compiler epilogue */

        return;
}

#endif /* CONFIG_DYNAMIC_FTRACE */

#ifdef CONFIG_DYNAMIC_FTRACE

int __init ftrace_dyn_arch_init(void *data)
{
        /* The return code is retured via data */
        *(unsigned long *)data = 0;

        return 0;
}

int ftrace_arch_code_modify_prepare(void)
{
        set_all_modules_text_rw();
        return 0;
}

int ftrace_arch_code_modify_post_process(void)
{
        set_all_modules_text_ro();
        return 0;
}

static unsigned long gen_sethi_insn(unsigned long addr)
{
        unsigned long opcode = 0x46000000;
        unsigned long imm = addr >> 12;
        unsigned long rt_num = 0xf << 20;
        return ENDIAN_CONVERT(opcode | rt_num | imm);
}

static unsigned long gen_ori_insn(unsigned long addr)
{
        unsigned long opcode = 0x58000000;
        unsigned long imm = addr & 0x0000fff;
        unsigned long rt_num = 0xf << 20;
        unsigned long ra_num = 0xf << 15;
        return ENDIAN_CONVERT(opcode | rt_num | ra_num | imm);
}

static unsigned long gen_jral_insn(unsigned long addr)
{
        unsigned long opcode = 0x4a000001;
        unsigned long rt_num = 0x1e << 20;
        unsigned long rb_num = 0xf << 10;
        return ENDIAN_CONVERT(opcode | rt_num | rb_num);
}

static void ftrace_gen_call_insns(unsigned long *call_insns,
                                  unsigned long pc, unsigned long addr)
{
        call_insns[0] = gen_sethi_insn(addr); /* sethi $r15, imm20u       */
        call_insns[1] = gen_ori_insn(addr);   /* ori   $r15, $r15, imm15u */
        call_insns[2] = gen_jral_insn(addr);  /* jral  $lp,  $r15         */
}

static int __ftrace_modify_code(unsigned long pc, unsigned long old_insn,
                              unsigned long new_insn, bool validate)
{
        unsigned long orig_insn;
        if (validate) {
                if (probe_kernel_read(&orig_insn,
                                      (void *)pc,
                                      (MCOUNT_INSN_SIZE / 3)))
                        return -EFAULT;

                if (orig_insn != old_insn)
                        return -EINVAL;
        }

        if (probe_kernel_write((void *)pc, &new_insn, (MCOUNT_INSN_SIZE / 3)))
                return -EPERM;

        return 0;
}

static int ftrace_modify_code(unsigned pc, unsigned long *old_insn,
                              unsigned long *new_insn, bool validate)
{
        int ret;

        ret =  __ftrace_modify_code(pc, old_insn[0], new_insn[0], validate);
        if (ret)
                return ret;
        ret = __ftrace_modify_code(pc+0x4, old_insn[1], new_insn[1], validate);
        if (ret)
                return ret;
        ret = __ftrace_modify_code(pc+0x8, old_insn[2], new_insn[2], validate);
        if (ret)
                return ret;

        flush_icache_range(pc, pc + MCOUNT_INSN_SIZE);

        return ret;
}

int ftrace_update_ftrace_func(ftrace_func_t func)
{
        unsigned long pc = (unsigned long)&ftrace_call;
        unsigned long old_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        unsigned long new_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        int ret;

        if (func != ftrace_stub)
                ftrace_gen_call_insns(new_insn, pc, (unsigned long)func);

        ret = ftrace_modify_code(pc, old_insn, new_insn, false);

        return ret;
}

int ftrace_make_call(struct dyn_ftrace *rec, unsigned long addr)
{
        unsigned long pc = rec->ip;
        unsigned long old_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        unsigned long new_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        int ret;

        ftrace_gen_call_insns(new_insn, pc, addr);

        ret = ftrace_modify_code(pc, old_insn, new_insn, true);

        return ret;
}

int ftrace_make_nop(struct module *mod,
                    struct dyn_ftrace *rec, unsigned long addr)
{
        unsigned long pc = rec->ip;
        unsigned long old_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        unsigned long new_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        int ret;

        ftrace_gen_call_insns(old_insn, pc, addr);

        ret = ftrace_modify_code(pc, old_insn, new_insn, true);

        return ret;
}

#endif /* CONFIG_DYNAMIC_FTRACE */

#ifdef CONFIG_FUNCTION_GRAPH_TRACER

void prepare_ftrace_return(unsigned long *parent, unsigned long self_addr,
			   unsigned long frame_pointer);
void noinline ftrace_graph_caller(void)
{
        unsigned long *parent_lp =
            (unsigned long *)(__builtin_frame_address(2) - 4);

        unsigned long selfpc =
            (unsigned long)(__builtin_return_address(1) - MCOUNT_INSN_SIZE);

        unsigned long frame_pointer = (unsigned long)__builtin_frame_address(3);

        prepare_ftrace_return(parent_lp, selfpc, frame_pointer);
}

extern unsigned long ftrace_return_to_handler(unsigned long frame_pointer);
void __naked return_to_handler(void)
{
        __asm__ __volatile__ (/* save state needed by the ABI     */
                              "smw.adm $r0,[$sp],$r1,#0x0  \n\t"

                              /* get original return address      */
                              "move $r0, $fp               \n\t"
                              "bal ftrace_return_to_handler\n\t"
                              "move $lp, $r0               \n\t"

                              /* restore state nedded by the ABI  */
                              "lmw.bim $r0,[$sp],$r1,#0x0  \n\t");
        return;
}

void prepare_ftrace_return(unsigned long *parent, unsigned long self_addr,
			   unsigned long frame_pointer)
{
        unsigned long return_hooker = (unsigned long)&return_to_handler;
        struct ftrace_graph_ent trace;
        unsigned long old;
        int err;

        if (unlikely(atomic_read(&current->tracing_graph_pause)))
                return;

        old = *parent;
        *parent = return_hooker;

        trace.func = self_addr;
        trace.depth = current->curr_ret_stack + 1;

        /* Only trace if the calling function expects to */
        if (!ftrace_graph_entry(&trace)) {
                *parent = old;
                return;
        }

        err = ftrace_push_return_trace(old,
                                       self_addr,
                                       &trace.depth,
                                       frame_pointer);
        if (err == -EBUSY) {
                *parent = old;
                return;
        }
}

#ifdef CONFIG_DYNAMIC_FTRACE

extern unsigned long ftrace_graph_call;

static int ftrace_modify_graph_caller(bool enable)
{
        unsigned long caller_fn = (unsigned long)&ftrace_graph_caller;
        unsigned long pc = (unsigned long)&ftrace_graph_call;
        unsigned long old_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        unsigned long new_insn[3] = {INSN_NOP, INSN_NOP, INSN_NOP};
        int ret;

        if (enable)
               ftrace_gen_call_insns(new_insn, pc, caller_fn);
        else
               ftrace_gen_call_insns(old_insn, pc, caller_fn);

        ret = ftrace_modify_code(pc, old_insn, new_insn, true);

        return ret;
}

int ftrace_enable_ftrace_graph_caller(void)
{
        return ftrace_modify_graph_caller(true);
}

int ftrace_disable_ftrace_graph_caller(void)
{
        return ftrace_modify_graph_caller(false);
}

#endif /* CONFIG_DYNAMIC_FTRACE */

#endif
