/*
 *  arch/nds32/kernel/head.S
 *
 *  NDS32 Kernel startup code
 *
 *  Copyright (C) 2007 Andes Technology Corporation
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 */
#include <linux/linkage.h>
#include <linux/init.h>
#include <asm/assembler.h>
#include <asm/procinfo.h>
#include <asm/ptrace.h>
#include <asm/asm-offsets.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/l2_cache.h>
#include <asm/sizes.h>
#include <asm/thread_info.h>

#ifdef CONFIG_CPU_BIG_ENDIAN
#define OF_DT_MAGIC 0xd00dfeed
#else
#define OF_DT_MAGIC 0xedfe0dd0
#endif

#ifdef CONFIG_PLAT_AG101P
#define UART0_PA_BASE 0x99600000
#elif CONFIG_PLAT_AE3XX
#define UART0_PA_BASE 0xf0300020
#endif

/*
 * We place the page tables 16K below TEXTADDR.  Therefore, we must make sure
 * that TEXTADDR is correctly set.  Currently, we expect the least significant
 * 16 bits to be 0x8000, but we could probably relax this restriction to
 * TEXTADDR >= PAGE_OFFSET + 0x4000
 *
 * Note that swapper_pg_dir is the virtual address of the page tables, and
 * pgtbl gives us a position-independent reference to these tables.  We can
 * do this because _stext == TEXTADDR
 */

	.globl  swapper_pg_dir
	.equ    swapper_pg_dir, TEXTADDR - 0x4000

/*
 * Kernel startup entry point.
 * ---------------------------
 *
 * This is normally called from the decompressor code.  The requirements
 * are: MMU = off, D-cache = disable, I-cache = disable
 *
 * This code is mostly position independent, so if you link the kernel at
 * 0xc0008000, you call this at __pa(0xc0008000).
 *
 */
	.section ".head.text", "ax"
	.type   _stext, %function
ENTRY(_stext)
	setgie.d                            ! Disable interrupt
	isb
/*
 * Disable I/D-cache
 */
	mfsr    $r0, $mr8
	li      $r1, #~0x3
	and     $r0, $r0, $r1
	mtsr    $r0, $mr8

/*
 * Process device tree blob
 */
	andi 	$r0,$r2,#0x3
	li	$r10, 0
	bne     $r0, $r10, _nodtb
	lwi	$r0, [$r2]
	li	$r1, OF_DT_MAGIC
	bne     $r0, $r1, _nodtb
	move	$r10, $r2
_nodtb:

/*
 * Create a temporary mapping area for booting, before start_kernel
 */
	sethi   $r4, hi20(swapper_pg_dir)
	li      $p0, (PAGE_OFFSET - PHYS_OFFSET)
	sub     $r4, $r4, $p0
	tlbop   FlushAll            ! invalidate TLB\n"
	isb
	mtsr    $r4, $L1_PPTB       ! load page table pointer\n"

/* set NTC0 cacheable/writeback, mutliple page size in use */
	mfsr    $r3, $MMU_CTL
	li      $r0, ~0x6
	and     $r3, $r3, $r0
#ifdef CONFIG_ANDES_PAGE_SIZE_4KB
	ori     $r3, $r3, 0x404
#else
	ori     $r3, $r3, 0x405
#endif
	mtsr    $r3, $MMU_CTL
	isb

/* set page size and size of kernel image */
        mfsr    $r0, $MMU_CFG
        srli    $r3, $r0, MMU_CFG_offfEPSZ
        zeb     $r3, $r3
        bnez    $r3, _extra_page_size_support
#ifdef CONFIG_ANDES_PAGE_SIZE_4KB
        li      $r5, #0x1000                 ! Use 4KB page size
#else
        li      $r5, #0x2000                 ! Use 8KB page size
        li      $r3, #1
#endif
        mtsr    $r3, $TLB_MISC
        b       _image_size_check

_extra_page_size_support:                    ! Use epzs pages size
        clz     $r6, $r3
        subri   $r2, $r6, #31
        li      $r3, #1
        sll     $r3, $r3, $r2
        /* MMU_CFG.EPSZ value -> meaning */
        mul     $r5, $r3, $r3
        slli    $r5, $r5, #14
        /* MMU_CFG.EPSZ  -> TLB_MISC.ACC_PSZ */
        addi    $r3, $r2, #0x2
        mtsr    $r3, $TLB_MISC

_image_size_check:
        /* calculate the image maximum size accepted by TLB config */
        andi    $r6, $r0, MMU_CFG_mskTBW
        andi    $r0, $r0, MMU_CFG_mskTBS
        srli    $r6, $r6, MMU_CFG_offTBW
        srli    $r0, $r0, MMU_CFG_offTBS
        /*
         * we just map the kernel to the maximum way - 1 of tlb
         * reserver one way for UART VA mapping
         * it will cause page fault if UART mapping cover the kernel mapping
         *
         * direct mapping is not supported now.
         */
        li      $r2, 't'
        beqz    $r6, __error                 ! MMU_CFG.TBW = 0 is direct mappin
        addi    $r0, $r0, #0x2               ! MMU_CFG.TBS value -> meaning
        sll     $r0, $r6, $r0                ! entries = k-way * n-set
        mul     $r6, $r0, $r5                ! max size = entries * page size
        /* check kernel image size */
        la      $r3, (_end - PAGE_OFFSET)
        li      $r2, 's'
        bgt     $r3, $r6, __error

#ifdef CONFIG_CPU_DCACHE_WRITETHROUGH
	li      $r2, #(PHYS_OFFSET + 0x7bf)
#else
	li      $r2, #(PHYS_OFFSET + 0x6bf) ! to remember here
#endif
        li      $r3, PAGE_OFFSET
        add     $r6, $r6, $r3

_tlb:
	mtsr    $r3, $TLB_VPN
	dsb
	tlbop   $r2, RWR
	isb
	add     $r3, $r3, $r5
	add     $r2, $r2, $r5
	bgt     $r6, $r3, _tlb
	mfsr    $r3, $TLB_MISC      ! setup access page size
	li      $r2, #~0xf
	and     $r3, $r3, $r2
#ifdef CONFIG_ANDES_PAGE_SIZE_8KB
	ori    $r3, $r3, #0x1
#endif
	mtsr    $r3, $TLB_MISC

	mfsr    $r0, $MISC_CTL      ! Enable BTB and RTP
	li      $r1, #~0x3
	and     $r0, $r0, $r1
	mtsr    $r0, $MISC_CTL

/*
 * Disable L2CC and wait until L2CC registers are mapped into memory to use L2$.
 */
#ifdef CONFIG_CACHE_L2
	mfsr	$p0, $MSC_CFG
	andi	$p0, $p0, #MSC_CFG_mskL2C
	beqz	$p0, 1f
	li  	$p0, L2CC_PA_BASE
	li  	$p1, 0
	swi 	$p1, [$p0 + L2CC_CTRL_OFF]
1:
#endif

	mfsr    $p1, $PSW
	li      $r15, #~0x43df          ! clear WBNA|DME|IME|DT|IT|POM|INTL|GIE
	and     $p1, $p1, $r15
#ifdef __NDS32_EB__
 #ifdef CONFIG_WBNA
	ori     $p1, $p1, #0x40ea       ! set   WBNA|DT|IT|BE|POM:super|INTL:1
 #else
	ori     $p1, $p1, #0xea         ! set   ----|DT|IT|BE|POM:super|INTL:1
 #endif
#else
 #ifdef CONFIG_WBNA
	ori     $p1, $p1, #0x40ca       ! set   WBNA|DT|IT|--|POM:super|INTL:1
 #else
	ori     $p1, $p1, #0xca         ! set   ----|DT|IT|--|POM:super|INTL:1
 #endif
#endif

	mtsr    $p1, $IPSW              ! when iret, it will automatically enable MMU
	la      $lp, __mmap_switched
	mtsr    $lp, $IPC
	iret
	nop

	.type   __switch_data, %object
__switch_data:
	.long   __bss_start                 ! $r6
	.long   _end                        ! $r7
	.long	__atags_pointer 	    ! $atag_pointer
	.long   init_thread_union + THREAD_SIZE    ! $sp


/*
 * The following fragment of code is executed with the MMU on in MMU mode,
 * and uses absolute addresses; this is not position independent.
 */
	.align
	.type   __mmap_switched, %function
__mmap_switched:
	la  $r3, __switch_data
	lmw.bim $r6, [$r3], $r8, #0b0001
	move    $fp, #0             ! Clear  BSS (and zero $fp)
	beq $r7, $r6, _RRT
1:	swi.bi  $fp, [$r6], #4
	bne $r7, $r6, 1b
	swi	$r10, [$r8]

_RRT:
	b   start_kernel

/*
 * Exception handling.
 *
 * s = exceed limit of  maximum size of kernel
 * t = tlb is direct mapping
 *
 */
__error:
        li  $r1, UART0_PA_BASE
        sw  $r2, [$r1]
die:    b   die

#ifdef CONFIG_SMP

		.type   secondary_startup, %function
ENTRY(secondary_startup)
		/*
		 * Common entry point for secondary CPUs.
		 *
		 * Lookup the processor type - there is no need to check the
		 * machine type as it has already been validated by the
		 * primary processor.
		 */
	mfsr    $r0, $MMU_CTL
	ori     $r0, $r0, #4
#ifndef CONFIG_NO_KERNEL_LARGE_PAGE
	ori     $r0, $r0, #0x400
#endif
	mtsr    $r0, $MMU_CTL

	movi    $r15, #0x01
	swi     $r15, [$p1+#0x10]
	lwi     $sp, [$p1+#0x18]

	/*
	 * Set stack, L1_PPTB, and enable mmu
	 */
	sethi   $r4, hi20(swapper_pg_dir)
	li      $p0, (PAGE_OFFSET - PHYS_OFFSET)
	sub     $r4, $r4, $p0

	tlbop   FlushAll
	isb
	mtsr    $r4, $L1_PPTB

#ifdef CONFIG_CACHE_L2
	mfsr	$p0, $MSC_CFG
	andi	$p0, $p0, #MSC_CFG_mskL2C
	beqz	$p0, 1f
	li      $r0, #0x1801
	mtsr    $r0, $HSMP_SADDR
	isb
	li      $r0, #(L2CC_PA_BASE+0x10)      ! L2CC Control
	lwi     $r1, [$r0]
	li      $r2, #~(0xf << 28)
	and     $r1, $r1, $r2
	bset    $r1, $r1, #29
	bset    $r1, $r1, #31
	swi     $r1, [$r0]
1:
#endif

	move    $fp, #0

	mfsr    $p1, $PSW
	li      $r15, #~0x43df          ! clear WBNA|DME|IME|DT|IT|POM|INTL|GIE
	and     $p1, $p1, $r15
#ifdef __NDS32_EB__
 #ifdef CONFIG_WBNA
	ori     $p1, $p1, #0x40ea       ! set   WBNA|DT|IT|BE|POM:super|INTL:1
 #else
	ori     $p1, $p1, #0xea         ! set   ----|DT|IT|BE|POM:super|INTL:1
 #endif
#else
 #ifdef CONFIG_WBNA
	ori     $p1, $p1, #0x40ca       ! set   WBNA|DT|IT|--|POM:super|INTL:1
 #else
	ori     $p1, $p1, #0xca         ! set   ----|DT|IT|--|POM:super|INTL:1
 #endif
#endif

	mtsr    $p1, $IPSW              ! when iret, it will automatically enable MMU
	la      $lp, secondary_start_kernel
	mtsr    $lp, $IPC
	iret
	nop
#endif
