/*
 *  linux/arch/nds32/kernel/irq.c
 *
 *  Copyright (C) 1992 Linus Torvalds
 *  Modifications for ARM processor Copyright (C) 1995-2000 Russell King.
 *  Copyright (C) 2009 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  This file contains the code used by various IRQ handling routines:
 *  asking for different IRQ's should be done through these routines
 *  instead of just grabbing them. Thus setups with different IRQ numbers
 *  shouldn't result in any weird surprises, and installing new handlers
 *  should be easier.
 *
 *  IRQ's are in fact implemented a bit like signal handlers for the kernel.
 *  Naturally it's not a 1:1 relation, but there are similarities.
 */
#include <linux/interrupt.h>
#include <linux/irqchip.h>

/*
 * do_IRQ handles all hardware IRQ's.  Decoded IRQs should not
 * come via this function.  Instead, they should provide their
 * own 'handler'
 */
extern struct irq_domain *root_domain;
#include <linux/irqdomain.h>
asmlinkage void asm_do_IRQ(unsigned int irq, struct pt_regs *regs)
{
	struct pt_regs *old_regs = set_irq_regs(regs);
	int virq;

	/*
	 * Some hardware gives randomly wrong interrupts.  Rather
	 * than crashing, do something sensible.
	 */
	if (unlikely(irq >= NR_IRQS)) {
		printk(KERN_EMERG "IRQ exceeds NR_IRQS\n");
		BUG();
	}

	irq_enter();
	virq = irq_find_mapping(root_domain, irq);
	generic_handle_irq(virq);
	irq_exit();
	set_irq_regs(old_regs);

}

void __init init_IRQ(void)
{
	irqchip_init();
}

#ifdef CONFIG_TRACE_IRQFLAGS
void notrace arch_trace_hardirqs_on(void)
{
	trace_hardirqs_on();
}

void notrace arch_trace_hardirqs_off(void)
{
	trace_hardirqs_off();
}
#endif
