/*
 *  linux/arch/nds32/kernel/sys_nds32.c
 *
 *  Copyright (C) People who wrote linux/arch/i386/kernel/sys_i386.c
 *  Copyright (C) 2007 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  This file contains various random system calls that
 *  have a non-standard calling sequence on the Linux/nds32
 *  platform.
 */
#include <linux/syscalls.h>

#include <asm/uaccess.h>
#include <asm/pfm.h>
#include <asm/cachectl.h>
#include <asm/proc-fns.h>

asmlinkage unsigned long sys_getpagesize(void)
{
	return PAGE_SIZE;
}

asmlinkage unsigned long sys_mmap2(unsigned long addr, unsigned long len,
				   unsigned long prot, unsigned long flags,
				   unsigned long fd, unsigned long pgoff)
{
	if (pgoff & (~PAGE_MASK >> 12))
		return -EINVAL;

	return sys_mmap_pgoff(addr, len, prot, flags, fd,
			      pgoff >> (PAGE_SHIFT - 12));
}

/*
 * Perform the select(nd, in, out, ex, tv) and mmap() system
 * calls.
 */

struct sel_arg_struct {
	unsigned long n;
	fd_set __user *inp, *outp, *exp;
	struct timeval __user *tvp;
};

asmlinkage int old_select(struct sel_arg_struct __user * arg)
{
	struct sel_arg_struct a;

	if (copy_from_user(&a, arg, sizeof(a)))
		return -EFAULT;
	/* sys_select() does the appropriate kernel locking */
	return sys_select(a.n, a.inp, a.outp, a.exp, a.tvp);
}

asmlinkage long sys_fadvise64_64_wrapper(int fd, int advice, loff_t offset,
					 loff_t len)
{
	return sys_fadvise64_64(fd, offset, len, advice);
}


int sys_cacheflush(unsigned int start, unsigned int end, int cache)
{
	struct vm_area_struct *vma;
	bool flushi = true, wbd = true;

	vma = find_vma(current->mm, start);
	if (!vma)
		return -EFAULT;
	switch (cache) {
	case ICACHE:
		wbd = false;
		break;
	case DCACHE:
		flushi = false;
		break;
	case BCACHE:
		break;
	default:
		return -EINVAL;
	}
	cpu_cache_wbinval_range_check(vma, start, end, flushi, wbd);

	return 0;
}

asmlinkage int nds32_syscall(int number, struct pt_regs *regs)
{
	switch (number) {
#ifdef CONFIG_OPROFILE
	case __NR_pfmctl:
		sys_pfmctl(regs->NDS32_r1, regs->NDS32_r2, regs->NDS32_r3,
			   regs->NDS32_r4);
		return 0;

	case __NR_getpfm:
		return sys_getpfm((struct pcounter __user *)regs->NDS32_r1);

	case __NR_setpfm:
		return sys_setpfm(regs->NDS32_r1, regs->NDS32_r2,
				  regs->NDS32_r3,
				  (struct pcounter __user *)regs->NDS32_r4);
#endif
	case __NR_wbna:
		if (regs->NDS32_r1)
			regs->NDS32_ipsw |= PSW_mskWBNA;
		else
			regs->NDS32_ipsw &= ~PSW_mskWBNA;
		return 0;

	default:
		return -ENOSYS;
	}
	return -ENOSYS;
}
