/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		MIPS specific IP/TCP/UDP checksumming routines
 *
 * Authors:	Ralf Baechle, <ralf@waldorf-gmbh.de>
 *		Lots of code moved from tcp.c and ip.c; see those files
 *		for more names.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */
#include <linux/module.h>
#include <linux/types.h>

#include <net/checksum.h>
#include <asm/byteorder.h>
#include <asm/string.h>
#include <asm/uaccess.h>

#define addc(_t,_r)                     \
	__asm__ __volatile__ (          \
        "add\t%0, %0, %1\n\t"		\
        "slt\t$p1, %0, %1\n\t"          \
        "add\r%0, %0, $p1\n\t"		\
	: "=r"(_t)                      \
	: "r"(_r), "0"(_t));

static inline unsigned short from32to16(unsigned int x)
{
	/* 32 bits --> 16 bits + carry */
	x = (x & 0xffff) + (x >> 16);
	/* 16 bits + carry --> 16 bits including carry */
	x = (x & 0xffff) + (x >> 16);
	return (unsigned short)x;
}

static inline unsigned int do_csum(const unsigned char *buff, int len)
{
	int odd;
	register unsigned int result = 0;
	register unsigned int count;
	if (len <= 0)
		goto out;
	odd = 1 & (unsigned long)buff;
	if (odd) {
		result = be16_to_cpu(*buff);
		len--;
		buff++;
	}
	count = len >> 1;	/* nr of 16-bit words.. */
	if (count) {
		if (2 & (unsigned long)buff) {
			result += *(unsigned short *)buff;
			count--;
			len -= 2;
			buff += 2;
		}
		count >>= 1;	/* nr of 32-bit words.. */
		if (count) {
			while (count >= 8) {
				__asm__
				    __volatile__("lmw.bi $r17, [%1], $r24\n\t"
						 "add\t%0, %0, $r17\n\t"
						 "slt\t$p1, %0, $r17\n\t"
						 "add\r%0, %0, $p1\n\t"
						 "add\t%0, %0, $r18\n\t"
						 "slt\t$p1, %0, $r18\n\t"
						 "add\r%0, %0, $p1\n\t"
						 "add\t%0, %0, $r19\n\t"
						 "slt\t$p1, %0, $r19\n\t"
						 "add\r%0, %0, $p1\n\t"
						 "add\t%0, %0, $r20\n\t"
						 "slt\t$p1, %0, $r20\n\t"
						 "add\r%0, %0, $p1\n\t"
						 "add\t%0, %0, $r21\n\t"
						 "slt\t$p1, %0, $r21\n\t"
						 "add\r%0, %0, $p1\n\t"
						 "add\t%0, %0, $r22\n\t"
						 "slt\t$p1, %0, $r22\n\t"
						 "add\r%0, %0, $p1\n\t"
						 "add\t%0, %0, $r23\n\t"
						 "slt\t$p1, %0, $r23\n\t"
						 "add\r%0, %0, $p1\n\t"
						 "add\t%0, %0, $r24\n\t"
						 "slt\t$p1, %0, $r24\n\t"
						 "add\r%0, %0, $p1\n\t":"=r"
						 (result)
						 :"r"(buff), "0"(result)
						 :"$r17", "$r18", "$r19",
						 "$r20", "$r21", "$r22", "$r23",
						 "$r24");
				count -= 8;
				buff += 32;
			}
			while (count) {
				unsigned int w = *(unsigned int *)buff;
				count--;
				buff += 4;
				addc(result, w);
			}
			result = (result & 0xffff) + (result >> 16);
		}
		if (len & 2) {
			result += *(unsigned short *)buff;
			buff += 2;
		}
	}
	if (len & 1)
		result += le16_to_cpu(*buff);
	result = from32to16(result);
	if (odd)
		result = swab16(result);
out:
	return result;
}

/*
 * computes a partial checksum, e.g. for TCP/UDP fragments
 */
/*
 * why bother folding?
 */
unsigned int csum_partial(const void *buff, int len, unsigned int sum)
{
	unsigned int result = 0;
//      printk("csum_partial %x %x %x\n", buff, len, sum);
	result = do_csum(buff, len);
	addc(result, sum);
	return (unsigned short)from32to16(result);
}
