/*
 *  linux/arch/nds32/lib/findbit.S
 *
 *  Copyright (C) 1995-2000 Russell King
 *  Copyright (C) 2006 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/linkage.h>

	.text

/*
 * Purpose  : Find a 'zero' bit
 * Prototype: int find_first_zero_bit(void *addr, int maxbit);
 */
ENTRY(_find_first_zero_bit)
	move	$r2, #0
	move	$p0, #0				! Reset bit count
next_word:
	move	$r4, #4
	lmw.bim	$p1, [$r0], $p1
	li	$r3, #0xffffffff		! Inversion mask
	xor	$p1, $p1, $r3			! Inverted the word
	addi	$p0, $p0, #32
	beqz	$p1, next_word
	addi	$p0, $p0, #-32
next_byte:
	beqz	$r4, next_word
	andi	$r3, $p1, #0xff			! Get the next byte
	bnez	$r3, found			! Found zero bit in this byte
	addi	$p0, $p0, #8			! Update bit count
	addi	$r4, $r4, #-1
	srli	$p1, $p1, #8
	bgt	$r1, $p0, next_byte
	addi	$r0, $r1, #1			! Return not found
	ret

/*
 * Purpose  : Find next 'zero' bit
 * Prototype: int find_next_zero_bit(void *addr, int maxbit, int offset)
 */
ENTRY(_find_next_zero_bit)
	beqz	$r2, _find_first_zero_bit	! If offset=0, goto find_first_zero_bit
	srli	$p0, $r2, #5			! Get offset byte count
	slli	$p0, $p0, #2
	add	$r0, $r0, $p0			! $r0 is the first word to load
	lmw.bim	$p1, [$r0], $p1
	li	$p0, #0xffffffff		! Inversion mask
	xor	$p1, $p1, $p0			! Inverted the word
	andi	$r4, $r2, #31			! Left bits in offset
	srl	$p1, $p1, $r4			! Shift out the left bits
	xor	$p0, $p0, $p0
	subri	$r4, $r4, #31
loop:
	andi	$r3, $p1, #0xff			! The first byte to check
	bnez	$r3, found			! Found zero bit in this byte
	addi	$p0, $p0, #8
	addi	$r4, $r4, #-8
	srli	$p1, $p1, #8			! Move on to the next byte
	bgtz	$r4, loop
	b	next_word

/*
 * One or more bits in the LSB of $p1 are assumed to be set.
 */

found:
	move	$p1, $r3
	xor	$r4, $r4, $r4
	andi	$r5, $p1, #0x0f			! Get bits 0-3
	move	$r3, #4				! For 0 case (no set bit found)
	cmovn	$r3, $r4, $r5			! Not 0 case (There's set bit in these 4 bits)
	add	$p0, $p0, $r3			! Update bit count
	slli	$r3, $p1, #4			! Not 0 case (Find set bit in these 4 bits)
	cmovn	$p1, $r3, $r5			! For 0 case (Find set bit in the rest 4 bits)
	andi	$r5, $p1, #0x30			! Get 4-5
	move	$r3, #2				! For 0 case (no set bit found)
	cmovn	$r3, $r4, $r5			! Not 0 case (There's set bit in these 2 bits)
	add	$p0, $p0, $r3			! Update bit count
	slli	$r3, $p1, #2			! Not 0 case (Find set bit in these 2 bits)
	cmovn	$p1, $r3, $r5			! For 0 case (Find set bit in the rest 2 bits)
	andi	$r5, $p1, #0x40			! Get bit 6
	move	$r3, #1				! For 0 case (bit 6 is not set)
	cmovn	$r3, $r4, $r5			! Not 0 case (bit 6 is set bit)
	add	$r5, $p0, $r3
	add	$r0, $r5, $r2
	ret

ENTRY(_ext2_find_first_zero_bit)
	pushm	$r2, $r4
	move	$r2, #0
	move	$p0, #0				! Reset bit count
1:
	lbi.bi	$r3, [$r0], #1
	xori	$r3, $r3, #0xff
	addi	$p0, $p0, #8
	beqz	$r3, 1b
	addi	$p0, $p0, #-8

2:
	b	found				! No return
	popm	$r2, $r4
	ret

