/*
 *  linux/arch/nds32/lib/getuser.S
 *
 *  Copyright (C) 2001 Russell King
 *  Copyright (C) 2009 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Idea from x86 version, (C) Copyright 1998 Linus Torvalds
 *
 * These functions have a non-standard call interface to make them more
 * efficient, especially as they return an error value in addition to
 * the "real" return value.
 *
 * __get_user_X
 *
 * Inputs:	$r0 contains the address
 * Outputs:	$r0 is the error code
 *		$r2, $r3 contains the zero-extended value
 *		lr corrupted
 *
 * No other registers must be altered.  (see include/asm-nds32/uaccess.h
 * for specific ASM register usage).
 *
 * Note that ADDR_LIMIT is either 0 or 0xc0000000.
 * Note also that it is intended that __get_user_bad is not global.
 */
#include <asm/asm-offsets.h>
#include <asm/thread_info.h>
#include <asm/errno.h>
#include <linux/linkage.h>


ENTRY(__get_user_1)
1:	lbi	$r2, [$r0]
	move	$r0, #0
	ret

ENTRY(__get_user_2)
2:	lbi.bi	$r2, [$r0], #1
3:	lbi	$r3, [$r0]
#ifndef __NDS32_EB__
	slli	$p1, $r3, #8
	or	$r2, $r2, $p1
#else
	slli	$p1, $r2, #8
	or	$r2, $r3, $p1
#endif
	move	$r0, #0
	ret

ENTRY(__get_user_4)
4:	lwi	$r2, [$r0]
	move	$r0, #0
	ret

ENTRY(__get_user_8)
5:	lwi.bi	$r2, [$r0], #4
6:	lwi	$r3, [$r0]
	move	$r0, #0
	ret

__get_user_bad_8:
	move	$r3, #0
__get_user_bad:
	move	$r2, #0
	move	$r0, #-EFAULT
	ret

.section __ex_table, "a"
	.long	1b, __get_user_bad
	.long	2b, __get_user_bad
	.long	3b, __get_user_bad
	.long	4b, __get_user_bad
	.long	5b, __get_user_bad_8
	.long	6b, __get_user_bad_8
.previous
