/*
 *  linux/arch/nds32/lib/putuser.S
 *
 *  Copyright (C) 2001 Russell King
 *  Copyright (C) 2009 Andes Technology Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Idea from x86 version, (C) Copyright 1998 Linus Torvalds
 *
 * These functions have a non-standard call interface to make
 * them more efficient, especially as they return an error
 * value in addition to the "real" return value.
 *
 * __put_user_X
 *
 * Inputs:	$r0 contains the address
 *		$r2, $r3 contains the value
 * Outputs:	$r0 is the error code
 *		lr corrupted
 *
 * No other registers must be altered.  (see include/asm-arm/uaccess.h
 * for specific ASM register usage).
 *
 * Note that ADDR_LIMIT is either 0 or 0xc0000000
 * Note also that it is intended that __put_user_bad is not global.
 */
#include <asm/asm-offsets.h>
#include <asm/thread_info.h>
#include <asm/errno.h>
#include <linux/linkage.h>

	.text

ENTRY(__put_user_1)
1:	sb	$r2, [$r0]
	move	$r0, #0
	ret

ENTRY(__put_user_2)
2:	shi	$r2, [$r0]	! Store input halfword
	move	$r0, #0
	ret

ENTRY(__put_user_4)
3:	sw	$r2, [$r0]
	move	$r0, #0
	ret

ENTRY(__put_user_8)
5:	swi.bi	$r2, [$r0], #4
6:	sw	$r3, [$r0]
	move	$r0, #0
	ret

__put_user_bad:
	move	$r0, #-EFAULT
	ret

.section __ex_table, "a"
	.long	1b, __put_user_bad
	.long	2b, __put_user_bad
	.long	3b, __put_user_bad
	.long	5b, __put_user_bad
	.long	6b, __put_user_bad
.previous
