/*
 *  linux/arch/armnommu/mach-faraday/platform-a320/apb_intc.c
 *
 *  Faraday AHB DMA Interrupt Process Driver Implementation
 *
 *  Copyright (C) 2005 Faraday Corp. (http://www.faraday-tech.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 * ChangeLog
 *
 *  Peter Liao 09/28/2005 Created
 */
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/dmad.h>

#ifdef CONFIG_PLATFORM_AHBDMA
extern int dmad_probe_irq_source_ahb(void);
void AHBDMA_irq_rounter(unsigned int irq, struct irq_desc *desc)
{
	int ahb_irq;
	struct irq_desc *ahb_desc;
	struct irq_data data;
	data.irq = irq;

	raw_spin_lock(&desc->lock);
	desc->irq_data.chip->irq_mask(&data);
	desc->irq_data.chip->irq_ack(&data);

	ahb_irq = dmad_probe_irq_source_ahb();
	if (ahb_irq >= 0) {
		ahb_irq += DMA_IRQ0;
		ahb_desc = irq_desc + ahb_irq;

		raw_spin_unlock(&desc->lock);
		ahb_desc->handle_irq(ahb_irq, ahb_desc);
		raw_spin_lock(&desc->lock);
	}

	desc->irq_data.chip->irq_unmask(&data);
	raw_spin_unlock(&desc->lock);
}

int intc_ftdmac020_init_irq(int irq)
{
	int i;
	int ret;
	/* Register all IRQ */
	for (i = DMA_IRQ0;
	     i < DMA_IRQ0 + DMA_IRQ_COUNT; i++) {
		// level trigger
		ret = irq_set_chip(i, &dummy_irq_chip);
		irq_set_handler(i, handle_simple_irq);
	}
	irq_set_chained_handler(irq, AHBDMA_irq_rounter);
	return 0;
}
#endif /* CONFIG_PLATFORM_AHBDMA */

#ifdef CONFIG_PLATFORM_APBDMA
extern int dmad_probe_irq_source_apb(void);
void APBDMA_irq_rounter(unsigned int irq, struct irq_desc *desc)
{
	int apb_irq;
	struct irq_desc *apb_desc;
	struct irq_data data;
	data.irq = irq;

	raw_spin_lock(&desc->lock);

	//mask_ack_irq(desc, irq);
	desc->irq_data.chip->irq_mask(&data);
	desc->irq_data.chip->irq_ack(&data);

	apb_irq = dmad_probe_irq_source_apb();

	if (apb_irq >= 0) {
		apb_irq += APBBRG_FTAPBBRG020S_IRQ0;
		apb_desc = irq_desc + apb_irq;

		raw_spin_unlock(&desc->lock);
		apb_desc->handle_irq(irq, apb_desc);
		raw_spin_lock(&desc->lock);
	}

	desc->irq_data.chip->irq_unmask(&data);
	raw_spin_unlock(&desc->lock);
}

int __init intc_ftapbbrg020s_init_irq(void)
{
	int i;

	/* Register all IRQ */
	for (i = APBBRG_FTAPBBRG020S_IRQ0;
	     i < APBBRG_FTAPBBRG020S_IRQ0 + APBBRG_FTAPBBRG020S_IRQ_COUNT;
	     i++) {
		// level trigger
		irq_set_chip(i, &dummy_irq_chip);
		irq_set_handler(i, handle_simple_irq);
	}

	irq_set_chained_handler(PLATFORM_APBDMA_IRQ, APBDMA_irq_rounter);

	return 0;
}

arch_initcall(intc_ftapbbrg020s_init_irq);
#endif /* CONFIG_PLATFORM_APBDMA */
