#ifndef SSP_FARADAY_H
#define SSP_FARADAY_H

#define XILINX_SPI_NAME "faraday-spi"

struct at_ts_platdata {
	resource_size_t regs;
	unsigned int x_max;
	unsigned int y_max;
};


/*
 * Register definitions as per "OPB Serial Peripheral Interface ( SPI) ( v1.00e)
 * Product Specification", DS464
 */
#define XSPI_CR_OFFSET		0x62	/* 16-bit Control Register */

#define XSPI_CR_ENABLE		0x02
#define XSPI_CR_MASTER_MODE	0x04
#define XSPI_CR_CPOL		0x08
#define XSPI_CR_CPHA		0x10
#define XSPI_CR_MODE_MASK	( XSPI_CR_CPHA | XSPI_CR_CPOL)
#define XSPI_CR_TXFIFO_RESET	0x20
#define XSPI_CR_RXFIFO_RESET	0x40
#define XSPI_CR_MANUAL_SSELECT	0x80
#define XSPI_CR_TRANS_INHIBIT	0x100

#define XSPI_SR_OFFSET		0x67	/* 8-bit Status Register */

#define XSPI_SR_RX_EMPTY_MASK	0x01	/* Receive FIFO is empty */
#define XSPI_SR_RX_FULL_MASK	0x02	/* Receive FIFO is full */
#define XSPI_SR_TX_EMPTY_MASK	0x04	/* Transmit FIFO is empty */
#define XSPI_SR_TX_FULL_MASK	0x08	/* Transmit FIFO is full */
#define XSPI_SR_MODE_FAULT_MASK	0x10	/* Mode fault error */

#define XSPI_TXD_OFFSET		0x6b	/* 8-bit Data Transmit Register */
#define XSPI_RXD_OFFSET		0x6f	/* 8-bit Data Receive Register */

#define XSPI_SSR_OFFSET		0x70	/* 32-bit Slave Select Register */

/* Register definitions as per "OPB IPIF ( v3.01c) Product Specification", DS414
 * IPIF registers are 32 bit
 */
#define XIPIF_V123B_DGIER_OFFSET	0x1c	/* IPIF global int enable reg */
#define XIPIF_V123B_GINTR_ENABLE	0x80000000

#define XIPIF_V123B_IISR_OFFSET		0x20	/* IPIF interrupt status reg */
#define XIPIF_V123B_IIER_OFFSET		0x28	/* IPIF interrupt enable reg */

#define XSPI_INTR_MODE_FAULT		0x01	/* Mode fault error */
#define XSPI_INTR_SLAVE_MODE_FAULT	0x02	/* Selected as slave while disabled */
#define XSPI_INTR_TX_EMPTY		0x04	/* TxFIFO is empty */
#define XSPI_INTR_TX_UNDERRUN		0x08	/* TxFIFO was underrun */
#define XSPI_INTR_RX_FULL		0x10	/* RxFIFO is full */
#define XSPI_INTR_RX_OVERRUN		0x20	/* RxFIFO was overrun */

#define XIPIF_V123B_RESETR_OFFSET	0x40	/* IPIF reset register */
#define XIPIF_V123B_RESET_MASK		0x0a	/* the value to write */

/*************************************************************************/
#define SSP_REG_CR0		0x00 /* SSP Control Register 0 */
#define SSP_REG_CR1		0x04 /* SSP Control Register 1 */
#define SSP_REG_CR2		0x08 /* SSP Control Register 2 */
#define SSP_REG_SR		0x0C /* SSP Status Register */
#define SSP_REG_ICR		0x10 /* SSP Interrupt Control Register */
#define SSP_REG_ISR		0x14 /* SSP Interrupt Status Register */
#define SSP_REG_DR		0x18 /* SSP Data Register */
#define SSP_REG_ACL		0x20 /* AC-Link Slot Valid Register */
#define SSP_REG_REV		0x60 /* SSP Revision Register */
#define SSP_REG_FEA		0x64 /* SSP Feature Register */

#define SSP_CR0_offFFMT		12	/* Frame Format */
#define SSP_CR0_offFSDIST	8	/* Frame/Sync and Data Distance */
#define SSP_CR0_offLBM		7	/* Loopback Mode */
#define SSP_CR0_offLSB		6	/* Bit Sequence Indicator */
#define SSP_CR0_offFSPO		5	/* Frame/Sync Polarity */
#define SSP_CR0_offFSJSFY	4	/* Data Justify */
#define SSP_CR0_offOPM		2	/* Operation Mode */
#define SSP_CR0_offSCLKPO	1	/* SCLK Polarity */
#define SSP_CR0_offSCLKPH	0	/* SCLK Phase */

#define SSP_CR0_mskFFMT		( 0x07UL << SSP_CR0_offFFMT)
#define SSP_CR0_mskFSDIST	( 0x03UL << SSP_CR0_offFSDIST)
#define SSP_CR0_mskLBM		( 0x01UL << SSP_CR0_offLBM)
#define SSP_CR0_mskLSB		( 0x01UL << SSP_CR0_offLSB)
#define SSP_CR0_mskFSPO		( 0x01UL << SSP_CR0_offFSPO)
#define SSP_CR0_mskFSJSFY	( 0x01UL << SSP_CR0_offFSJSFY)
#define SSP_CR0_mskOPM		( 0x03UL << SSP_CR0_offOPM)
#define SSP_CR0_mskSCLKPO	( 0x01UL << SSP_CR0_offSCLKPO)
#define SSP_CR0_mskSCLKPH	( 0x01UL << SSP_CR0_offSCLKPH)

#define SSP_CR1_offPDL		24	/* Padding Data Length */
#define SSP_CR1_offSDL		16	/* Serial Data Length */
#define SSP_CR1_offSCLKDIV	0	/* SCLK Divider */

#define SSP_CR1_mskPDL		( 0xFFUL << SSP_CR1_offPDL)
#define SSP_CR1_mskSDL		( 0x1FUL << SSP_CR1_offSDL)
#define SSP_CR1_mskSCLKDIV	( 0xFFUL << SSP_CR1_offSCLKDIV)

#define SSP_CR2_offSSPRST	6	/* SSP Reset */
#define SSP_CR2_offACCRST	5	/* AC-Link Cold Reset Enable */
#define SSP_CR2_offACWRST	4	/* AC-Link Warm Reset Enable */
#define SSP_CR2_offTXFCLR	3	/* Transmit FIFO Clear */
#define SSP_CR2_offRXFCLR	2	/* Recieve FIFO clear */
#define SSP_CR2_offTXDOE	1	/* Transmit Data Output Enable */
#define SSP_CR2_offSSPEN	0	/* The SSP Enable */

#define SSP_CR2_mskSSPRST	( 0x01UL << SSP_CR2_offSSPRST)
#define SSP_CR2_mskACCRST	( 0x01UL << SSP_CR2_offACCRST)
#define SSP_CR2_mskACWRST	( 0x01UL << SSP_CR2_offACWRST)
#define SSP_CR2_mskTXFCLR	( 0x01UL << SSP_CR2_offTXFCLR)
#define SSP_CR2_mskRXFCLR	( 0x01UL << SSP_CR2_offRXFCLR)
#define SSP_CR2_mskTXDOE	( 0x01UL << SSP_CR2_offTXDOE)
#define SSP_CR2_mskSSPEN	( 0x01UL << SSP_CR2_offSSPEN)

#define SSP_SR_offTFVE		12	/* Transmit FIFO Valid Entries */
#define SSP_SR_offRFVE		4	/* Recieve FIFO Valid Entries */
#define SSP_SR_offBUSY		2	/* Busy Indicator */
#define SSP_SR_offTFNF		1	/* Transmit FIFO not full */
#define SSP_SR_offRFF		0	/* Recieve FIFO full */

#define SSP_SR_mskTFVE		( 0x1FUL << SSP_SR_offTFVE)
#define SSP_SR_mskRFVE		( 0x1FUL << SSP_SR_offRFVE)
#define SSP_SR_mskBUSY		( 0x01UL << SSP_SR_offBUSY)
#define SSP_SR_mskTFNF		( 0x01UL << SSP_SR_offTFNF)
#define SSP_SR_mskRFF		( 0x01UL << SSP_SR_offRFF)

#define SSP_ICR_offTFTHOD	12	/* Transmit FIFO Threshold */
#define SSP_ICR_offRFTHOD	8	/* Recieve FIFO Threshold */
#define SSP_ICR_offAC97FCEN	6	/* AC97 Frame Complete */
#define SSP_ICR_offTFDMAEN	5	/* Transmit DMA Request Enable */
#define SSP_ICR_offRFDMAEN	4	/* Recieve DMA Request Enable */
#define SSP_ICR_offTFTHIEN	3	/* Transmit FIFO Threshold Interrupt */
#define SSP_ICR_offRFTHIEN	2	/* Recieve FIFO Threshold Interrupt */
#define SSP_ICR_offTFURIEN	1	/* Transmit FIFO Underrun Interrupt Enable */
#define SSP_ICR_offRFORIEN	0	/* Recieve FIFO Overrun Interrupt Enable */

#define SSP_ICR_mskTFTHOD	( 0x0FUL << SSP_ICR_offTFTHOD)
#define SSP_ICR_mskRFTHOD	( 0x0FUL << SSP_ICR_offRFTHOD)
#define SSP_ICR_mskAC97FCEN	( 0x01UL << SSP_ICR_offAC97FCEN)
#define SSP_ICR_mskTFDMAEN	( 0x01UL << SSP_ICR_offTFDMAEN)
#define SSP_ICR_mskRFDMAEN	( 0x01UL << SSP_ICR_offRFDMAEN)
#define SSP_ICR_mskTFTHIEN	( 0x01UL << SSP_ICR_offTFTHIEN)
#define SSP_ICR_mskRFTHIEN	( 0x01UL << SSP_ICR_offRFTHIEN)
#define SSP_ICR_mskTFURIEN	( 0x01UL << SSP_ICR_offTFURIEN)
#define SSP_ICR_mskRFORIEN	( 0x01UL << SSP_ICR_offRFORIEN)

#define SSP_ISR_offAC97FCI	4	/* AC97 Frame Complete Interrupt */
#define SSP_ISR_offTFTHI	3	/* Transmit FIFO Threshold Interrupt */
#define SSP_ISR_offRFTHI	2	/* Recieve FIFO Threshold Interrupt */
#define SSP_ISR_offTFURI	1	/* Transmit FIFO underrun Interrupt */
#define SSP_ISR_offRFORI	0	/* Recieve FIFO Overun Interrupt */

#define SSP_ISR_mskAC97FCI	( 0x01UL << SSP_ISR_offAC97FCI)
#define SSP_ISR_mskTFTHI	( 0x01UL << SSP_ISR_offTFTHI)
#define SSP_ISR_mskRFTHI	( 0x01UL << SSP_ISR_offRFTHI)
#define SSP_ISR_mskTFURI	( 0x01UL << SSP_ISR_offTFURI)
#define SSP_ISR_mskRFORI	( 0x01UL << SSP_ISR_offRFORI)

#define SSP_ACL_offSLOT1V	14	/* The 1st Slot is Valid */
#define SSP_ACL_offSLOT2V	13	/* The 2nd Slot is Valid */
#define SSP_ACL_offSLOT3V	12	/* The 3th Slot is Valid */
#define SSP_ACL_offSLOT4V	11	/* The 4th Slot is Valid */
#define SSP_ACL_offSLOT5V	10	/* The 5th Slot is Valid */
#define SSP_ACL_offSLOT6V	9	/* The 6th Slot is Valid */
#define SSP_ACL_offSLOT7V	8	/* The 7th Slot is Valid */
#define SSP_ACL_offSLOT8V	7	/* The 8th Slot is Valid */
#define SSP_ACL_offSLOT9V	6	/* The 9th Slot is Valid */
#define SSP_ACL_offSLOT10V	5	/* The 10th Slot is Valid */
#define SSP_ACL_offSLOT11V	4	/* The 11th Slot is Valid */
#define SSP_ACL_offSLOT12V	3	/* The 12th Slot is Valid */
#define SSP_ACL_offCODECID	0	/* Codec ID, which will be shifted out as tag slot */

#define SSP_ACL_mskSLOT1V	( 0x01UL << SSP_ACL_offSLOT1V)
#define SSP_ACL_mskSLOT2V	( 0x01UL << SSP_ACL_offSLOT2V)
#define SSP_ACL_mskSLOT3V	( 0x01UL << SSP_ACL_offSLOT3V)
#define SSP_ACL_mskSLOT4V	( 0x01UL << SSP_ACL_offSLOT4V)
#define SSP_ACL_mskSLOT5V	( 0x01UL << SSP_ACL_offSLOT5V)
#define SSP_ACL_mskSLOT6V	( 0x01UL << SSP_ACL_offSLOT6V)
#define SSP_ACL_mskSLOT7V	( 0x01UL << SSP_ACL_offSLOT7V)
#define SSP_ACL_mskSLOT8V	( 0x01UL << SSP_ACL_offSLOT8V)
#define SSP_ACL_mskSLOT9V	( 0x01UL << SSP_ACL_offSLOT9V)
#define SSP_ACL_mskSLOT10V	( 0x01UL << SSP_ACL_offSLOT10V)
#define SSP_ACL_mskSLOT11V	( 0x01UL << SSP_ACL_offSLOT11V)
#define SSP_ACL_mskSLOT12V	( 0x01UL << SSP_ACL_offSLOT12V)
#define SSP_ACL_mskCODECID	( 0x03UL << SSP_ACL_offCODECID)

#define SSP_REV_offMAJOR_REV	16	/* Major Revision Number */
#define SSP_REV_offMINOR_REV	8	/* Minor Revision Number */
#define SSP_REV_offREL_REV	0	/* Release Number */

#define SSP_REV_mskMAJOR_REV	( 0xFFUL << SSP_REV_offMAJOR_REV)
#define SSP_REV_mskMINOR_REV	( 0xFFUL << SSP_REV_offMINOR_REV)
#define SSP_REV_mskREL_REV	( 0xFFUL << SSP_REV_offREL_REV)

#define SSP_FEA_offSSP_FCFG	27	/* The SSP Functional Configurations */
#define SSP_FEA_offSPIMWR_FCFG	26	/* Motorola's SPI and National Semiconductor's Microwire Configurations */
#define SSP_FEA_offI2S_FCFG	25	/* Philips's I2S Functional Configurations */
#define SSP_FEA_offAC97_FCFG	24	/* Intel's AC-Link Functional Configurations */
#define SSP_FEA_offTXFIFO_WIDTH	16	/* Transmit FIFO Size Configurations */
#define SSP_FEA_offRXFIFO_WIDTH	8	/* Recieve FIFO Size Configuration */
#define SSP_FEA_offFIFO_WIDTH	0	/* Transmit/Recieve FIFO Width */

#define SSP_FEA_mskSSP_FCFG	( 0x01UL << SSP_FEA_offSSP_FCFG)
#define SSP_FEA_mskSPIMWR_FCFG	( 0x01UL << SSP_FEA_offSPIMWR_FCFG)
#define SSP_FEA_mskI2S_FCFG	( 0x01UL << SSP_FEA_offI2S_FCFG)
#define SSP_FEA_mskAC97_FCFG	( 0x01UL << SSP_FEA_offAC97_FCFG)
#define SSP_FEA_mskTXFIFO_WIDTH	( 0xFFUL << SSP_FEA_offTXFIFO_WIDTH)
#define SSP_FEA_mskRXFIFO_WIDTH	( 0xFFUL << SSP_FEA_offRXFIFO_WIDTH)
#define SSP_FEA_mskFIFO_WIDTH	( 0xFFUL << SSP_FEA_offFIFO_WIDTH)
#endif /* SSP_FARADAY */
