/* FTSSP010 - UDA1345TS module:
 *
 *	$log$
 *
 *	2006/02/23:	I-Jui Sung:	OSS emulation half-duplex
 *					playback/capture at 48K, 44.1K, 8K
 *					with mono/stereo 16bit/8bit
 *
 *	2006/02/22:	I-Jui Sung:	OSS emulation playback at 44.1KHz
 *                                      16-bit mono completed. Relying ALSA to
 *                                      resample
 *  2009/02/24: dma upgrade checking list:
 *               - ac97 mode playback ................. ok
 *               - ac97 mode capture .................. ok
 *               - i2s mode playback .................. ok
 *               - i2s mode capture ................... ok
 *               - mixer support (snd_ctl_add, ...) ... todo
 *               - debug /proc entry .................. ok
 */
#include <linux/init.h>
#include <linux/module.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <asm/dmad.h>
#include <linux/dma-mapping.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>
#include <sound/asound.h>
#include <linux/i2c.h>
#include <sound/control.h>
#include <linux/interrupt.h>
#include <linux/of.h>
#include <linux/of_device.h>
#include <asm/irq.h>
#include <asm/io.h>
#include "FTSSP010_UDA1345TS.h"

struct alc5630_data;
int init_hw(unsigned int cardno,unsigned int ac97, struct i2c_client *client);

#if (!defined(CONFIG_PLATFORM_AHBDMA) && !defined(CONFIG_PLATFORM_APBDMA))
#warning needs ahb/apb dma to wrok
#endif

/* ---------------------------------------------------------------------------
 * Define the debug level of FTSSP_DEBUG
 */
#define FTSSP_DEBUG     	0
#define FTSSP_DEBUG_VERBOSE	0
#define FTSSP_PROC_FS   	0

#undef VVDBG
#if (FTSSP_DEBUG_VERBOSE)
#define VVDBG(vvar...)	(void)0
//#define VVDBG(vvar...)	printk(KERN_INFO vvar)
#else
#define VVDBG(vvar...)	(void)0
#endif

#undef ERR
#define ERR(vvar...)	printk(KERN_ERR vvar)

#undef INFO
#define INFO(vvar...)	printk(KERN_INFO vvar)

#if (FTSSP_DEBUG)
#undef DBG
#define DBG(vvar...)	printk(KERN_INFO vvar)
#else
#define DBG(vvar...)	(void)0
#endif

#if (FTSSP_DEBUG_VERBOSE)
#undef VDBG
#define VDBG(vvar...)	printk(KERN_INFO vvar)
#else
#define VDBG(vvar...)	(void)0
#endif

#ifdef CONFIG_SND_FTSSP010_I2S
static struct i2c_client *g_i2c_client;
struct alc5630_data {
	struct i2c_client *client;
	struct delayed_work     work;
	unsigned long gpio2_value;
	struct mutex mtx;
};

static int alc5630_i2c_attach(struct i2c_adapter *adapter);
static int alc5630_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id);
static int alc5630_i2c_suspend(struct i2c_client *i2c_client, pm_message_t mesg);
static int alc5630_i2c_resume(struct i2c_client *i2c_client);

static int alc5630_i2c_remove(struct i2c_client *client);
static int ftssp_alsa_init(struct i2c_client *client);

static int i2s_alc5630_read(unsigned int raddr, char *data, struct i2c_client *client)
{
	struct i2c_adapter *adap =  client->adapter;
	struct i2c_msg msg;
	int ret, i2c_value;

	//Reading ALC5630 register
	msg.addr = raddr;
	msg.flags = (client->flags & I2C_M_TEN) | I2C_M_RD;
	msg.len = 1;
	msg.buf = (char *)data;

	//ret = i2c_transfer(adap, &msg, 1);
	#ifndef CONFIG_SND_FTSSP010_AC97
		ret = i2c_transfer(adap, &msg, 1);
	#endif

	if (ret != 0) {
			printk("i2c read failed\n");
			return -1;
    }
    else
    {
	        i2c_value = (data[0]&0xff) << 8 | (data[1]&0xff);
            return i2c_value;	   
	}

}

static void i2s_alc5630_write(unsigned int addr, unsigned int data, struct i2c_client *client)
{

	struct i2c_adapter *adap =  client->adapter;
	struct i2c_msg msg;
	int ret, i2c_value;
	char buf[3];

	//Writing ALC5630 register
	i2c_value = 0x0;
	msg.addr = addr;
	msg.flags = (client->flags & I2C_M_TEN) | ~I2C_M_RD;
	msg.len = 1;

	buf[0] = (data >> 8) & 0xff;
	buf[1] =  data & 0xff;
	msg.buf = (char *)buf;

	//ret = i2c_transfer(adap, &msg, 1);
	#ifndef CONFIG_SND_FTSSP010_AC97
		ret = i2c_transfer(adap, &msg, 1);
	#endif

	if (ret != 0) {
			printk("i2c write failed\n");
    }

}

static void i2s_alc5630_read_test(struct i2c_client *client)
{
	char data[3];
	//printk(">>>>> : i2s_alc5630_read_test().....\n");
	printk("Reg 0x%02x = 0x%08x\n", 0x0, i2s_alc5630_read(0x0, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x02, i2s_alc5630_read(0x02, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x04, i2s_alc5630_read(0x04, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x06, i2s_alc5630_read(0x06, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x08, i2s_alc5630_read(0x08, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x0a, i2s_alc5630_read(0x0a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x0c, i2s_alc5630_read(0x0c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x0e, i2s_alc5630_read(0x0e, data,client));

    printk("Reg 0x%02x = 0x%08x\n", 0x10, i2s_alc5630_read(0x10, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x12, i2s_alc5630_read(0x12, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x14, i2s_alc5630_read(0x14, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x16, i2s_alc5630_read(0x16, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x18, i2s_alc5630_read(0x18, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x1a, i2s_alc5630_read(0x1a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x1c, i2s_alc5630_read(0x1c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x1e, i2s_alc5630_read(0x1e, data,client));

    printk("Reg 0x%02x = 0x%08x\n", 0x20, i2s_alc5630_read(0x20, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x22, i2s_alc5630_read(0x22, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x24, i2s_alc5630_read(0x24, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x26, i2s_alc5630_read(0x26, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x28, i2s_alc5630_read(0x28, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x2a, i2s_alc5630_read(0x2a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x2c, i2s_alc5630_read(0x2c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x2e, i2s_alc5630_read(0x2e, data,client));

    printk("Reg 0x%02x = 0x%08x\n", 0x30, i2s_alc5630_read(0x30, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x32, i2s_alc5630_read(0x32, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x34, i2s_alc5630_read(0x34, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x36, i2s_alc5630_read(0x36, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x38, i2s_alc5630_read(0x38, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x3a, i2s_alc5630_read(0x3a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x3c, i2s_alc5630_read(0x3c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x3e, i2s_alc5630_read(0x3e, data,client));

    printk("Reg 0x%02x = 0x%08x\n", 0x40, i2s_alc5630_read(0x40, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x42, i2s_alc5630_read(0x42, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x44, i2s_alc5630_read(0x44, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x46, i2s_alc5630_read(0x46, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x48, i2s_alc5630_read(0x48, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x4a, i2s_alc5630_read(0x4a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x4c, i2s_alc5630_read(0x4c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x4e, i2s_alc5630_read(0x4e, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x50, i2s_alc5630_read(0x50, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x52, i2s_alc5630_read(0x52, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x54, i2s_alc5630_read(0x54, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x56, i2s_alc5630_read(0x56, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x58, i2s_alc5630_read(0x58, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x5a, i2s_alc5630_read(0x5a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x5c, i2s_alc5630_read(0x5c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x5e, i2s_alc5630_read(0x5e, data,client));

    printk("Reg 0x%02x = 0x%08x\n", 0x60, i2s_alc5630_read(0x60, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x62, i2s_alc5630_read(0x62, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x64, i2s_alc5630_read(0x64, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x66, i2s_alc5630_read(0x66, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x68, i2s_alc5630_read(0x68, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x6a, i2s_alc5630_read(0x6a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x6c, i2s_alc5630_read(0x6c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x6e, i2s_alc5630_read(0x6e, data,client));

    printk("Reg 0x%02x = 0x%08x\n", 0x70, i2s_alc5630_read(0x70, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x72, i2s_alc5630_read(0x72, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x74, i2s_alc5630_read(0x74, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x76, i2s_alc5630_read(0x76, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x78, i2s_alc5630_read(0x78, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x7a, i2s_alc5630_read(0x7a, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x7c, i2s_alc5630_read(0x7c, data,client));
    printk("Reg 0x%02x = 0x%08x\n", 0x7e, i2s_alc5630_read(0x7e, data,client));

}

static int alc5630_i2c_attach(struct i2c_adapter *adapter){

	struct i2c_board_info info;
	struct i2c_client *client;

	int ret=0;

	//i2c_dbg( 1, "alc5630_i2c_attach() is called.\n");
	//printk(">>>>>>>>>> (2) alc5630_i2c_attach() is called.\n");

	//ret = i2c_probe(adapter, &addr_data, alc5630_detect);
	memset(&info, 0, sizeof(struct i2c_board_info));
	strlcpy(info.type, "alc5630_codec", I2C_NAME_SIZE);
	//info.addr = 0x3e;
	info.addr = 0x66;
	//info.platform_data = node;

	//client = i2c_new_device(adapter, &info);
	#ifndef CONFIG_SND_FTSSP010_AC97
		client = i2c_new_device(adapter, &info);
	#endif
	if (!client) {
		printk("$$$$$????? : i2c_new_device() failed.....\n");
		return -ENODEV;
	}
	/*
	 * We know the driver is already loaded, so the device should be
	 * already bound. If not it means binding failed, and then there
	 * is no point in keeping the device instantiated.
	 */
	if (!client->driver) {
		//i2c_unregister_device(client);
		#ifndef CONFIG_SND_FTSSP010_AC97
			i2c_unregister_device(client);
		#endif
		return -ENODEV;
	}

	/*
	 * Let i2c-core delete that device on driver removal.
	 * This is safe because i2c-core holds the core_lock mutex for us.
	 */
	list_add_tail(&client->detected, &client->driver->clients);

	return ret;
}

static irqreturn_t gpio2_irq(int irq, void *dev_id)
{

	struct i2c_client *client = (struct i2c_client *)dev_id;
	struct alc5630_data *alc5630 = i2c_get_clientdata(client);

	unsigned long org_mask, org_dir;
	//printk("@@@@@#####$$$$$!!!!! : gpio2_irq is detected ???....\n");

	//Mask GPIO interrupt
	//REG32(AMIC_VA_BASE + 0x80) = REG32(AMIC_VA_BASE + 0x80) & ~(1 << 13); //GPIO interrupt disable

	//Dump AMIC contents and ir~
	unsigned long ir0,ir1,ir2, ir3, ir4, ir5;
	unsigned long ir6, ir7, ir8, ir9, ir10;
	unsigned long ir11, ir12, ir13, ir14, ir15;
	unsigned long gpio2_value, value;
	char data[3];


	//__asm__ volatile ("setgie.d\n\t");
	//REG32(AMIC_VA_BASE + 0x80) = 0;

	//Get original gpio direction
	org_dir  = REG32(GPIO_VA_BASE + 0x08);

	//Get original gpio interrupt mask
	org_mask = REG32(GPIO_VA_BASE + 0x2c);

	//Mask all gpio interrupt
	REG32(GPIO_VA_BASE + 0x2c) = 0x0000FFFF;
	//Clear all gpio interrupt
	REG32(GPIO_VA_BASE + 0x30) = 0x0000FFFF;

	//Get gpio pin value
	REG32(GPIO_VA_BASE + 0x08) = REG32(GPIO_VA_BASE + 0x08) & ~(0x1UL << 2); //Set gpio2 direction => input
	gpio2_value = (REG32(GPIO_VA_BASE + 0x04) >> 2 & 1);
	//printk("The gpio2 value is 0x%08x\n",gpio2_value);


	alc5630->gpio2_value = gpio2_value;
	schedule_delayed_work(&alc5630->work, 1);

	/*if (gpio3_value==0x1) {
		i2s_alc5630_write(0x02, 0x0000, g_i2c_client);
		i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,g_i2c_client)|0x0040 , g_i2c_client);
		printk("External speaker is plugged in... and Try to mute internal speaker......\n");
    }
	else {
		i2s_alc5630_write(0x02, 0x5F5F, g_i2c_client);
		i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,g_i2c_client)|0x0440 , g_i2c_client);
	    printk("External speaker is pulled ... and Try to unmute internal speaker......\n");
    }*/


	//Set gpio3 pin value to 0
	//REG32(GPIO_VA_BASE + 0x08) = REG32(GPIO_VA_BASE + 0x08) | (0x1UL << 3); //Set gpio3 direction => output
	//value = REG32(GPIO_VA_BASE + 0x00) & ~(0x1UL << 3);
	//REG32(GPIO_VA_BASE + 0x00) = value;

	//Get gpio pin value again
	//REG32(GPIO_VA_BASE + 0x08) = REG32(GPIO_VA_BASE + 0x08) & ~(0x1UL << 3); //Set gpio3 direction => input
	//gpio3_value = (REG32(GPIO_VA_BASE + 0x04) >> 3 & 1);
	//printk("The gpio3 value is 0x%08x\n",gpio3_value);

	//restore original gpio interrupt mask
	REG32(GPIO_VA_BASE + 0x2c) = org_mask;

	//restore original gpio direction
	REG32(GPIO_VA_BASE + 0x08) = org_dir;

	//unmask GPIO interrupt
	//REG32(AMIC_VA_BASE + 0x80) = REG32(AMIC_VA_BASE + 0x80) | (1 << 13); //GPIO interrupt enable
	REG32(AMIC_VA_BASE + 0x84) = REG32(AMIC_VA_BASE + 0x84) | (1 << 13); //Clear GPIO interrupt
	REG32(GPIO_VA_BASE + 0x30) = 0x0000FFFF; //Clear all gpio interrupt

	//printk("@@@@@#####$$$$$!!!!! : gpio2_irq() is exited....\n");

	return IRQ_HANDLED;
}

static void iic_work(struct work_struct *work)
{
	char data[3];
	struct alc5630_data *alc5630 =
		container_of(to_delayed_work(work), struct alc5630_data, work);


	if (alc5630->gpio2_value==0x0) {
		//i2s_alc5630_write(0x02, 0x0000, alc5630->client);
		i2s_alc5630_write(0x02, 0x5F5F, alc5630->client);
		i2s_alc5630_write(0x3A, (i2s_alc5630_read(0x3A, data,alc5630->client) & 0xFBFF) | 0x0040 , alc5630->client);
		//printk("External speaker is plugged in... and Try to mute internal speaker......\n");
		//i2s_alc5630_read_test(alc5630->client);
    }
	else {
		//i2s_alc5630_write(0x02, 0x5F5F, alc5630->client);
		i2s_alc5630_write(0x02, 0x0000, alc5630->client);
		i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,alc5630->client)|0x0440 , alc5630->client);
	    //printk("External speaker is pulled ... and Try to unmute internal speaker......\n");
	    //i2s_alc5630_read_test(alc5630->client);
    }
	//ftsdc_enable_irq(host, false);

	//ftsdc_enable_irq(host, true);
}

static int alc5630_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
{

	struct alc5630_data *alc5630;
	struct i2c_adapter *adap =  client->adapter;
	//unsigned char ret1, ret2;
	struct i2c_msg msg;
	char rbuf[3], wbuf[3];
	int ret, i2c_rvalue, i2c_wvalue;
	unsigned int gpio2_value;
	char data[3];

	//i2c_dbg( 1, "alc5630_i2c_probe() is called.\n");
	//printk(">>>>>>>>>> (3) alc5630_i2c_probe() is called.\n");

	alc5630 = kzalloc(sizeof(struct alc5630_data), GFP_KERNEL);

	if (!alc5630)
		return -ENOMEM;

	mutex_init(&alc5630->mtx);
	client-> flags = 0;
	alc5630->client = client;
	i2c_set_clientdata(client, alc5630);

	//ADD by river 2011.05.18 for GPIO2 interrupt => edge trigger and rising edge
	REG32(GPIO_VA_BASE + 0x20) = 0x00000000;
	REG32(GPIO_VA_BASE + 0x2c) = 0x00000000;
	REG32(GPIO_VA_BASE + 0x2c) = 0x00000000;
	REG32(GPIO_VA_BASE + 0x30) = 0x0000FFFF;
	REG32(GPIO_VA_BASE + 0x38) = 0x0000FFFF;
	REG32(GPIO_VA_BASE + 0x40) = 0x0000FFFF;


	REG32(GPIO_VA_BASE + 0x08) = REG32(GPIO_VA_BASE + 0x08) & ~(1 << 2); //GPIO2 as input
	REG32(GPIO_VA_BASE + 0x34) = REG32(GPIO_VA_BASE + 0x34) & ~(1 << 2); //GPIO2 edge trigger
	//REG32(GPIO_VA_BASE + 0x38) = REG32(GPIO_VA_BASE + 0x38) & ~(1 << 2); //GPIO2 single edge
	REG32(GPIO_VA_BASE + 0x38) = REG32(GPIO_VA_BASE + 0x38) | (1 << 2); //GPIO2 both edge
	REG32(GPIO_VA_BASE + 0x3c) = REG32(GPIO_VA_BASE + 0x3c) & ~(1 << 2); //GPIO2 rising edge
	REG32(GPIO_VA_BASE + 0x20) = REG32(GPIO_VA_BASE + 0x20) | (1 << 2); //GPIO2 pin interrupt enable

    REG32(AMIC_VA_BASE + 0x20) = REG32(AMIC_VA_BASE + 0x20) | (1 << 13); //Interrupt Trigger Mode (edge trigger)
    REG32(AMIC_VA_BASE + 0x24) = REG32(AMIC_VA_BASE + 0x24) & ~(1 << 13); //Interrupt Trigger edge(rising edge) 
    REG32(AMIC_VA_BASE + 0x80) = REG32(AMIC_VA_BASE + 0x80) | (1 << 13); //GPIO interrupt enabled
    //REG32(AMIC_VA_BASE + 0x80) = 0; //GPIO interrupt enabled

	if (request_irq(GPIO_IRQ, gpio2_irq, IRQF_SHARED, "gpio2", client)) {
		printk("Failed to request GPIO2 interrupt.\n");
		goto fail;
	}

	//ADD by river 2011.07.11
	//REG32(GPIO_VA_BASE + 0x08) = REG32(GPIO_VA_BASE + 0x08) & ~(0x1UL << 3); //Set gpio3 direction => input
	gpio2_value = (REG32(GPIO_VA_BASE + 0x04) >> 2 & 1);
	//printk("The gpio2 value is 0x%08x\n",gpio2_value);
	alc5630->gpio2_value = gpio2_value;

	if (gpio2_value==0x0) {
		i2s_alc5630_write(0x02, 0x5F5F, alc5630->client);
		i2s_alc5630_write(0x3A, (i2s_alc5630_read(0x3A, data,alc5630->client) & 0xFBFF) | 0x0040 , alc5630->client);
		//printk("@@@@@ alc5630_i2c_probe : External speaker is plugged in... and Try to mute internal speaker......\n");
		//i2s_alc5630_read_test(alc5630->client);
    }
	else {
		i2s_alc5630_write(0x02, 0x0000, alc5630->client);
		i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,alc5630->client)|0x0440 , alc5630->client);
	    //printk("@@@@@ alc5630_i2c_probe : External speaker is pulled ... and Try to unmute internal speaker......\n");
	    //i2s_alc5630_read_test(alc5630->client);
    }

	//End ADD by river 2011.07.11

	//ADD by river 2011.06.02
	INIT_DELAYED_WORK(&alc5630->work, iic_work);
	//End ADD by river 2011.06.02
	//End ADD by river 2011.05.18

	ftssp_alsa_init(client);

	return 0;
 fail:
	mutex_destroy(&alc5630->mtx);
	kfree(alc5630);
	return -EINVAL;

}

////////End ADD by river 2011.01.26
#endif

/* ---------------------------------------------------------------------------
 * Preserved size of memory space for audio DMA ring
 */
#define FTSSP_HW_DMA_SIZE		(512 * 1024)

/* Buffer sizes reported to ALSA layer - AC97 mode */

/* ring size, exported to application */
#define AC97_HW_BUFFER_BYTES_MAX	(42 * 1024)
/* should not exceed AC97_HW_PERIOD_BYTES_MAX */
#define AC97_HW_PERIOD_BYTES_MIN	(2 * 1024)
/* AC97_HW_PERIOD_BYTES_MAX * AC97_HW_PERIODS_MAX <= AC97_HW_BUFFER_SIZE */
#define AC97_HW_PERIOD_BYTES_MAX	(8 * 1024)
/* 3 <= AC97_HW_PERIODS_MIN <= AC97_HW_PERIODS_MAX */
#define AC97_HW_PERIODS_MIN		3
/* AC97_HW_PERIOD_BYTES_MAX * AC97_HW_PERIODS_MAX <= AC97_HW_BUFFER_SIZE */
#define AC97_HW_PERIODS_MAX		5

/* Driver internal dma buffer size, x2 for S16_LE(16-bits) to AC97 (20-bits),
 * x6 for sampling rate converion from minimum 8k to AC97 48k.
 *
 * Note that AC97 mode cannot do playback and recording simultaneouly. So we
 * use up all FTSSP_HW_DMA_SIZE of memory.
 */
#define AC97_HW_DMA_SIZE		(AC97_HW_BUFFER_BYTES_MAX * 2 * 6)

/* Buffer sizes reported to ALSA layer - I2S mode */

/* ring size, exported to application */
#define I2S_HW_BUFFER_BYTES_MAX		(256 * 1024)
/* should not exceed I2S_HW_PERIOD_BYTES_MAX */
#define I2S_HW_PERIOD_BYTES_MIN		(2 * 1024)
/* I2S_HW_PERIOD_BYTES_MAX * I2S_HW_PERIODS_MAX <= I2S_HW_BUFFER_SIZE */
#define I2S_HW_PERIOD_BYTES_MAX		(32 * 1024)
/* 3 <= I2S_HW_PERIODS_MIN <= I2S_HW_PERIODS_MAX */
#define I2S_HW_PERIODS_MIN		3
/* I2S_HW_PERIOD_BYTES_MAX * I2S_HW_PERIODS_MAX <= I2S_HW_BUFFER_SIZE */
#define I2S_HW_PERIODS_MAX		8

/* Page-in size for playback and capture each.  Note that I2S mode can do
 * playback and recording simultaneouly, so this size should be less than or
 * equal to FTSSP_HW_DMA_SIZE/2
 */
#define I2S_HW_DMA_SIZE			(I2S_HW_BUFFER_BYTES_MAX)

/* ---------------------------------------------------------------------------
 * Audio formats
 */
#define AC97_CODEC_FORMATS		(SNDRV_PCM_FMTBIT_S16_LE)
#define AC97_CODEC_SAMPLE_RATES		(SNDRV_PCM_RATE_48000 | \
					 SNDRV_PCM_RATE_16000 | \
					 SNDRV_PCM_RATE_8000)
#define AC97_CODEC_SAMPLE_RATE_MIN	(8000)
#define AC97_CODEC_SAMPLE_RATE_MAX	(48000)

#define I2S_CODEC_FORMATS		(SNDRV_PCM_FMTBIT_S16_LE)
#define I2S_CODEC_SAMPLE_RATES		(SNDRV_PCM_RATE_48000 | \
					 SNDRV_PCM_RATE_44100 | \
					 SNDRV_PCM_RATE_32000 | \
					 SNDRV_PCM_RATE_22050 | \
					 SNDRV_PCM_RATE_16000 | \
					 SNDRV_PCM_RATE_11025 | \
					 SNDRV_PCM_RATE_8000)
#define I2S_CODEC_SAMPLE_RATE_MIN	(8000)
#define I2S_CODEC_SAMPLE_RATE_MAX	(48000)


/* ---------------------------------------------------------------------------
 * Configuration
 */
#if (CONFIG_PROC_FS == 0)
#undef FTSSP_PROC_FS
#define FTSSP_PROC_FS 0
#else
#if (FTSSP_PROC_FS)
#include <sound/info.h>
#endif  /* FTSSP_PROC_FS */
#endif  /* CONFIG_PROC_FS */

#define FTSSP_CARD_ID		"ftssp010"
#define FTSSP_DRIVER_NAME	"ftssp"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Faraday Technology Corp.");
MODULE_DESCRIPTION("FTSSP010 Linux 2.6 Driver");

static int cardno = 0;
//static const unsigned int SSP_FTSSP010_pa_base[SSP_FTSSP010_IRQ_COUNT] =
//	{ SSP_FTSSP010_PA_BASE };

/* Driver mode */
#ifdef CONFIG_SND_FTSSP010_AC97
static int ac97 = 1;
#else
static int ac97 = 0;
#endif

// ----------------------------------------------
module_param(cardno, int, 0);
MODULE_PARM_DESC(cardno, "FTSSP No.");

module_param(ac97, int, 0);
MODULE_PARM_DESC(ac97, "AC97 mode");
// ----------------------------------------------

/* ---------------------------------------------------------------------------
 * Structures
 */

/* private data for card */
typedef struct {
	struct snd_card *card;
	struct snd_pcm  *pcm;
	struct snd_pcm_substream *substream_tx;
	struct snd_pcm_substream *substream_rx;
#if (FTSSP_PROC_FS)
	struct snd_info_entry *info_buf_max;
	struct snd_info_entry *info_period_min;
	struct snd_info_entry *info_period_max;
	struct snd_info_entry *info_periods_min;
	struct snd_info_entry *info_periods_max;
#endif
} ftssp_chip;

/* dma request descriptors */
dmad_chreq dma_chreq_tx = {
	.channel = -1,
	.drq     = NULL,
};

dmad_chreq dma_chreq_rx = {
	.channel = -1,
	.drq     = NULL,
};

/* Holds ALSA card instance pointers */
struct snd_card *ftssp_cards[SSP_FTSSP010_COUNT];

/* snd_pcm_hardware */
static struct snd_pcm_hardware snd_ftssp_pcm_hw =
{
	.info               = SNDRV_PCM_INFO_INTERLEAVED,
	.formats            = I2S_CODEC_FORMATS,
	.rates              = I2S_CODEC_SAMPLE_RATES,
	.rate_min           = I2S_CODEC_SAMPLE_RATE_MIN,
	.rate_max           = I2S_CODEC_SAMPLE_RATE_MAX,
	.channels_min       = 1,
	.channels_max       = 2,
	.buffer_bytes_max   = I2S_HW_BUFFER_BYTES_MAX,
	.period_bytes_min   = I2S_HW_PERIOD_BYTES_MIN,
	.period_bytes_max   = I2S_HW_PERIOD_BYTES_MAX,
	.periods_min        = I2S_HW_PERIODS_MIN,
	.periods_max        = I2S_HW_PERIODS_MAX,
};

/* private data for a substream (playback or capture) */
/* function pointer for set up AHBDMA for this substream */
typedef void (*start_t)(int cardno, unsigned use_dma);
typedef void (*pmu_set_clocking_t)(unsigned int);
typedef void (*ftssp010_config_t)(int cardno, unsigned is_stereo,
				  unsigned speed, int use8bit);

typedef struct {
	u32                busy;
	spinlock_t         dma_lock;
	unsigned           dma_area_va;
	int                dma_width;
	unsigned int       tx_period;
	unsigned int       rx_period;

	start_t            start;
	pmu_set_clocking_t pmu_set_clocking;
	ftssp010_config_t  hw_config;
} ftssp_substream;

static ftssp_substream ftssp010_substreams[2] = {
	/* Playback substream */
	{
		busy             : 0,
		start            : ftssp010_start_tx,
		pmu_set_clocking : pmu_set_i2s_clocking,
		hw_config        : ftssp010_config_tx,
	},
	/* Capture substream */
	{
		busy             : 0,
		start            : ftssp010_start_rx,
		pmu_set_clocking : pmu_set_i2s_clocking,
		hw_config        : ftssp010_config_rx,
	}
};

/* (AC97 only) Convert 16 bits PCM data in user buffer to/from 20 bits PCM data
 * (32 bits actaully in dma buffer) for AC97 codec.
 */
static int snd_ftssp_playback_copy(struct snd_pcm_substream *substream,
	int channel, snd_pcm_uframes_t pos, void *usr_buf,
	snd_pcm_uframes_t count)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	ftssp_substream *ftssp010_substream =
		(ftssp_substream *) runtime->private_data;


	//printk("~~~~~ : snd_ftssp_playback_copy() is invoked....\n");
	u32 *dma_va = NULL;
	u16 *usr_va = usr_buf;
	int copy_words;
	int pcm_data;

	dmad_chreq *dma_chreq;

	/* frames_to_bytes(runtime, pos + count) * 2(bytes/per pcm) /
	 * 4(bytes per dma unit) */
	u32 sw_ptr = (u32)frames_to_bytes(runtime, pos + count) >> 1;

	switch (runtime->rate) {
	case 8000:
		sw_ptr *= 6;

		dma_va = (unsigned *)(ftssp010_substream->dma_area_va +
				frames_to_bytes(runtime, pos * 6) * 2);

		VVDBG("%s: pos(0x%08x) count(0x%08x) next_pos(0x%08x)\n",
			__func__, (u32)pos, (u32)count, (u32)(pos + count));
		VVDBG("%s: va base(0x%08x) range (0x%08x ~ 0x%08x)\n",
			__func__, (u32)ftssp010_substream->dma_area_va,
			(u32)dma_va,
			(u32)dma_va +
			(u32)2 * frames_to_bytes(runtime, count * 6));

		if (runtime->channels == 1) {
			while (count--) {
				dma_va[0] = (u32)(*usr_va++) << 4;
				dma_va[1] = dma_va[2] = dma_va[3] =
				dma_va[4] = dma_va[5] = dma_va[0];
				//memcpy(&dma_va[1], &dma_va[0], 5 * 4 * 1);
				dma_va += 6;
			}
		} else {  // assume 2 channels
			while (count--) {
				dma_va[0] = (u32)(*usr_va++) << 4;
				dma_va[1] = (u32)(*usr_va++) << 4;
				dma_va[2] = dma_va[4] = dma_va[6] =
				dma_va[8] = dma_va[10] = dma_va[0];
				dma_va[3] = dma_va[5] = dma_va[7] =
				dma_va[9] = dma_va[11] = dma_va[0];
				//memcpy(&dma_va[2], &dma_va[0], 5 * 4 * 2);
				dma_va += 12;
			}
		}
		break;
	case 16000:
		sw_ptr *= 3;

		dma_va = (unsigned *)(ftssp010_substream->dma_area_va +
				frames_to_bytes(runtime, pos * 3) * 2);

		VVDBG("%s: pos(0x%08x) count(0x%08x) next_pos(0x%08x)\n",
			__func__, (u32)pos, (u32)count, (u32)(pos + count));
		VVDBG("%s: va base(0x%08x) range (0x%08x ~ 0x%08x)\n",
			__func__, (u32)ftssp010_substream->dma_area_va,
			(u32)dma_va,
			(u32)dma_va +
			(u32)2 * frames_to_bytes(runtime, count * 3));

		if (runtime->channels == 1) {
			while (count--) {
				dma_va[0] = (u32)(*usr_va++) << 4;
				dma_va[1] = dma_va[2] = dma_va[0];
				//memcpy(&dma_va[1], &dma_va[0], 2 * 4 * 1);
				dma_va += 3;
			}
		} else {  // assume 2 channels
			while (count--) {
				dma_va[0] = (u32)(*usr_va++) << 4;
				dma_va[1] = (u32)(*usr_va++) << 4;
				dma_va[2] = dma_va[4] = dma_va[0];
				dma_va[3] = dma_va[5] = dma_va[1];
				//memcpy(&dma_va[2], &dma_va[0], 2 * 4 * 2);
				dma_va += 6;
			}
		}
		break;
	case 48000:
	default:
		dma_va = (unsigned *)(ftssp010_substream->dma_area_va +
				frames_to_bytes(runtime, pos) * 2);
		copy_words = 2 * frames_to_bytes(runtime, count) / sizeof(u32);

		VVDBG("%s: pos(0x%08x) count(0x%08x) next_pos(0x%08x)\n",
			__func__, (u32)pos, (u32)count, (u32)(pos + count));
		VVDBG("%s: va base(0x%08x) range (0x%08x ~ 0x%08x)\n",
			__func__, (u32)ftssp010_substream->dma_area_va,
			(u32)dma_va,
			(u32)dma_va + (u32)copy_words*4);

		while (copy_words--) {
			pcm_data = (*usr_va++);
			*dma_va++= pcm_data << 4;
		}
		break;
	}

	dma_chreq = &dma_chreq_tx;

	if (dmad_update_ring_sw_ptr(dma_chreq, sw_ptr,
		(runtime->status->state == SNDRV_PCM_STATE_RUNNING) ? 1:0) != 0)
	{
		ERR("%s: failed to update sw-pointer!\n", __func__);
		return -ENODEV;
	}

	return 0;
}

static int snd_ftssp_capture_copy(struct snd_pcm_substream *substream,
	int channel, snd_pcm_uframes_t pos, void *usr_buf,
	snd_pcm_uframes_t count)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	ftssp_substream *ftssp010_substream =
		(ftssp_substream *) runtime->private_data;

	//printk("~~~~~ : snd_ftssp_capture_copy() is invoked....\n");
	//printk(">>>>>>>>>> : snd_ftssp_capture_copy() for recording....\n");
	u32 *dma_va = NULL;
	u16 *usr_va = usr_buf;

	switch (runtime->rate) {
	case 8000:
		dma_va = (unsigned *)(ftssp010_substream->dma_area_va +
				 frames_to_bytes(runtime, pos * 6) * 2);

		VVDBG("%s: pos(0x%08x) count(0x%08x) next_pos(0x%08x)\n",
			__func__, (u32)pos, (u32)count, (u32)(pos + count));
		VVDBG("%s: va base(0x%08x) range (0x%08x ~ 0x%08x)\n",
			__func__, (u32)ftssp010_substream->dma_area_va,
			(u32)dma_va,
			(u32)dma_va +
			(u32)2 * frames_to_bytes(runtime, count * 6));

		if (runtime->channels == 1) {
			while (count--) {
				*usr_va++ = (u16)(dma_va[0] >> 4);
				dma_va += 6;
			}
		} else {
			while (count--) {
				usr_va[0] = (u16)(dma_va[0] >> 4);

				/* [hw-limit] only slot-3 has valid data in
				 *   recording mode -- check TAG_DATA_MONO
				 *   defined in "FTSSP010_lib.c".  Mask out
				 *   one channel to avoid hi-freq noise.
				 */
				usr_va[1] = usr_va[0];
				usr_va += 2;
				dma_va += 12;
			}
		}
		break;
	case 16000:
		dma_va = (unsigned *)(ftssp010_substream->dma_area_va +
				 frames_to_bytes(runtime, pos * 3) * 2);

		VVDBG("%s: pos(0x%08x) count(0x%08x) next_pos(0x%08x)\n",
			__func__, (u32)pos, (u32)count, (u32)(pos + count));
		VVDBG("%s: va base(0x%08x) range (0x%08x ~ 0x%08x)\n",
			__func__, (u32)ftssp010_substream->dma_area_va,
			(u32)dma_va,
			(u32)dma_va +
			(u32)2 * frames_to_bytes(runtime, count * 3));

		if (runtime->channels == 1) {
			while (count--) {
				*usr_va++ = (u16)(dma_va[0] >> 4);
				dma_va += 3;
			}
		} else {
			while (count--) {
				usr_va[0] = (u16)(dma_va[0] >> 4);

				/* [hw-limit] only slot-3 has valid data in
				 *   recording mode -- check TAG_DATA_MONO
				 *   defined in "FTSSP010_lib.c".  Mask out
				 *   one channel to avoid hi-freq noise.
				 */
				usr_va[1] = usr_va[0];
				usr_va += 2;
				dma_va += 6;
			}
		}
		break;
	case 48000:
	default:
		dma_va = (unsigned *)(ftssp010_substream->dma_area_va +
				frames_to_bytes(runtime, pos) * 2);

		VVDBG("%s: pos(0x%08x) count(0x%08x) next_pos(0x%08x)\n",
			__func__, (u32)pos, (u32)count, (u32)(pos + count));
		VVDBG("%s: va base(0x%08x) range (0x%08x ~ 0x%08x)\n",
			__func__, (u32)ftssp010_substream->dma_area_va,
			(u32)dma_va,
			(u32)dma_va +
			(u32)2 * frames_to_bytes(runtime, count));

		if (runtime->channels == 1) {
			while (count--) {
				*usr_va++ = (u16)(*dma_va++ >> 4);
			}
		} else {
			while (count--) {
				usr_va[0] = (u16)(dma_va[0] >> 4);

				/* [hw-limit] only slot-3 has valid data in
				 *   recording mode -- check TAG_DATA_MONO
				 *   defined in "FTSSP010_lib.c".  Mask out
				 *   one channel to avoid hi-freq noise.
				 */
				usr_va[1] = usr_va[0];
				usr_va += 2;
				dma_va += 2;
			}
		}
		break;
	}

	return 0;
}

/**
 * These dma callbacks are called in interrupt context.
 * @data: pointer to the chip-wide structure.
 *        TODO: use stream-specifc data
 */
__attribute__((__unused__))
static void ftssp_dma_callback_tx(int ch, u16 int_status, void *data)
{
	ftssp_chip *chip = (ftssp_chip *)data;

	//printk("~~~~~ : ftssp_dma_callback_tx() is invoked....\n");
	if (!ac97) {
		/* in i2s mode, no indication to driver for user data length.
		 * For simplicity, just go ahead by one period */

		struct snd_pcm_runtime *runtime = chip->substream_tx->runtime;
		ftssp_substream *ftssp010_substream =
			(ftssp_substream *)runtime->private_data;
		u32 sw_ptr;
		u32 tx_period = ftssp010_substream->tx_period + 1;

		if (tx_period == runtime->periods)
			sw_ptr = runtime->buffer_size;
		else
			sw_ptr = tx_period * runtime->period_size;

		sw_ptr = (u32)frames_to_bytes(runtime, sw_ptr) >> 1;

		if (dmad_update_ring_sw_ptr(&dma_chreq_tx, (u32)sw_ptr, 0)) {
			ERR("%s: failed to update sw-pointer!\n", __func__);
		}

		ftssp010_substream->tx_period = tx_period % runtime->periods;
	}

	snd_pcm_period_elapsed(chip->substream_tx);
}

__attribute__((__unused__))
static void ftssp_dma_callback_rx(int ch, u16 int_status, void *data)
{
	ftssp_chip *chip = (ftssp_chip *)data;
	struct snd_pcm_runtime *runtime = chip->substream_rx->runtime;
	ftssp_substream *ftssp010_substream =
		(ftssp_substream *)runtime->private_data;

	//printk("~~~~~ : ftssp_dma_callback_rx() is invoked....\n");
	//printk(">>>>>>>>>> : ftssp_dma_callback_rx() for recording....\n");

	u32 sw_ptr;
	u32 rx_period = ftssp010_substream->rx_period + 1;

	if (rx_period == runtime->periods)
		sw_ptr = runtime->buffer_size;
	else
		sw_ptr = rx_period * runtime->period_size;

	if (ac97) {
		switch (runtime->rate) {
		case 8000:
			sw_ptr = sw_ptr * 6;
			break;
		case 16000:
			sw_ptr = sw_ptr * 3;
			break;
		case 48000:
		default:
			break;
		}
	}
	sw_ptr = (u32)frames_to_bytes(runtime, sw_ptr) >> 1;

	if (dmad_update_ring_sw_ptr(&dma_chreq_rx, (u32)sw_ptr, 0) != 0) {
		ERR("%s: failed to update sw-pointer!\n", __func__);
	}

	ftssp010_substream->rx_period = rx_period % runtime->periods;

	snd_pcm_period_elapsed(chip->substream_rx);
}

static inline int snd_ftssp_dma_ch_alloc(struct snd_pcm_substream *substream)
{
	dmad_chreq *ch_req __attribute__((__unused__)) = 0;

	//printk("~~~~~ WATCH : snd_ftssp_dma_ch_alloc() is invoked....\n");
#ifdef CONFIG_PLATFORM_APBDMA

	if (substream->pstr->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		ch_req = &dma_chreq_tx;
		ch_req->completion_cb    = ftssp_dma_callback_tx;
		ch_req->apb_req.tx_dir   = DMAD_DIR_A0_TO_A1;
		/*for amerald ac97 ssp2 */
#if defined(CONFIG_PLAT_AG101P)
		if((inl(PMU_BASE) & AMERALD_MASK) == AMERALD_PRODUCT_ID)
		{
			ch_req->apb_req.dev_reqn = APBBR_REQN_I2SAC97TX_AMERALD;
		}
		else
#endif
			ch_req->apb_req.dev_reqn = APBBR_REQN_I2SAC97TX;
	} else {
		ch_req = &dma_chreq_rx;
		ch_req->completion_cb    = ftssp_dma_callback_rx;
		ch_req->apb_req.tx_dir   = DMAD_DIR_A1_TO_A0;
		/*for amerald ac97 ssp2 */
#if defined(CONFIG_PLAT_AG101P)
		if((inl(PMU_BASE) & AMERALD_MASK) == AMERALD_PRODUCT_ID)
		{
			ch_req->apb_req.dev_reqn = APBBR_REQN_I2SAC97RX_AMERALD;
		}
		else
#endif
			ch_req->apb_req.dev_reqn = APBBR_REQN_I2SAC97RX;
	}

	ch_req->controller         = DMAD_DMAC_APB_CORE;
	ch_req->flags              = DMAD_FLAGS_RING_MODE;
	ch_req->ring_base          = 0;
	ch_req->dev_addr           = (dma_addr_t)FTSSP010_DATA_PA(cardno);
	ch_req->periods            = 0;
	ch_req->period_size        = 0;

	if (ac97) {
		ch_req->apb_req.ring_ctrl  = APBBR_ADDRINC_I4X;
		ch_req->apb_req.ring_reqn  = APBBR_REQN_NONE;
		ch_req->apb_req.dev_ctrl   = APBBR_ADDRINC_FIXED;
		ch_req->apb_req.burst_mode = 0;
		ch_req->apb_req.data_width = APBBR_DATAWIDTH_4;
	} else {
		ch_req->apb_req.ring_ctrl  = APBBR_ADDRINC_I2X;
		ch_req->apb_req.ring_reqn  = APBBR_REQN_NONE;
		ch_req->apb_req.dev_ctrl   = APBBR_ADDRINC_FIXED;
		ch_req->apb_req.burst_mode = 0;
		ch_req->apb_req.data_width = APBBR_DATAWIDTH_2;
	}

	ch_req->completion_data    = (void *)snd_pcm_substream_chip(substream);

	if (dmad_channel_alloc(ch_req) != 0) {
		ERR("%s: APBDMA channel allocation failed\n",__func__);
		goto _try_ahb;
	}

	DBG("%s: APBDMA channel allocated (ch: %d) ring_mode\n",
		__func__, ch_req->channel);

	return 0;

_try_ahb:

#endif /* CONFIG_PLATFORM_APBDMA */

#ifdef CONFIG_PLATFORM_AHBDMA
	if (substream->pstr->stream == SNDRV_PCM_STREAM_PLAYBACK) {
		ch_req = &dma_chreq_tx;
		ch_req->completion_cb    = ftssp_dma_callback_tx;
		ch_req->ahb_req.tx_dir   = DMAD_DIR_A0_TO_A1;
#if defined(CONFIG_PLAT_AG101P)
		if((inl(PMU_BASE) & AMERALD_MASK) == AMERALD_PRODUCT_ID)
		{
			ch_req->ahb_req.dev_reqn = DMAC_REQN_I2SAC97TX_AMERALD;
		}
		else
#endif
			ch_req->ahb_req.dev_reqn = DMAC_REQN_I2SAC97TX;
	} else {
		ch_req = &dma_chreq_rx;
		ch_req->completion_cb    = ftssp_dma_callback_rx;
		ch_req->ahb_req.tx_dir   = DMAD_DIR_A1_TO_A0;
#if defined(CONFIG_PLAT_AG101P)
		if((inl(PMU_BASE) & AMERALD_MASK) == AMERALD_PRODUCT_ID)
		{
			ch_req->ahb_req.dev_reqn = DMAC_REQN_I2SAC97RX_AMERALD;
		}
		else
#endif
			ch_req->ahb_req.dev_reqn = DMAC_REQN_I2SAC97RX;
	}

	ch_req->controller           = DMAD_DMAC_AHB_CORE;
	ch_req->flags                = DMAD_FLAGS_RING_MODE;
	ch_req->ring_base            = 0;
	ch_req->dev_addr             = (dma_addr_t)FTSSP010_DATA_PA(cardno);
	ch_req->periods              = 0;
	ch_req->period_size          = 0;

	ch_req->ahb_req.sync         = 1;
	ch_req->ahb_req.priority     = DMAC_CSR_CHPRI_2;
	ch_req->ahb_req.hw_handshake = 1;
	ch_req->ahb_req.burst_size   = DMAC_CSR_SIZE_1;

	if (ac97) {
		ch_req->ahb_req.ring_width   = DMAC_CSR_WIDTH_32;
		ch_req->ahb_req.ring_ctrl    = DMAC_CSR_AD_INC;
		ch_req->ahb_req.ring_reqn    = DMAC_REQN_NONE;
		ch_req->ahb_req.dev_width    = DMAC_CSR_WIDTH_32;
		ch_req->ahb_req.dev_ctrl     = DMAC_CSR_AD_FIX;
	} else {
		ch_req->ahb_req.ring_width   = DMAC_CSR_WIDTH_16;
		ch_req->ahb_req.ring_ctrl    = DMAC_CSR_AD_INC;
		ch_req->ahb_req.ring_reqn    = DMAC_REQN_NONE;
		ch_req->ahb_req.dev_width    = DMAC_CSR_WIDTH_16;
		ch_req->ahb_req.dev_ctrl     = DMAC_CSR_AD_FIX;
	}

	ch_req->completion_data = (void *)snd_pcm_substream_chip(substream);

	if (dmad_channel_alloc(ch_req) != 0) {
		ERR("%s: AHBDMA channel allocation failed\n", __func__);
		goto _err_exit;
	}

	DBG("%s: AHBDMA channel allocated (ch: %d) ring_mode\n",
		__func__, ch_req->channel);

	return 0;

_err_exit:

#endif /* CONFIG_PLATFORM_AHBDMA */

	return -ENODEV;
}

static inline ftssp_substream *ftssp010_substream_new(int stream_id)
{
	ftssp_substream *s = NULL;

	//printk("~~~~~ : ftssp010_substream_new() is invoked....\n");

	switch (stream_id) {
	case SNDRV_PCM_STREAM_PLAYBACK:
		s = &ftssp010_substreams[0];
		break;
	case SNDRV_PCM_STREAM_CAPTURE:
		s = &ftssp010_substreams[1];
		break;
	default:
		ERR("%s: wrong stream type (%d)\n", __func__, stream_id);
		return NULL;
	}

	if (s->busy) {
		ERR("%s: device busy!\n", __func__);
		return NULL;
	}
	s->busy = 1;

	spin_lock_init(&s->dma_lock);

	return s;
}

static int snd_ftssp_pcm_open(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	int stream_id = substream->pstr->stream;

	VDBG("%s, %s\n", __func__,
		(substream->pstr->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
		"playback" : "capture");

	//printk("~~~~~ : snd_ftssp_pcm_open() is invoked....\n");

	/* Both playback and capture share a hardware description */
	runtime->hw = snd_ftssp_pcm_hw;

	/* Allocate & Initialize stream-specific data */
	runtime->private_data = ftssp010_substream_new(stream_id);

	if (runtime->private_data) {
		//printk("~~~~~ YAYAYA @@@@@ : Calling snd_ftssp_dma_ch_alloc().\n");
		return snd_ftssp_dma_ch_alloc(substream);
    }
	else
		return -EBUSY;
}

static int snd_ftssp_pcm_close(struct snd_pcm_substream *substream)
{
	int stream_id = substream->pstr->stream;
	ftssp_substream *ftssp010_substream =
		(ftssp_substream *)substream->runtime->private_data;

	VDBG("%s, %s\n", __func__,
		(substream->pstr->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
		"playback" : "capture");

    //printk("~~~~~ : snd_ftssp_pcm_close() is invoked....\n");

	if (stream_id == SNDRV_PCM_STREAM_PLAYBACK)
		dmad_channel_free(&dma_chreq_tx);
	else
		dmad_channel_free(&dma_chreq_rx);

	ftssp010_substream->busy = 0;
	return 0;
}

static int snd_ftssp_pcm_hw_params(struct snd_pcm_substream *substream,
	struct snd_pcm_hw_params *hw_params)
{
	VDBG("%s, %s\n", __func__,
		(substream->pstr->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
		"playback" : "capture");

	//printk("~~~~~ : snd_ftssp_pcm_hw_params() is invoked....\n");

	if (ac97)
		return snd_pcm_lib_malloc_pages(substream, AC97_HW_DMA_SIZE);
	else
		return snd_pcm_lib_malloc_pages(substream, I2S_HW_DMA_SIZE);
}

static int snd_ftssp_pcm_hw_free(struct snd_pcm_substream *substream)
{
	VDBG("%s, %s\n", __func__,
		(substream->pstr->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
		"playback" : "capture");
	//printk("~~~~~ : snd_ftssp_pcm_hw_free() is invoked....\n");

	if (substream->pstr->stream == SNDRV_PCM_STREAM_PLAYBACK)
		dmad_drain_requests(&dma_chreq_tx, 1);
	else
		dmad_drain_requests(&dma_chreq_rx, 1);

	return snd_pcm_lib_free_pages(substream);
}

/* Prepare FTSSP010 AHBDMA for playback & capture */
static int snd_ftssp_pcm_prepare(struct snd_pcm_substream *substream)
{
	ftssp_chip *chip = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;

	//printk("~~~~~ : snd_ftssp_pcm_prepare() is invoked....\n");
	//printk("@@@@@ >>>>> : snd_ftssp_pcm_prepare() is called.\n");

	ftssp_substream *ftssp010_substream =
		(ftssp_substream *)runtime->private_data;

	int stream_id = substream->pstr->stream;
	dmad_chreq *dma_chreq;
	unsigned period_size, buffer_size;

	if (stream_id == SNDRV_PCM_STREAM_PLAYBACK)
		dma_chreq = &dma_chreq_tx;
	else
		dma_chreq = &dma_chreq_rx;

	period_size = frames_to_bytes(runtime, runtime->period_size);
	buffer_size = frames_to_bytes(runtime, runtime->buffer_size);

	if (runtime->format != SNDRV_PCM_FORMAT_S16_LE)
		return -ENODEV;

	if (ac97) {
		switch (runtime->rate) {
		case 8000:
			period_size *= 12;
			buffer_size *= 12;
			break;
		case 16000:
			period_size *= 6;
			buffer_size *= 6;
			break;
		case 48000:
		default:
			period_size *= 2;
			buffer_size *= 2;
			break;
		}

		ftssp010_substream->dma_width = 4;
	} else {
		ftssp010_substream->dma_width = 2;
	}

	dmad_drain_requests(dma_chreq, 1);

	dma_chreq->ring_base   = (dma_addr_t)runtime->dma_addr;
	dma_chreq->periods     = (dma_addr_t)runtime->periods;
	if (ac97) {
		dma_chreq->period_size = (dma_addr_t)(period_size >> 2);
		dma_chreq->ring_size   = (dma_addr_t)(buffer_size >> 2);
	} else {
		dma_chreq->period_size = (dma_addr_t)(period_size >> 1);
		dma_chreq->ring_size   = (dma_addr_t)(buffer_size >> 1);
	}
	dmad_update_ring(dma_chreq);

	/* Set PMU, FTSSP010, and DMA */
	spin_lock(&ftssp010_substream->dma_lock);

	/* keep DMA buffer VA for copy() callback */
	// todo: support playback/capture simultaneously
	ftssp010_substream->dma_area_va = (u32)runtime->dma_area;

	if (ac97) {
		ftssp010_substream->pmu_set_clocking(48000);
		ftssp010_substream->hw_config(cardno,
			runtime->channels > 1 ? 1 : 0, /* 1: stereo, 0: mono */
			48000, ftssp010_substream->dma_width);
	} else {
		ftssp010_substream->pmu_set_clocking(runtime->rate);
		ftssp010_substream->hw_config(cardno,
			runtime->channels > 1 ? 1 : 0, /* 1: stereo, 0: mono */
			runtime->rate, ftssp010_substream->dma_width);
	}

	if (stream_id == SNDRV_PCM_STREAM_PLAYBACK) {
		ftssp010_substream->tx_period = 0;
		chip->substream_tx = substream;
	} else {
		ftssp010_substream->rx_period = 0;
		chip->substream_rx = substream;
	}

	spin_unlock(&ftssp010_substream->dma_lock);

	VVDBG("%s <<\n", __func__);

	return 0;
}

static inline int snd_ftssp_start_play(ftssp_substream *ftssp010_substream,
	struct snd_pcm_runtime *runtime)
{
	int err = 0;

	//printk("~~~~~ : snd_ftssp_start_play() is invoked....\n");

	if (ac97) {
		/* in ac97 mode, user data was fed to dma buffer through
		 * driver-provided copy callback */
		err = dmad_kickoff_requests(&dma_chreq_tx);
		if (err != 0) {
			ERR("%s: failed to kickoff dma!\n", __func__);
			return err;
		}
	} else {
		/* in i2s mode, no indication to driver for user data length
		 * (except start threshold). For simplicity at start, just go
		 * ahead by one cycle */

		u32 sw_ptr =
			(u32)frames_to_bytes(runtime, runtime->buffer_size) >>1;

		err = dmad_update_ring_sw_ptr(&dma_chreq_tx, sw_ptr, 0);
		if (err != 0) {
			ERR("%s: failed to update sw-pointer!\n", __func__);
			return err;
		}

		err = dmad_kickoff_requests(&dma_chreq_tx);
		if (err != 0) {
			ERR("%s: failed to kickoff dma!\n", __func__);
			return err;
		}
	}

	ftssp010_substream->start(cardno, 1);

	//ADD by river 2011.03.08
	//i2s_alc5630_write(0x02, 0x0000, g_i2c_client);
	//i2s_alc5630_write(0x04, 0x0000, g_i2c_client);
	//i2s_alc5630_write(0x0c, 0x1010, g_i2c_client);
	//i2s_alc5630_write(0x10, 0xff03, g_i2c_client);
	//End ADD by river 2011.03.08

	return 0;
}

static inline int snd_ftssp_start_record(ftssp_substream *ftssp010_substream,
	struct snd_pcm_runtime *runtime)
{
	int err = 0;

	//printk("~~~~~ : snd_ftssp_start_record() is invoked....\n");

	u32 sw_ptr = (u32)frames_to_bytes(runtime, runtime->buffer_size);

	if (ac97) {
		switch (runtime->rate) {
		case 8000:
			sw_ptr = (sw_ptr * 3);
			break;
		case 16000:
			sw_ptr = (sw_ptr * 3) >> 1;
			break;
		case 48000:
		default:
			sw_ptr = sw_ptr >> 1;
			break;
		}
	} else {

		//printk(">>>>>>>>>> : snd_ftssp_start_record() for recording....\n");
		sw_ptr = sw_ptr >> 1;
	}

	err = dmad_update_ring_sw_ptr(&dma_chreq_rx, sw_ptr, 0);
	if (err != 0) {
		ERR("%s: failed to update sw-pointer!\n", __func__);
		return err;
	}

	err = dmad_kickoff_requests(&dma_chreq_rx);
	if (err != 0) {
		ERR("%s: failed to kickoff dma!\n", __func__);
		return err;
	}

	ftssp010_substream->start(cardno, 1);

	return 0;
}

/* Triggers AHBDMA for playback & capture */
static int snd_ftssp_pcm_trigger(struct snd_pcm_substream * substream, int cmd)
{
	ftssp_substream *ftssp010_substream =
		(ftssp_substream *)substream->runtime->private_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	int err = 0;
	int stream_id = substream->pstr->stream;

	//printk("~~~~~ : snd_ftssp_pcm_trigger() is invoked....\n");

	/* note local interrupts are already disabled in the midlevel code */
	spin_lock(&ftssp010_substream->dma_lock);

	switch (cmd) {
	case SNDRV_PCM_TRIGGER_START:

		VDBG("%s: SNDRV_PCM_TRIGGER_START state(0x%08x)\n",
			__func__, (u32)runtime->status->state);

		if (stream_id == SNDRV_PCM_STREAM_PLAYBACK) {
			err = snd_ftssp_start_play(ftssp010_substream, runtime);
		} else {
			err = snd_ftssp_start_record(ftssp010_substream,
				runtime);
		}
		break;

	case SNDRV_PCM_TRIGGER_STOP:

		VDBG("%s: SNDRV_PCM_TRIGGER_STOP state(0x%08x)\n",
			__func__, (u32)substream->runtime->status->state);

		if (stream_id == SNDRV_PCM_STREAM_PLAYBACK) {
			ftssp010_stop_tx(cardno);
			dmad_drain_requests(&dma_chreq_tx, 1);
		} else {
			ftssp010_stop_rx(cardno);
			dmad_drain_requests(&dma_chreq_rx, 1);
		}
		break;
	default:
		err = -EINVAL;
		break;
	}

	spin_unlock(&ftssp010_substream->dma_lock);
	return err;
}

// pcm middle-layer call this function within irq (snd_pcm_period_elapsed) or
// with local irq disabled (snd_pcm_lib_write1)
static snd_pcm_uframes_t snd_ftssp_pcm_pointer(
	struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	u32 hw_ptr;
	snd_pcm_uframes_t ret;
	int stream_id = substream->pstr->stream;

	//printk("~~~~~ : snd_ftssp_pcm_pointer() is invoked....\n");


	/* Fetch DMA pointer, with spin lock */
	//spin_lock_irqsave(&ftssp010_substream->dma_lock, flags);

	if (stream_id == SNDRV_PCM_STREAM_PLAYBACK) {
		hw_ptr = dmad_probe_ring_hw_ptr(&dma_chreq_tx);
	} else {
		hw_ptr = dmad_probe_ring_hw_ptr(&dma_chreq_rx);
	}

	//spin_unlock_irqrestore(&ftssp010_substream->dma_lock, flags);

	if (ac97) {
		ret = bytes_to_frames(runtime, hw_ptr << 1);

		switch (runtime->rate) {
		case 8000:
			ret = ret / 6;
			break;
		case 16000:
			ret = ret / 3;
			break;
		case 48000:
		default:
			break;
		}
	} else {
		ret = bytes_to_frames(runtime, hw_ptr << 1);
	}


	VVDBG("%s: hw_ptr(0x%08x) ret(0x%08x)\n",
		(stream_id == SNDRV_PCM_STREAM_PLAYBACK) ? "p" : "c",
		(u32)hw_ptr, (u32)ret);

	/* ALSA requires return value 0 <= ret < buffer_size */
	if (ret >= runtime->buffer_size)
		return 0;
	return ret;
}

/* For FTSSP010 driver, operations are shared among playback & capture */
static struct snd_pcm_ops snd_ftssp_playback_ops = {
	.open      = snd_ftssp_pcm_open,
	.close     = snd_ftssp_pcm_close,
	.ioctl     = snd_pcm_lib_ioctl,
	.hw_params = snd_ftssp_pcm_hw_params,
	.hw_free   = snd_ftssp_pcm_hw_free,
	.prepare   = snd_ftssp_pcm_prepare,
	.trigger   = snd_ftssp_pcm_trigger,
	.pointer   = snd_ftssp_pcm_pointer,
	.copy      = NULL,
};

static struct snd_pcm_ops snd_ftssp_capture_ops = {
	.open      = snd_ftssp_pcm_open,
	.close     = snd_ftssp_pcm_close,
	.ioctl     = snd_pcm_lib_ioctl,
	.hw_params = snd_ftssp_pcm_hw_params,
	.hw_free   = snd_ftssp_pcm_hw_free,
	.prepare   = snd_ftssp_pcm_prepare,
	.trigger   = snd_ftssp_pcm_trigger,
	.pointer   = snd_ftssp_pcm_pointer,
	.copy      = NULL,
};

/* ALSA PCM constructor */
static int snd_ftssp_new_pcm(ftssp_chip *chip)
{
	struct snd_pcm *pcm;
	int err;

	//printk("~~~~~ : snd_ftssp_new_pcm() is invoked....\n");

	/* PCM device #0 with 1 playback and 1 capture */
	if ((err = snd_pcm_new(chip->card, "ftssp_pcm", 0, 1, 1, &pcm)) < 0)
		return err;

	pcm->private_data = chip;
	strcpy(pcm->name, "ftssp_pcm device");
	chip->pcm = pcm;

	/* set operators for playback and capture*/
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK,
		&snd_ftssp_playback_ops);

	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE,
		&snd_ftssp_capture_ops);

	/* Pre-allocate buffer, as suggested by ALSA driver document */
	// todo: support playback/capture simultaneously
	snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV,
	NULL, FTSSP_HW_DMA_SIZE, FTSSP_HW_DMA_SIZE);

	/* Force half-duplex (on A320D, or AC97 mode) */
	if (ac97)
		pcm->info_flags |= SNDRV_PCM_INFO_HALF_DUPLEX;

	return 0;
}

#if (FTSSP_PROC_FS)
static void snd_ftssp_buf_max_read(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	snd_iprintf(buffer, "%d\n", snd_ftssp_pcm_hw.buffer_bytes_max);
}

static void snd_ftssp_buf_max_write(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	char tmp[128];
	char *ptr_e;
	u32 val;

	if (buffer->size == 0)
		return;

	memset(tmp, 0, 128);
	snd_info_get_str(tmp, buffer->buffer, 127);

	val = simple_strtoul(tmp, &ptr_e, 10);
	if (*ptr_e == 'k')
		val *= 1024;
	else if (*ptr_e == 'm')
		val *= 1024 * 1024;

	if (ac97) {
		if (val > AC97_HW_BUFFER_BYTES_MAX)
			val = AC97_HW_BUFFER_BYTES_MAX;
	} else {
		if (val > I2S_HW_BUFFER_BYTES_MAX)
			val = I2S_HW_BUFFER_BYTES_MAX;
	}

	snd_ftssp_pcm_hw.buffer_bytes_max = (size_t)val;
}

static void snd_ftssp_period_min_read(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	snd_iprintf(buffer, "%d\n", snd_ftssp_pcm_hw.period_bytes_min);
}

static void snd_ftssp_period_min_write(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	char tmp[128];
	char *ptr_e;
	u32 val;

	if (buffer->size == 0)
		return;

	memset(tmp, 0, 128);
	snd_info_get_str(tmp, buffer->buffer, 127);

	val = simple_strtoul(tmp, &ptr_e, 10);
	if (*ptr_e == 'k')
		val *= 1024;
	else if (*ptr_e == 'm')
		val *= 1024 * 1024;

	snd_ftssp_pcm_hw.period_bytes_min = (size_t)val;

	if ((val * snd_ftssp_pcm_hw.periods_max) >
	    snd_ftssp_pcm_hw.buffer_bytes_max) {
		INFO("\nWarning: period_bytes(%d) * periods(%d) exceeds "
			"hw_buffer_size(%d).\n",
			snd_ftssp_pcm_hw.period_bytes_min,
			snd_ftssp_pcm_hw.periods_max,
			snd_ftssp_pcm_hw.buffer_bytes_max);
		INFO("         Unexpected access violation may occur!\n");
	}
}

static void snd_ftssp_period_max_read(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	snd_iprintf(buffer, "%d\n", snd_ftssp_pcm_hw.period_bytes_max);
}

static void snd_ftssp_period_max_write(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	char tmp[128];
	char *ptr_e;
	u32 val;

	if (buffer->size == 0)
		return;

	memset(tmp, 0, 128);
	snd_info_get_str(tmp, buffer->buffer, 127);

	val = simple_strtoul(tmp, &ptr_e, 10);
	if (*ptr_e == 'k')
		val *= 1024;
	else if (*ptr_e == 'm')
		val *= 1024 * 1024;

	snd_ftssp_pcm_hw.period_bytes_max = (size_t)val;

	if ((val * snd_ftssp_pcm_hw.periods_max) >
	    snd_ftssp_pcm_hw.buffer_bytes_max) {
		INFO("\nWarning: period_bytes(%d) * periods(%d) exceeds "
			"hw_buffer_size(%d).\n",
			snd_ftssp_pcm_hw.period_bytes_max,
			snd_ftssp_pcm_hw.periods_max,
			snd_ftssp_pcm_hw.buffer_bytes_max);
		INFO("         Unexpected access violation may occur!\n");
	}
}

static void snd_ftssp_periods_min_read(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	snd_iprintf(buffer, "%d\n", snd_ftssp_pcm_hw.periods_min);
}

static void snd_ftssp_periods_min_write(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	char tmp[128];
	char *ptr_e;
	u32 val;

	if (buffer->size == 0)
		return;

	memset(tmp, 0, 128);
	snd_info_get_str(tmp, buffer->buffer, 127);

	val = simple_strtoul(tmp, &ptr_e, 10);
	if (*ptr_e == 'k')
		val *= 1024;
	else if (*ptr_e == 'm')
		val *= 1024 * 1024;

	snd_ftssp_pcm_hw.periods_min = (size_t)val;

	if ((val * snd_ftssp_pcm_hw.period_bytes_max) >
	    snd_ftssp_pcm_hw.buffer_bytes_max) {
		INFO("\nWarning: period_bytes(%d) * periods(%d) exceeds "
			"hw_buffer_size(%d).\n",
			snd_ftssp_pcm_hw.period_bytes_max,
			snd_ftssp_pcm_hw.periods_min,
			snd_ftssp_pcm_hw.buffer_bytes_max);
		INFO("         Unexpected access violation may occur!\n");
	}
}

static void snd_ftssp_periods_max_read(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	snd_iprintf(buffer, "%d\n", snd_ftssp_pcm_hw.periods_max);
}

static void snd_ftssp_periods_max_write(struct snd_info_entry *entry,
	struct snd_info_buffer *buffer)
{
	char tmp[128];
	char *ptr_e;
	u32 val;

	if (buffer->size == 0)
		return;

	memset(tmp, 0, 128);
	snd_info_get_str(tmp, buffer->buffer, 127);

	val = simple_strtoul(tmp, &ptr_e, 10);
	if (*ptr_e == 'k')
		val *= 1024;
	else if (*ptr_e == 'm')
		val *= 1024 * 1024;

	snd_ftssp_pcm_hw.periods_max = (size_t)val;

	if ((val * snd_ftssp_pcm_hw.period_bytes_max) >
	    snd_ftssp_pcm_hw.buffer_bytes_max) {
		INFO("\nWarning: period_bytes(%d) * periods(%d) exceeds "
			"hw_buffer_size(%d).\n",
			snd_ftssp_pcm_hw.period_bytes_max,
			snd_ftssp_pcm_hw.periods_max,
			snd_ftssp_pcm_hw.buffer_bytes_max);
		INFO("         Unexpected access violation may occur!\n");
	}
}
#endif  //FTSSP_PROC_FS

static inline void ftssp_ac97_init(void)
{
	//driver_name = AC97_DRIVER_NAME;
	//codec_info = AC97_CODEC_NAME;

	/* Change codec-dependent callbacks to AC97 */
	ftssp010_substreams[0].pmu_set_clocking = pmu_set_ac97_clocking;
	ftssp010_substreams[0].hw_config        = ftssp010_config_ac97_play;
	ftssp010_substreams[1].pmu_set_clocking = pmu_set_ac97_clocking;
	ftssp010_substreams[1].hw_config        = ftssp010_config_ac97_rec;

	snd_ftssp_playback_ops.copy = snd_ftssp_playback_copy;
	snd_ftssp_capture_ops.copy  = snd_ftssp_capture_copy;

	snd_ftssp_pcm_hw.rates    		= AC97_CODEC_SAMPLE_RATES;
	snd_ftssp_pcm_hw.rate_min 		= AC97_CODEC_SAMPLE_RATE_MIN;
	snd_ftssp_pcm_hw.rate_max 		= AC97_CODEC_SAMPLE_RATE_MAX;
	snd_ftssp_pcm_hw.formats		= AC97_CODEC_FORMATS;
	snd_ftssp_pcm_hw.buffer_bytes_max	= AC97_HW_BUFFER_BYTES_MAX;
	snd_ftssp_pcm_hw.period_bytes_min	= AC97_HW_PERIOD_BYTES_MIN;
	snd_ftssp_pcm_hw.period_bytes_max	= AC97_HW_PERIOD_BYTES_MAX;
	snd_ftssp_pcm_hw.periods_min		= AC97_HW_PERIODS_MIN;
	snd_ftssp_pcm_hw.periods_max		= AC97_HW_PERIODS_MAX;
}

static int ftssp_alsa_init(struct i2c_client *client)
{
	ftssp_chip *chip;
	int err;

	//ADD by river 2011.03.08
	//g_i2c_client = client;
	//End ADD by river 2011.03.08
	//printk(">>>>>>>>> (4) ftssp_alsa_init().\n");
	if(init_hw(cardno, ac97, client))
		return -EIO;

	if (ac97)
		ftssp_ac97_init();

	DBG("%s: FTSSP010 #%d (Physical Addr=0x%08X), mode: %s\n",
		__func__,
		cardno, SSP_FTSSP010_pa_base[cardno],
		ac97 ? "ac97" : "i2s");

	err = snd_card_create(cardno, FTSSP_CARD_ID, THIS_MODULE,
			sizeof(ftssp_chip), &ftssp_cards[cardno]);
	if (err < 0)
		return err;

	if (ac97) {
		sprintf(ftssp_cards[cardno]->driver, FTSSP_DRIVER_NAME);
		sprintf(ftssp_cards[cardno]->shortname,
			FTSSP_DRIVER_NAME "_ac97");
		sprintf(ftssp_cards[cardno]->longname,
			FTSSP_DRIVER_NAME "_ac97 controller");
	} else {
		sprintf(ftssp_cards[cardno]->driver, FTSSP_DRIVER_NAME);
		sprintf(ftssp_cards[cardno]->shortname,
			FTSSP_DRIVER_NAME "_i2s");
		sprintf(ftssp_cards[cardno]->longname,
			FTSSP_DRIVER_NAME "_i2s controller");
	}

	// PCM
	chip = (ftssp_chip *)(ftssp_cards[cardno]->private_data);
	chip->card = ftssp_cards[cardno];

	if ((err = snd_ftssp_new_pcm(chip))) {
		ERR("%s, Can't new PCM devices\n",__func__);
		return -ENODEV;
	}

#if (FTSSP_PROC_FS)
	// new a proc entries subordinate to card->proc_root for debugging
	// /proc/card#/buf_max
	snd_card_proc_new(chip->card, "buf_max", &chip->info_buf_max);
	if (chip->info_buf_max) {
		chip->info_buf_max->c.text.read = snd_ftssp_buf_max_read;
		chip->info_buf_max->c.text.write = snd_ftssp_buf_max_write;
	}
	// /proc/card#/period_min
	snd_card_proc_new(chip->card, "period_size_min",
		&chip->info_period_min);
	if (chip->info_period_min) {
		chip->info_period_min->c.text.read = snd_ftssp_period_min_read;
		chip->info_period_min->c.text.write =
			snd_ftssp_period_min_write;
	}
	// /proc/card#/period_max
	snd_card_proc_new(chip->card, "period_size_max",
		&chip->info_period_max);
	if (chip->info_period_max) {
		chip->info_period_max->c.text.read = snd_ftssp_period_max_read;
		chip->info_period_max->c.text.write =
			snd_ftssp_period_max_write;
	}
	// /proc/card#/periods_min
	snd_card_proc_new(chip->card, "periods_min", &chip->info_periods_min);
	if (chip->info_periods_min) {
		chip->info_periods_min->c.text.read =
			snd_ftssp_periods_min_read;
		chip->info_periods_min->c.text.write =
			snd_ftssp_periods_min_write;
	}
	// /proc/card#/periods_max
	snd_card_proc_new(chip->card, "periods_max", &chip->info_periods_max);
	if (chip->info_periods_max) {
		chip->info_periods_max->c.text.read =
			snd_ftssp_periods_max_read;
		chip->info_periods_max->c.text.write =
			snd_ftssp_periods_max_write;
	}
#endif

	// Register the card to ALSA
	if ((err = snd_card_register(chip->card)) == 0) {
		INFO("%s card registered!\n", FTSSP_CARD_ID);
	}
	return err;
}

#ifdef CONFIG_SND_FTSSP010_I2S
//ADD by river 2011.01.26
static int alc5630_i2c_remove(struct i2c_client *client)
{
	struct alc5630_data *alc5630 = i2c_get_clientdata(client);

	// power down codec chip
	//tas_write_reg(tas, TAS_REG_ACR, 1, &tmp);

	mutex_destroy(&alc5630->mtx);
	kfree(alc5630);
	return 0;
}

static const struct i2c_device_id alc5630_i2c_id[] = {
	{ "alc5630_codec", 0 },
	{ }
};

// This is the driver that will be inserted
static struct i2c_driver alc5630_driver = {
	.driver = {
		.name	= "alc5630_codec",
		.owner = THIS_MODULE,
	},
	.attach_adapter	= alc5630_i2c_attach,
	.probe = alc5630_i2c_probe,
	.remove = alc5630_i2c_remove,
	.suspend = alc5630_i2c_suspend,
	.resume = alc5630_i2c_resume,
	.id_table = alc5630_i2c_id,
};

static int alc5630_i2c_suspend(struct i2c_client *i2c_client, pm_message_t mesg)
{
	//printk("@@@@@ TRACE by river 2011.05.10 : alc5630_i2c_suspend() is invoked.\n");

	//i2c_del_driver(&alc5630_driver);
	//#ifndef CONFIG_SND_FTSSP010_AC97
	//	i2c_del_driver(&alc5630_driver);
	//#endif
	//printk("@@@@@ TRACE by river 2011.05.10 : alc5630_i2c_suspend() -2 is invoked.\n");
	dmad_channel_free(&dma_chreq_tx);
	dmad_channel_free(&dma_chreq_rx);
    //printk("@@@@@ TRACE by river 2011.05.10 : alc5630_i2c_suspend() -3 is invoked.\n");
	//snd_card_free(ftssp_cards[cardno]);

	return 0;

}

static int alc5630_i2c_resume(struct i2c_client *i2c_client)
{

    //printk("@@@@@ TRACE by river 2011.05.10 : alc5630_i2c_resume() is invoked.\n");
    //#ifdef CONFIG_SND_FTSSP010_AC97
	//	ftssp_alsa_init(NULL);
	//#else
	//	return i2c_add_driver(&alc5630_driver);
	//#endif

    return 0;

}
#endif

static __init int ftssp_alsa_i2c_i2s_init(void)
{
	#ifdef CONFIG_SND_FTSSP010_AC97
		return ftssp_alsa_init(NULL);
	#else
		return i2c_add_driver(&alc5630_driver);
	#endif
}

static int ftssp_alsa_i2c_i2s_exit(void)
{
	DBG("%s, cleaning up\n",__func__);

	#ifndef CONFIG_SND_FTSSP010_AC97
		i2c_del_driver(&alc5630_driver);
	#endif

	dmad_channel_free(&dma_chreq_tx);
	dmad_channel_free(&dma_chreq_rx);

	snd_card_free(ftssp_cards[cardno]);

	return 0;
}

static int atf_ac97_probe(struct platform_device *pdev)
{
	struct resource *r;
	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	ssp2_pbase = r->start;
	ssp2_vbase = (resource_size_t)devm_ioremap_resource(&pdev->dev, r);
	return ftssp_alsa_i2c_i2s_init();
}

static int atf_ac97_remove(struct platform_device *pdev)
{
	return ftssp_alsa_i2c_i2s_exit();
}

static const struct of_device_id atf_ac97_of_match[] = {
	{ .compatible = "andestech,atfac97", },
	{},
};

static struct platform_driver atf_ac97_driver = {
	.driver = {
		.name = "atfssp",
		.owner = THIS_MODULE,
		.of_match_table = atf_ac97_of_match,
	},
	.probe = atf_ac97_probe,
	.remove = atf_ac97_remove,
};
module_platform_driver(atf_ac97_driver);

