/* FTSSP010 - UDA1345TS module
 *
 *      $log$
 */

#include <linux/init.h>
#include <linux/module.h>
#include <asm/io.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <asm/dmad.h>
#include <linux/dma-mapping.h>
#include "FTSSP010_UDA1345TS.h"

#if 0
MODULE_LICENSE("Faraday License");
MODULE_AUTHOR("Faraday Technology Corp.");
MODULE_DESCRIPTION("FTSSP010 - UDA1345TS Linux 2.6 Library");
#endif

resource_size_t	ssp2_vbase;
resource_size_t	ssp2_pbase;

#ifdef CONFIG_PLAT_AG101P
#define PMU_PDLLCR1		(pmu_base+0x34)
#define PMU_MFPSR		(pmu_base+0x28)
#define PMU_I2SAC97_REQACKCFG	(pmu_base+0xbc)
#define PMU_C4			(pmu_base+0xc4)
#endif
#define SSPCLK_TO_SCLKDIV(sspclk_div2,bps)	((sspclk_div2)/(bps)-1)

// Each client has this additional data
struct alc5630_data {
	struct i2c_client *client;
	struct delayed_work     work;
	unsigned long gpio2_value;
	struct mutex mtx;
};

//ADD by river 2011.01.26
static int i2s_alc5630_read(unsigned int raddr, char *data, struct i2c_client *client)
{
#ifndef CONFIG_SND_FTSSP010_AC97
	struct i2c_adapter *adap =  client->adapter;
	int ret;
#endif
	struct i2c_msg msg;
	int i2c_value;

	//Reading ALC5630 register
	msg.addr = raddr;
	msg.flags = (client->flags & I2C_M_TEN) | I2C_M_RD;
	msg.len = 1;
	msg.buf = (char *)data;

	//ret = i2c_transfer(adap, &msg, 1);
#ifndef CONFIG_SND_FTSSP010_AC97
	ret = i2c_transfer(adap, &msg, 1);
	if (ret != 0) {
		printk("i2c read failed\n");
		return -1;
	}
	else
#endif
	{
		i2c_value = (data[0]&0xff) << 8 | (data[1]&0xff);
		return i2c_value;
	}
}

static void i2s_alc5630_write(unsigned int raddr, unsigned int data, struct i2c_client *client)
{
#ifndef CONFIG_SND_FTSSP010_AC97
	struct i2c_adapter *adap =  client->adapter;
	int ret;
#endif
	struct i2c_msg msg;
	char buf[3];
	//Writing ALC5630 register
	msg.addr = raddr;
	msg.flags = (client->flags & I2C_M_TEN) | ~I2C_M_RD;
	msg.len = 1;
	buf[0] = (data >> 8) & 0xff;
	buf[1] =  data & 0xff;
	msg.buf = (char *)buf;
	//ret = i2c_transfer(adap, &msg, 1);
#ifndef CONFIG_SND_FTSSP010_AC97
	ret = i2c_transfer(adap, &msg, 1);
	if (ret != 0)
	{
		printk("i2c write failed\n");
	}
#endif
}

static void i2s_al5630_slave_stereo_mode(struct i2c_client *client)
{
	i2s_alc5630_write(0x34, 0x8000, client); // codec slave mode
	i2s_alc5630_write(0x0c, 0x1010, client);
	i2s_alc5630_write(0x10, 0xee03, client);
    i2s_alc5630_write(0x1c, 0x0748, client);
    //i2s_alc5630_write(0x02, 0x8080, client);
    //i2s_alc5630_write(0x04, 0x8888, client);
    i2s_alc5630_write(0x62, 0x0000, client);

}

//End ADD by river 2011.01.26


/* Drive PMU to generate I2S main clocking signal. Also configures PMU to set correct DMA REQ/ACK pair */
void pmu_set_i2s_clocking(unsigned int speed)
{
	unsigned int pmu_pdllcr1; /* PLL/DLL Control Register 1 */
	/* Configure PMU to generate I2S main clock */
	#ifdef CONFIG_PLAT_AG101
	pmu_pdllcr1 = inl(PMU_PDLLCR1)&0xfff0ffff; /* Bit 19-16 are relevent */
	#endif

	switch (speed) {
	case 8000:
		pmu_pdllcr1 |= 0x00000000; /* 2.048MHz x2 */
		break;
	case 11025:
		pmu_pdllcr1 |= 0x00010000; /* 2.8224MHz x2 */
		break;
	case 16000:
		pmu_pdllcr1 |= 0x00020000; /* 4.096MHz x2 */
		break;
	case 22050:
		pmu_pdllcr1 |= 0x00030000; /* 5.6448MHz x2 */
		break;
	case 32000:
		pmu_pdllcr1 |= 0x00040000; /* 8.192MHz x2 */
		break;
	case 44100:
		pmu_pdllcr1 |= 0x00050000; /* 11.2896Mhz x2 */
		break;
	case 48000:
		pmu_pdllcr1 |= 0x00060000; /* 12.2880MHz x2 */
		break;
	default:
		printk("%s: Unknown i2s speed %d\n",__func__,speed);
	};

	#ifdef CONFIG_PLAT_AG101
	outl(pmu_pdllcr1, PMU_PDLLCR1);
	/* Configure PMU to select I2S output (instead of AC97) */
	outl(inl(PMU_MFPSR)&(~(1<<3)), PMU_MFPSR); /* clear bit 3 of MFPSR*/
	#endif
}

/* Drive PMU to generate AC97 main clocking signal. Also configures PMU to set correct DMA REQ/ACK pair */
void pmu_set_ac97_clocking(unsigned int speed)
{
	/* Configure PMU to select AC97 output (instead of I2S) */
	/* Set GPIO[26] to AC97 clock, use 49.152MHz main clock (AC97 CLK1) */
	//outl(inl(PMU_MFPSR)|((1<<13)|(1<<3)), PMU_MFPSR); /* Set bit 13 & 3 of MFPSR*/
	#ifndef CONFIG_PLAT_AG102
        #ifndef CONFIG_PLAT_AE3XX
		outl(inl(PMU_MFPSR)|((1<<13)|(1<<3)), PMU_MFPSR); /* Set bit 13 & 3 of MFPSR*/
	#endif
        #endif
}

/* Programs PMU to set I2S/AC97 DMA Channel, ch=0-7 */
void pmu_set_i2s_dma_channel(unsigned ch)
{
	#ifdef CONFIG_PLAT_AG101
	ch&=0x7;
	//outl((inl(PMU_I2SAC97_REQACKCFG)&(~0x7))|ch, PMU_I2SAC97_REQACKCFG);
	outl(0xa, PMU_I2SAC97_REQACKCFG);
	outl(0xb, PMU_C4);
	#endif
}

//ADD by river 2011.02.11
static struct i2c_client *g_i2c_client;




void ftssp010_set_int_control(int cardno, unsigned val)
{
	outl(val, FTSSP010_INT_CONTROL(cardno));
}

unsigned ftssp010_get_int_status(int cardno)
{
	return (inl(FTSSP010_INT_STATUS(cardno)));
}

int ftssp010_get_status(int cardno)
{
	return (inl(FTSSP010_STATUS(cardno)));
}

int  ftssp010_tx_fifo_not_full(int cardno)
{
	return (inl(FTSSP010_STATUS(cardno))&0x2)==0x2;
}

int ftssp010_tx_fifo_vaild_entries(int cardno)
{
	return (inl(FTSSP010_STATUS(cardno))>>12) & 0x1f;
}

#include "FTSSP010_W83972D.h"

// AC97 codec tags
#define TAG_COMMAND	0xe000
#define TAG_DATA	0x9800 /* Slot 3/4 */
#define TAG_DATA_MONO	0x9000 /* Slot 3 */
//#define TAG_DATA_LINE_IN	0x9000	/* Slot 3 */

void ftssp010_ac97_write_codec_start(unsigned int cardno)
{
	outl(0x0,  FTSSP010_INT_CONTROL(cardno));/*Disable interrupts & DMA req */
	outl(0xC, FTSSP010_CONTROL2(cardno));  /* Disable FTSSP010, clear RX/TX Fifo.  */
	outl(TAG_COMMAND, FTSSP010_ACLINK_SLOT_VALID(cardno));
}

void ftssp010_ac97_write_codec(unsigned int cardno,unsigned int reg,unsigned int data)
{
	outl(reg << 12, FTSSP010_DATA(cardno));
	//mdelay(50);
	outl(data << 4, FTSSP010_DATA(cardno));
	//mdelay(50);
}

void ftssp010_ac97_write_codec_commit(unsigned int cardno)
{
	while((inl(FTSSP010_CONTROL2(cardno))&0x1)==0) {
		outl(0x3 , FTSSP010_CONTROL2(cardno)); /* SSPEN + TXDOE */
	}
	while(ftssp010_tx_fifo_vaild_entries(cardno))
		;
	/* Wait for frame completion */
	while((inl(FTSSP010_INT_STATUS(cardno))&0x10)==0)
		;
	outl(0x0, FTSSP010_CONTROL2(cardno));
	/* Wait for busy indicator */
	while((inl(FTSSP010_STATUS(cardno))&0x4)==0x4);
}

/*	Configure FTSSP010 to a given sampling rate and channel number
 *	for AC97 mode in playback mode
 */
int init_hw(unsigned int cardno,unsigned int ac97, struct i2c_client *client)
{
	int cnt = 0x1000000;
	//printk(">>>>>>>>>> (5) init_hw() is called.\n");
	g_i2c_client = client;

	if(ac97)
	{
		//pmu_set_ac97_clocking(48000);
		#ifndef CONFIG_PLAT_AG102
			pmu_set_ac97_clocking(48000);
		#endif
		outl(0x400c, FTSSP010_CONTROL0(cardno)); /* set FTSSP010 to AC97 mode */
		//mdelay(50);
		outl(0xc400,  FTSSP010_INT_CONTROL(cardno));
		if((inl(FTSSP010_INT_CONTROL(cardno)))!=0xc400){
			return -EIO;
		}
		//mdelay(50);
		outl(0x20, FTSSP010_CONTROL2(cardno));  /* Cold Reset AC-Link */
		//mdelay(50);
		while(inl(FTSSP010_CONTROL2(cardno))&&cnt--);
		if(!cnt){
			return -EIO;
		}
		//mdelay(50);
		outl(0x40, FTSSP010_CONTROL2(cardno));  /* Reset AC-Link */
		//mdelay(1500);
	}
	else
	{

		//printk(">>>>>>>>> (6) I2S mode selected....YAYAYA......\n");
		#ifdef CONFIG_PLAT_AG101
		outl(inl(PMU_MFPSR)&(~(1<<3)), PMU_MFPSR); /* clear bit 3 of MFPSR*/
		outl(0xa, PMU_I2SAC97_REQACKCFG);
		outl(0xb, PMU_C4);
		#endif

		//MOD by river 2011.01.26
		//i2s_alc5630_master_stereo_mode(client);
		i2s_al5630_slave_stereo_mode(client);
		//ssp_slave_stereo_mode();
		//End MOD by river 2011.01.26
		outl(0x311c, FTSSP010_CONTROL0(cardno));	/* I2S Master */
		outl(0, FTSSP010_CONTROL1(cardno));	        /* I2S Master */
		outl(0xc400, FTSSP010_INT_CONTROL(cardno));	/* I2S Master */
		outl(0x40, FTSSP010_CONTROL2(cardno));  /* Reset AC-Link */

		//i2s_alc5630_read_test(client);
	}
	return 0;
}
static void _ftssp010_config_ac97(int cardno, unsigned is_stereo, unsigned speed, int is_rec)
{
	int cnt = 0x1000000;
	/* Codec initialization */
	ftssp010_ac97_write_codec_start(cardno);
	ftssp010_ac97_write_codec(cardno, W83972D_RESET, 0);
	ftssp010_ac97_write_codec_commit(cardno);

	//msleep_interruptible(10);
	ftssp010_ac97_write_codec_start(cardno);

	if (is_rec) {	/* Recording */
		/* Mute output */
		//ftssp010_ac97_write_codec(cardno, W83972D_STEREO_OUTPUT_CONTROL, 0x8000);
		/* Mute PCM */
		//ftssp010_ac97_write_codec(cardno, W83972D_PCM_OUTPUT_CONTROL, 0x8000);

		/* Register 0x10, Line-In/Mic Gain */
		ftssp010_ac97_write_codec(cardno, W83972D_LINE_IN_VOLUME, 0x808);
		//ftssp010_ac97_write_codec(cardno, W83972D_AUX_INPUT_CONTROL, 0x808);
		ftssp010_ac97_write_codec(cardno, W83972D_MIC_VOLUME, 0x8);
		/* FIXME: REC from line-in only */

		/* Register 0x1A, Record Select=StereoMix */
		ftssp010_ac97_write_codec(cardno, W83972D_RECORD_SELECT, 0x505 /*404*/);
		/* Register 0x1C, Record Gain=0db */
		ftssp010_ac97_write_codec(cardno, W83972D_RECORD_GAIN, 0x808);
		ftssp010_ac97_write_codec(cardno, W83972D_RECORD_GAIN_MIC, 0x8);
	} else {	/* Playback */
		/* Register 0x10, Mute Line-In/Mic Gain */
		ftssp010_ac97_write_codec(cardno, W83972D_LINE_IN_VOLUME, 0x8000);
		ftssp010_ac97_write_codec(cardno, W83972D_MIC_VOLUME, 0x8000);

		/* Register 0x1A, Mute Record Gains */
		ftssp010_ac97_write_codec(cardno, W83972D_RECORD_GAIN, 0x8000);
		ftssp010_ac97_write_codec(cardno, W83972D_RECORD_GAIN_MIC, 0x8000);

		/* Output */
		ftssp010_ac97_write_codec(cardno, W83972D_STEREO_OUTPUT_CONTROL, 0);
		ftssp010_ac97_write_codec(cardno, W83972D_PCM_OUTPUT_CONTROL, 0x808);
	}

#if 0
	ftssp010_ac97_write_codec(cardno, W83972D_EXT_AUDIO_CONTROL, 0x1);
	ftssp010_ac97_write_codec(cardno, W83972D_DAC_SAMPLE_RATE_CONTROL, speed);
#endif

	ftssp010_ac97_write_codec_commit(cardno);
	//msleep_interruptible(10);

	/* Start data transfer */
//	if(is_rec) {
//		outl(TAG_DATA_LINE_IN, FTSSP010_ACLINK_SLOT_VALID(cardno));
//	} else {
		if(is_stereo)
			outl(TAG_DATA, FTSSP010_ACLINK_SLOT_VALID(cardno));
		else
			outl(TAG_DATA_MONO, FTSSP010_ACLINK_SLOT_VALID(cardno));
//	}
	while((inl(FTSSP010_INT_STATUS(cardno))&0x3)&&cnt--);
//		msleep_interruptible(10);
}

void ftssp010_config_ac97_play(int cardno, unsigned is_stereo, unsigned speed, int use8bit)
{
	_ftssp010_config_ac97(cardno, is_stereo, speed, 0);
}

void ftssp010_config_ac97_rec(int cardno, unsigned is_stereo, unsigned speed, int use8bit)
{
	_ftssp010_config_ac97(cardno, is_stereo, speed, 1);
}
/*
 *	Configure FTSSP010 to a given sampling rate and channel number
 *	for I2S mode
 */
void ftssp010_config(int cardno, unsigned is_stereo, unsigned speed, int width, int is_rec)
{
	int use8bit = (width == 1 ? 1 : 0);
	unsigned opm, bps = 2 * (use8bit ? 8 : 16);	/* bits per 1 second audio data. */
	unsigned fpclkdiv = 0;

	//ADD by river 2011.06.02
	struct alc5630_data *alc5630;
	char data[3];
	opm = is_stereo ? FTSSP010_CONTROL0_OPM_STEREO : FTSSP010_CONTROL0_OPM_MONO;
	//MOD by river 2011.01.27
	outl(0x3100 | opm, FTSSP010_CONTROL0(cardno));	/* I2S Master */
	//End MOD by river 2011.01.27

#if 0
	printk("%s: use %dHz %d-bit %s \n",__func__,
		speed,
		use8bit?8:16,
		is_stereo?"stereo":"mono"
		);
#endif

	/* configures CONTROL1 to use suitable clock divider.
           the I2S clock is generated from PMU. */
	bps *= speed;
	switch(speed) {
	case 8000:	/* SCLK : 256KHZ */
	    i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client); //?
		i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
	    //i2s_alc5630_write(0x44, 0x6a0, g_i2c_client); //?
		//i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    //i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
	    fpclkdiv = 0xBB;

		//fpclkdiv=SSPCLK_TO_SCLKDIV(2048000, bps);
		break;
	case 11025:	/* SCLK : 352.8KHZ */
	    i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client); //?
	    i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
		fpclkdiv = 0x88;
		//fpclkdiv=SSPCLK_TO_SCLKDIV(2822400, bps);
		break;
	case 16000:	/* SCLK : 512KHZ */
	    i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client); //?
		i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
	    fpclkdiv = 0x5f;
		//fpclkdiv=SSPCLK_TO_SCLKDIV(4096000, bps);
		break;
	case 22050:	/* SCLK : 705.6KHZ */
	    i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client);
	    i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
	    fpclkdiv = 0x45;
		//fpclkdiv=SSPCLK_TO_SCLKDIV(5644800, bps);
		break;
	case 24000:	/* SCLK : 768KHZ */
	    i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client);
	    i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
	    fpclkdiv = 0x3e;
		//fpclkdiv=SSPCLK_TO_SCLKDIV(5644800, bps);
		break;
	case 32000:	/* SCLK : 1024KHZ */
	    i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client);
		i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
		//fpclkdiv = 0x2e;
		fpclkdiv = 0x2f;
		//fpclkdiv=SSPCLK_TO_SCLKDIV(8192000, bps);
		break;
	case 44100:	/* SCLK : 1.4112 MHZ */ /* 96 MHZ */
		i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client);
		i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
		fpclkdiv = 0x22;
		//fpclkdiv=SSPCLK_TO_SCLKDIV(11289600, bps);
		break;
	case 48000: /* SCLK : 1.536 MHZ */
	    i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client);
		i2s_alc5630_write(0x60, 0x3174, g_i2c_client);
	    i2s_alc5630_write(0x62, 0x1010, g_i2c_client);
	    fpclkdiv = 0x1f;
		//fpclkdiv=SSPCLK_TO_SCLKDIV(12288000, bps);

		break;
	default:
		printk("%s: unsupported speed %d\n", __func__,speed);
		return;
	};


	if(!use8bit) {
		outl(0xf0000|fpclkdiv, FTSSP010_CONTROL1(cardno));	/*  16bits */
		//outl(0xf0000|0x10, FTSSP010_CONTROL1(cardno));	/*  16bits */
		//outl(0xf0000|0x22, FTSSP010_CONTROL1(cardno));	/*  16bits */
	} else {
		outl(0x70000|fpclkdiv, FTSSP010_CONTROL1(cardno));	/*  8bits */
	}

	//printk("#####$$$$$ : bps = %d\n", bps);
	//printk("#####$$$$$ : speed = %d\n", speed);
	//printk("#####$$$$$ : fpclkdiv = 0x%08x\n", fpclkdiv);
	//printk("#####$$$$$ : FTSSP010_CONTROL1(cardno) = 0x%08x\n", inl(FTSSP010_CONTROL1(cardno)));

	if(is_rec)
		outl(inl(FTSSP010_INT_CONTROL(cardno))&(~0x0f15),  FTSSP010_INT_CONTROL(cardno));		/* Disable all interrupts */
	else
		outl(inl(FTSSP010_INT_CONTROL(cardno))&(~0xf02a) ,  FTSSP010_INT_CONTROL(cardno));		/* Disable all interrupts */

	outl(0xc, FTSSP010_CONTROL2(cardno));	/* clear FIFOs */

	//ADD by river 2011.06.02
	alc5630 = i2c_get_clientdata(g_i2c_client);
	//End

	if(is_rec) {
		printk("ftssp010_config() for I2S mode in record.\n");
		//ADD by river 2011.03.21
		//TEST by river 2011.03.22 for recording => workable
		//i2s_alc5630_write(0x0e, 0x8888, g_i2c_client);
		i2s_alc5630_write(0x0e, 0x0808, g_i2c_client);
		i2s_alc5630_write(0x10, 0xee03, g_i2c_client);
		i2s_alc5630_write(0x22, 0x0500, g_i2c_client);
		i2s_alc5630_write(0x1c, 0x0748, g_i2c_client);
		i2s_alc5630_write(0x14, 0x1f1f, g_i2c_client);
		i2s_alc5630_write(0x12, 0xdfdf, g_i2c_client);

		i2s_alc5630_write(0x26, 0x000f, g_i2c_client);
		i2s_alc5630_write(0x3a, 0xffff, g_i2c_client);
		i2s_alc5630_write(0x3c, 0xffff, g_i2c_client);
		//i2s_alc5630_write(0x3e, 0xffff, g_i2c_client);
		i2s_alc5630_write(0x3e, 0x80cf, g_i2c_client);

		i2s_alc5630_write(0x44, 0x3ea0, g_i2c_client);
		i2s_alc5630_write(0x42, 0x2000, g_i2c_client);
		i2s_alc5630_write(0x40, 0x8c0a, g_i2c_client);

		//i2s_alc5630_write(0x02, 0x0000, g_i2c_client);
		i2s_alc5630_write(0x02, 0x8080, g_i2c_client);
		i2s_alc5630_write(0x04, 0x0000, g_i2c_client);
		//End TEST by river 2011.03.22

		//printk("ftssp010_config() for I2S mode (Recording) ===> Dump register.\n");
		//i2s_alc5630_read_test(g_i2c_client);

    }
    else {
	    //printk("ftssp010_config() for I2S mode in playback.\n");
	    //ADD by river 2011.03.24 for record and playback case
	    i2s_alc5630_write(0x0e, 0x0808, g_i2c_client);
	    i2s_alc5630_write(0x12, 0xcbcb, g_i2c_client);
	    i2s_alc5630_write(0x14, 0x7f7f, g_i2c_client);
	    i2s_alc5630_write(0x22, 0x0000, g_i2c_client);
	    i2s_alc5630_write(0x3e, 0x8000, g_i2c_client);
	    i2s_alc5630_write(0x40, 0x0c0a, g_i2c_client);
	    i2s_alc5630_write(0x42, 0x0000, g_i2c_client);
	    //End ADD by river 2011.03.24 for record and playback case

	    //TEST by river 2011.03.23
	i2s_alc5630_write(0x26, 0x0000, g_i2c_client);
	i2s_alc5630_write(0x3c, 0x2000, g_i2c_client);
	i2s_alc5630_write(0x3a, 0x0002, g_i2c_client);
	i2s_alc5630_write(0x3c, 0xa330, g_i2c_client);
	i2s_alc5630_write(0x3a, 0xc843, g_i2c_client);
	//End TEST by river 2011.03.23

	//ADD by river 2011.03.23 for HP Out De-pop
	i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,g_i2c_client)|0x0002 , g_i2c_client);
	i2s_alc5630_write(0x04, i2s_alc5630_read(0x04, data,g_i2c_client)|0x8080 , g_i2c_client);
	i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,g_i2c_client)|0x0040 , g_i2c_client);
	i2s_alc5630_write(0x3c, i2s_alc5630_read(0x3C, data,g_i2c_client)|0x2000 , g_i2c_client);
	i2s_alc5630_write(0x3E, i2s_alc5630_read(0x3E, data,g_i2c_client)|0xfC00 , g_i2c_client);
	i2s_alc5630_write(0x5E, i2s_alc5630_read(0x5E, data,g_i2c_client)|0x0100 , g_i2c_client);
	//mdelay(500);
	i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,g_i2c_client)|0x0200 , g_i2c_client);
	i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,g_i2c_client)|0x0100 , g_i2c_client);
	i2s_alc5630_write(0x5E, i2s_alc5630_read(0x5E, data,g_i2c_client)& 0xfeff ,g_i2c_client);
	//End ADD by river 2011.03.23 for HP Out De-pop


	    //TEST by river 2011.03.22
	    i2s_alc5630_write(0x1c, 0x0748, g_i2c_client);
	    //End TEST by river 2011.03.22
		//TEST by river 2011.03.16
	    //i2s_alc5630_write(0x02, 0x8080, g_i2c_client);
		i2s_alc5630_write(0x26, 0x000f, g_i2c_client);
		//End TEST by river 2011.03.16

		//ADD by river 2011.03.23
		if (alc5630->gpio2_value==0x0)
		    i2s_alc5630_write(0x3A, (i2s_alc5630_read(0x3A, data,g_i2c_client) & 0xFBFF)|0x0040 , g_i2c_client);
		else
			i2s_alc5630_write(0x3A, i2s_alc5630_read(0x3A, data,g_i2c_client)|0x0440 , g_i2c_client);


		i2s_alc5630_write(0x5E, i2s_alc5630_read(0x5E, data,g_i2c_client)|0x0020 , g_i2c_client);
		i2s_alc5630_write(0x5E, i2s_alc5630_read(0x5E, data,g_i2c_client)|0x00c0 , g_i2c_client);
		i2s_alc5630_write(0x04, i2s_alc5630_read(0x04, data,g_i2c_client)& 0x7f7f , g_i2c_client);
		//mdelay(30);

		if (alc5630->gpio2_value==0x0) {
			//printk(">>>>>>>>>>>> Turn off internal speaker.....\n");
			i2s_alc5630_write(0x02, 0x5F5F, g_i2c_client);
		    //i2s_alc5630_write(0x02, 0x0000, g_i2c_client);
	    }
		else  {
			//printk(">>>>>>>>>>>> Turn on internal speaker.....\n");
			i2s_alc5630_write(0x02, 0x0000, g_i2c_client);
		    //i2s_alc5630_write(0x02, 0x5F5F, g_i2c_client);
	    }
		//End ADD by river 2011.03.23


		//printk("ftssp010_config() for I2S mode (Playback) ===> Dump register.\n");
		//i2s_alc5630_read_test(g_i2c_client);
    }

#if 0
	/* Stuff TX fifo */
	while(ftssp010_tx_fifo_not_full(cardno)) {
		outl(0x0, FTSSP010_DATA(cardno));
	}
#endif
}

void ftssp010_config_tx(int cardno, unsigned is_stereo, unsigned speed, int width)
{
	return ftssp010_config(cardno, is_stereo, speed, width, 0);
}

void ftssp010_config_rx(int cardno, unsigned is_stereo, unsigned speed, int width)
{
	return ftssp010_config(cardno, is_stereo, speed, width, 1);
}

/* Configures FTSSP010 to start TX. If use_dma being nonzero,
 * FTSSP010 will use hardware handshake for DMA */
void ftssp010_start_tx(int cardno, unsigned use_dma)
{
	unsigned bogus=0x800*3;
	if(use_dma) {
		/* Enable H/W DMA Request and set TX DMA threshold to 12*/
//		outl(0xC022, FTSSP010_INT_CONTROL(cardno));
                outl(inl(FTSSP010_INT_CONTROL(cardno)) | 0xc422, FTSSP010_INT_CONTROL(cardno));
#if 0
		printk("%s: enable DMA request\n", __func__);
#endif
	}
//	outl(0x3, FTSSP010_CONTROL2(cardno));
        outl(inl(FTSSP010_CONTROL2(cardno)) | 0x3, FTSSP010_CONTROL2(cardno));
//	printk("%s\n",__func__);
//	printk("int_control 0x%x\n",inl(FTSSP010_INT_CONTROL(cardno)));
//	printk("control2 0x%x\n",inl(FTSSP010_CONTROL2(cardno)));
	if(!use_dma) {
		while(bogus>0) {
			while(!ftssp010_tx_fifo_not_full(cardno))
				udelay(50);
			outl(0, FTSSP010_DATA(cardno));
			bogus--;
		}
	}
}

/* Configures FTSSP010 to start RX. If use_dma being nonzero,
 * FTSSP010 will use hardware handshake for DMA */
void ftssp010_start_rx(int cardno, unsigned use_dma)
{
        if(use_dma) {
                /* Enable H/W DMA Request and set RX DMA threshold to 2*/
//		outl(0x0111, FTSSP010_INT_CONTROL(cardno));
                outl(inl(FTSSP010_INT_CONTROL(cardno)) | 0xc411, FTSSP010_INT_CONTROL(cardno));
#if 0
                printk("%s: enable DMA request\n", __func__);
#endif
        }
//	outl(0x3, FTSSP010_CONTROL2(cardno));
        outl(inl(FTSSP010_CONTROL2(cardno)) | 0x3, FTSSP010_CONTROL2(cardno));
//	printk("%s\n",__func__);
//	printk("int_control 0x%x\n",inl(FTSSP010_INT_CONTROL(cardno)));
//	printk("control2 0x%x\n",inl(FTSSP010_CONTROL2(cardno)));
}

void ftssp010_stop_tx(int cardno)
{
//	outl(0, FTSSP010_CONTROL2(cardno));
	outl(inl(FTSSP010_INT_CONTROL(cardno)) & (~0x22), FTSSP010_INT_CONTROL(cardno));
//	printk("%s\n",__func__);
//	printk("int_control 0x%x\n",inl(FTSSP010_INT_CONTROL(cardno)));
//	printk("control2 0x%x\n",inl(FTSSP010_CONTROL2(cardno)));
}

void ftssp010_stop_rx(int cardno)
{
	//outl(0, FTSSP010_CONTROL2(cardno));
	outl(inl(FTSSP010_INT_CONTROL(cardno)) & (~0x11), FTSSP010_INT_CONTROL(cardno));
	//printk("%s\n",__func__);
	//printk("int_control 0x%x\n",inl(FTSSP010_INT_CONTROL(cardno)));
	//printk("control2 0x%x\n",inl(FTSSP010_CONTROL2(cardno)));
}

